﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Globalization;

using CoboltMonitor.Laser;

namespace CoboltMonitor.Gui
{
    public partial class PIDMonitor1Tec : Form
    {
        #region Data
        public delegate void UpdateSerialData(Object objSender, LaserUpdatedEvent dataReceivedEventArgs);

        List<float>[] _temps;				// temperature history
        List<float>[] _drives;				// drive % history
        Tec[] _tecs;				// tec values
        float[] _time;

        protected int _tec = 0;
        protected LaserCommon laser;
        #endregion

        #region Properties
        public int MonitoredTec
        {
            get { return _tec; }
        }
        #endregion

        #region Constructors

        public PIDMonitor1Tec()
        {
            InitializeComponent();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PIDMonitor"/> class.
        /// </summary>
        public PIDMonitor1Tec(LaserCommon laser)
        {
            InitializeComponent();
            this.laser = laser;
            laser.LaserUpdatedEvent += new EventHandler<LaserUpdatedEvent>(UpdateGui);

            _time = new float[laser.NumberOfTecs];
            _temps = new List<float>[laser.NumberOfTecs];
            _drives = new List<float>[laser.NumberOfTecs];
            _tecs = new Tec[laser.NumberOfTecs];
            for (int i = 0; i < laser.NumberOfTecs; i++)
            {
                _temps[i] = new List<float>();
                _temps[i].Capacity = 120 * 5;
                _time[i] = 0f;
                _drives[i] = new List<float>();
                _drives[i].Capacity = 120 * 5;
                _tecs[i] = new Tec();
            }

            toolTip1.SetToolTip(usGraph1, "Orange curve is temperature.\rOlive curve is percent drive.");
            toolTip1.SetToolTip(textBoxKp, "Proportional constant");
            toolTip1.SetToolTip(textBoxKi, "Integral constant");
            toolTip1.SetToolTip(textBoxKd, "Differential constant");
            toolTip1.SetToolTip(vScrollBar1, "Use scrollbar to change\rdisplayed temperature range");
            toolTip1.SetToolTip(buttonSaveImage, "Click to save a snapshot of the image.\rYou might want to go to the main form first,\r to disconnect the serial comms.");
        }
        #endregion

        #region Public member functions
        /// <summary>
        /// Plots the specified idx.
        /// </summary>
        /// <param name="idx">The idx.</param>
        /// <param name="tec">The tec.</param>
        public void Plot(int idx, Tec tec)
        {
            _tecs[idx] = tec;
            float temp = tec.Temp;
            float target = tec.SetPoint;
            float drive = tec.Drive;

            // Keep max 120 temperatures in history.
            if (_temps[idx].Count == 120 * 5)
                _temps[idx].RemoveAt(0);

            // Add new temperature.
            _temps[idx].Add(temp);

            // Update time.
            _time[idx] += 0.2F;

            // Keep max 120 drive values in history.
            if (_drives[idx].Count == 120 * 5)
                _drives[idx].RemoveAt(0);

            // Add new drive value.
            _drives[idx].Add(drive);

            // Plot history if length greater than one.
            if (_temps[idx].Count > 1)
            {
                if (idx == _tec)
                {
                    // Get entire temperature history into an array.
                    float[] a = _temps[idx].ToArray();

                    // Get entire drives history into an array.
                    float[] d = _drives[idx].ToArray();

                    float time;

                    if (_time[idx] < 120F)
                    {
                        time = 0F;
                    }
                    else
                    {
                        time = _time[idx] - 120F;
                    }

                    //usGraph1.SetHorizontalBounds(time, time + 120F);

                    if (!Visible)
                        return;

                    // Clear display.
                    usGraph1.Clear();

                    Font fntLegend = new Font("Arial", 8);
                    Brush brLegend = new SolidBrush(Color.Gray);

                    float t = usGraph1.YMin + usGraph1.YTick;
                    while (t <= usGraph1.YMax - usGraph1.YTick)
                    {
                        usGraph1.Legend(t.ToString(), fntLegend, brLegend, new PointF(0, t + 0.4f));
                        t += usGraph1.YTick;
                    }

                    PointF tp0;
                    PointF tp1;

                    // Plot target temperature axis with ticks.
                    //tp0 = new PointF(0, target);
                    //tp1 = new PointF(120, target);
                    //usGraph1.Axis(tp0, tp1, Color.LightBlue);

                    // Plot target temperature upper limit axis with ticks.
                    tp0 = new PointF(0, target + tec.Delta);
                    tp1 = new PointF(120, target + tec.Delta);
                    usGraph1.Axis(tp0, tp1, Color.White);

                    // Plot target temperature lower limit axis with ticks.
                    tp0 = new PointF(0, target - tec.Delta);
                    tp1 = new PointF(120, target - tec.Delta);
                    usGraph1.Axis(tp0, tp1, Color.White);

                    // Plot entire history.
                    time = 0;
                    for (int i = 0; i < _temps[idx].Count - 1; i++)
                    {
                        // Plot temperature curve.
                        PointF p0 = new PointF(time, a[i]);
                        PointF p1 = new PointF(time + 0.2f, a[i + 1]);
                        usGraph1.Line(p0, p1, Color.Orange);

                        // Plot drive curve.
                        float h = usGraph1.YMax - usGraph1.YMin;
                        p0 = new PointF(time, d[i] * h * 0.005f + usGraph1.YMin + h * 0.5f);
                        p1 = new PointF(time + 0.2f, d[i + 1] * h * 0.005f + usGraph1.YMin + h * 0.5f);
                        usGraph1.Line(p0, p1, Color.Olive);

                        // Increment time.
                        time += 0.2F;
                    }

                    // Make it appear on screen.
                    if (idx == 3)
                        usGraph1.Invalidate();
                }
            }
        }

        /// <summary>
        /// Resets this instance.
        /// </summary>
        public void Reset()
        {
            for (int i = 0; i < laser.NumberOfTecs; i++)
                _temps[i].Clear();

            usGraph1.Clear();
        }

        /// <summary>
        /// Fills in the PID constants in the appropriate TextBoxes.
        /// </summary>
        /// <param name="idx">The idx.</param>
        /// <param name="tec">The tec.</param>
        public void PidData(int idx, Tec tec)
        {
            if (idx == _tec)
            {
                textBoxKp.Text = tec.Kp.ToString(new CultureInfo("en-US"));
                textBoxKi.Text = tec.Ki.ToString(new CultureInfo("en-US"));
                textBoxKd.Text = tec.Kd.ToString(new CultureInfo("en-US"));
            }
        }
        #endregion

        #region Private member functions

        private void GetPid()
        {
            laser.GetPidForTec();
        }
        #endregion

        #region Form Events
        private void PIDMonitor_Paint(object sender, PaintEventArgs e)
        {
            usGraph1.Invalidate();
        }
        #endregion

        #region Control Events

        private void buttonReset_Click(object sender, EventArgs e)
        {
            Reset();
        }

        private void buttonClose_Click(object sender, EventArgs e)
        {
            Hide();
        }

        private void textBoxKp_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)Keys.Return)
            {
                float res;

                if (float.TryParse(textBoxKp.Text, NumberStyles.Float, new CultureInfo("en-US"), out res))
                {
                    _tecs[_tec].Kp = res;
                    string args = "";
                    args += _tecs[_tec].Kp.ToString(new CultureInfo("en-US"));
                    args += " ";
                    args += _tecs[_tec].Ki.ToString(new CultureInfo("en-US"));
                    args += " ";
                    args += _tecs[_tec].Kd.ToString(new CultureInfo("en-US"));

                    laser.SetPidForTec(_tec, args);
                }
                else
                {
                    MessageBox.Show("'" + textBoxKp.Text + "' is not a valid floating point number.", "Bad number format");
                    GetPid();
                }
                e.Handled = true;
            }
        }

        private void textBoxKi_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)Keys.Return)
            {
                float res;

                if (float.TryParse(textBoxKi.Text, NumberStyles.Float, new CultureInfo("en-US"), out res))
                {
                    _tecs[_tec].Ki = res;
                    string args = "";
                    args += _tecs[_tec].Kp.ToString(new CultureInfo("en-US"));
                    args += " ";
                    args += _tecs[_tec].Ki.ToString(new CultureInfo("en-US"));
                    args += " ";
                    args += _tecs[_tec].Kd.ToString(new CultureInfo("en-US"));

                    laser.SetPidForTec(_tec, args);
                }
                else
                {
                    MessageBox.Show("'" + textBoxKi.Text + "' is not a valid floating point number.", "Bad number format");
                    GetPid();
                }
                e.Handled = true;
            }
        }

        private void textBoxKd_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)Keys.Return)
            {
                float res;

                if (float.TryParse(textBoxKd.Text, NumberStyles.Float, new CultureInfo("en-US"), out res))
                {
                    _tecs[_tec].Kd = res;
                    string args = "";
                    args += _tecs[_tec].Kp.ToString(new CultureInfo("en-US"));
                    args += " ";
                    args += _tecs[_tec].Ki.ToString(new CultureInfo("en-US"));
                    args += " ";
                    args += _tecs[_tec].Kd.ToString(new CultureInfo("en-US"));

                    laser.SetPidForTec(_tec, args);
                }
                else
                {
                    MessageBox.Show("'" + textBoxKd.Text + "' is not a valid floating point number.", "Bad number format");
                    GetPid();
                }
                e.Handled = true;
            }
        }

        private void buttonSet_Click(object sender, EventArgs e)
        {
            string args = "";
            args += _tecs[_tec].Kp.ToString(new CultureInfo("en-US"));
            args += " ";
            args += _tecs[_tec].Ki.ToString(new CultureInfo("en-US"));
            args += " ";
            args += _tecs[_tec].Kd.ToString(new CultureInfo("en-US"));

            laser.SetPidForTec(args);
        }

        private void buttonGet_Click(object sender, EventArgs e)
        {
            GetPid();
        }

        private void SetVertGraphRange(int tmin, int tmax)
        {
            //usGraph1.YMin = tmin;
            //usGraph1.YMax = tmax;
            usGraph1.SetVerticalBounds(tmin, tmax);
            labelMinTemp.Text = tmin.ToString() + "°";
            labelMaxTemp.Text = tmax.ToString() + "°";
        }

        private void PanToSetPoint()
        {
            int setpoint = (int)Math.Floor(_tecs[_tec].SetPoint);
            SetVertGraphRange(setpoint - 10, setpoint + 6);
        }

        /// <summary>
        /// Handles the Scroll event of the vScrollBar1 control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.Windows.Forms.ScrollEventArgs"/> instance containing the event data.</param>
        private void vScrollBar1_Scroll(object sender, ScrollEventArgs e)
        {
            int tmin = vScrollBar1.Value;
            int tmax = tmin + (int)(usGraph1.YTick * 8);

            SetVertGraphRange(tmin, tmax);
        }

        /// <summary>
        /// Handles the Click event of the buttonSaveImage control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void buttonSaveImage_Click(object sender, EventArgs e)
        {
            saveFileDialog1.ShowDialog();
            if (saveFileDialog1.FileName != "")
            {
                try
                {
                    usGraph1.SaveImage(saveFileDialog1.FileName);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message, "Error saving image to file.");
                }
            }
        }

        private void PIDMonitor_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (e.CloseReason == CloseReason.UserClosing)
                e.Cancel = true;

            Hide();
        }

        protected void UpdateGui(object sender, LaserUpdatedEvent e)
        {
            if (this.InvokeRequired == true)
            {
                UpdateSerialData updateHandler = new UpdateSerialData(UpdateGui);
                this.BeginInvoke(updateHandler, new Object[] { sender, e });
                return;
            }

            Tec[] tecs = laser.GetTecs();
            switch (e.CommandResponse.CommandId)
            {
                case CommandId.GetPidForTec:
                    PidData(0, tecs[0]);
                    break;

                case CommandId.ReadTecTemperature:
                    Plot(0, tecs[0]);
                    break;

                case CommandId.ReadAll:
                    // TEC PID settings.
                    for (int i = 0; i < laser.NumberOfTecs; i++)
                    {
                        Plot(i, tecs[i]);
                    }
                    break;
                case CommandId.ReadAllSetPoints:
                    // TEC PID settings.
                    PidData(_tec, tecs[_tec]);
                    break;
            }

        }
        #endregion

    }
}