﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using System.Globalization;

namespace CoboltMonitor.Laser
{
    /// <summary>
    /// Class handling generation06_81 lasers
    /// </summary>
    public class Gen06_81Laser : LaserCommon
    {
        public Gen06_81Laser(SerialHandler sh)
            : base(sh, 3)
        {
            powerRangeMax = 500.0f;
            currentRangeMin = 500.0f;
            currentRangeMax = 5000.0f;
        }

        enum ReadAllOffset
        {
            BOARD_TEMP      = 0,
            TEC_TEMP        = 1,
            TEC_DRIVE       = 4,
            LASER_PDIODE_V  = 7,
            LASER_CURRENT   = 8,
            LASER_POWER     = 9,
            LED_BITS        = 10,
            LASER_OP_MODE   = 11,
            AUTOSTART_MODE  = 12,
            TEC_FAULT       = 13,
            LASER_FAULT     = 16,
            TEC_RUNNING     = 17
        };

        enum ReadAllSetPointsOffset
        {
            TEC_TEMP_SETPOINT       = 0,
            FLAGS                   = 3,
            TEC_PID                 = 4,
            LASER_CURRENT_SETPOINT  = 13,
            LASER_POWER_SETPOINT    = 14,
            TEC_DELTA               = 15,
            LASER_WARMUP_DELAY      = 16,
            LASER_WARMUP_CURR       = 17
        };

        /// <summary>
        /// Parses the laser read all response.
        /// </summary>
        /// <param name="resp">String with readall data</param>
        protected override void ParseReadAllResponse(string resp)
        {
            String[] readings = resp.Split(new char[] { ' ' });

            if (readings.Length == 20)
            {
                ParseFloatResponse(readings[(int)ReadAllOffset.BOARD_TEMP], out boardTemp, 0);

                // TEC temperature readings.
                for (int i = 0; i < tecs.Length; i++)
                {
                    ParseReadTempResponse(readings[(int)ReadAllOffset.TEC_TEMP + i], i);
                }

                // TEC drive readings.
                for (int i = 0; i < tecs.Length; i++)
                {
                    ParseReadDriveResponse(readings[(int)ReadAllOffset.TEC_DRIVE + i], i);
                }

                // Laser Current set point.
                ParseReadLaserCurrentResponse(readings[(int)ReadAllOffset.LASER_CURRENT]);

                // Laser Power set point.
                ParseReadLaserPowerResponse(readings[(int)ReadAllOffset.LASER_POWER]);

                ParseReadPhotoDiodeVoltageResponse(readings[(int)ReadAllOffset.LASER_PDIODE_V]);

                // Laser Operating mode.
                int om;
                if (int.TryParse(readings[(int)ReadAllOffset.LASER_OP_MODE], out om))
                {
                    OperatingMode = om;
                }

                // Constant Current or Constant Power.
                int auto;
                int.TryParse(readings[(int)ReadAllOffset.AUTOSTART_MODE], out auto);
                if (auto == 0)
                {
                    ConstantCurrent = true;
                    ConstantPower = false;
                }
                else if (auto == 1)
                {
                    ConstantCurrent = false;
                    ConstantPower = true;
                }

                //TEC Faults
                int fault = 0;
                for (int i = 0; i < tecs.Length; ++i)
                {
                    lock (valueLock)
                    {
                        if (int.TryParse(readings[(int)ReadAllOffset.TEC_FAULT + i], out fault))
                            tecs[i].Fault = fault;
                    }
                }

                if (int.TryParse(readings[(int)ReadAllOffset.LASER_FAULT], out fault))
                    LaserFault = fault;

                // TEC running.
                for (int i = 0; i < tecs.Length; i++)
                {
                    ParseReadTecRunningResponse(readings[(int)ReadAllOffset.TEC_RUNNING + i], i);
                }

                int leds;
                if (int.TryParse(readings[(int)ReadAllOffset.LED_BITS], out leds))
                    LedBits = leds;
            }
        }

        /// <summary>
        /// Parses the laser read all setpoints response.
        /// </summary>
        /// <param name="resp">String with readallsetpoints data</param>
        protected override void ParseReadAllSetPointsResponse(string resp)
        {
            String[] readings = resp.Split(new char[] { ' ' });
            if (readings.Length == 18)
            {
                // TEC temperature set points.
                for (int i = 0; i < tecs.Length; i++)
                {
                    ParseGetTempResponse(readings[(int)ReadAllSetPointsOffset.TEC_TEMP_SETPOINT + i], i);
                }
                // TEC Enable states.
                int flags;
                int.TryParse(readings[(int)ReadAllSetPointsOffset.FLAGS], out flags);

                lock (tecs.SyncRoot)
                {
                    tecs[0].Enabled = (flags & 1 << 0) != 0;
                    tecs[1].Enabled = (flags & 1 << 1) != 0;
                    tecs[2].Enabled = (flags & 1 << 2) != 0;
                }
                // Autostart Enable state.
                AutoStart = (flags & 1 << 3) != 0;
                Cdrh = ((flags & 1 << 4) != 0);
                DirectInput = ((flags & 1 << 5) != 0);

                // TEC PID settings.
                for (int i = 0; i < tecs.Length; i++)
                {
                    ParseGetPidResponse(readings, (int)ReadAllSetPointsOffset.TEC_PID + i * 3, i);
                }

                // Laser Current set point.
                ParseGetLaserCurrentResponse(readings[(int)ReadAllSetPointsOffset.LASER_CURRENT_SETPOINT]);

                // Laser Power set point.
                ParseGetLaserPowerResponse(readings[(int)ReadAllSetPointsOffset.LASER_POWER_SETPOINT]);

                // TEC delta setting.
                ParseGetDeltaTempResponse(readings[(int)ReadAllSetPointsOffset.TEC_DELTA]);

                // Laser Warm Up Delay setting.
                ParseGetWarmUpTimeResponse(readings[(int)ReadAllSetPointsOffset.LASER_WARMUP_DELAY]);

                // Laser autostart warm up current set point.
                ParseGetWarmUpCurrentResponse(readings[(int)ReadAllSetPointsOffset.LASER_WARMUP_CURR]);
            }
        }

        /// <summary>
        /// Handles the Tick event of the timerComm control.
        /// Handles Gen0681 specific functionality.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        protected override void timerComm_Tick(object sender, EventArgs e)
        {
            if (timerComm.Enabled == false)
                return;
            if (GetValues == true && serialHandler.Resends == 0)
            {
                SendCommand(CommandId.GetTec1MaxDrive, "");
                SendCommand(CommandId.GetTec2MaxDrive, "");
                SendCommand(CommandId.GetTec3MaxDrive, "");
                SendCommand(CommandId.GetCalibrateLdCurr, "");
            }
            base.timerComm_Tick(sender, e);
        }

        /// <summary>
        /// Adds gen0681 specific commands to command dictionary.
        /// </summary>
        protected override void BuildDictionary()
		{
            base.BuildDictionary();

            cmdDict.Add(CommandId.GetPidForTec1, new CommandInfo("gpidtec1?", "Get PID parameters for TEC1.", true, true));
            cmdDict.Add(CommandId.GetPidForTec2, new CommandInfo("gpidtec2?", "Get PID parameters for TEC2.", true, true));
            cmdDict.Add(CommandId.GetPidForTec3, new CommandInfo("gpidtec3?", "Get PID parameters for TEC3.", true, true));
            cmdDict.Add(CommandId.GetPidForTecld, new CommandInfo("gpidtecld?", "Get PID parameters for TECLD.", true, true));

            cmdDict.Add(CommandId.SetPidForTec1, new CommandInfo("spidtec1", "Set PID parameters for TEC1.", true, true));
            cmdDict.Add(CommandId.SetPidForTec2, new CommandInfo("spidtec2", "Set PID parameters for TEC2.", true, true));
            cmdDict.Add(CommandId.SetPidForTec3, new CommandInfo("spidtec3", "Set PID parameters for TEC3.", true, true));
            cmdDict.Add(CommandId.SetPidForTecld, new CommandInfo("spidtecld", "Set PID parameters for TECLD.", true, true));

            cmdDict.Add(CommandId.SetTec1Temperature, new CommandInfo("stec1t", "Set TEC1 temperature.", true, true));
            cmdDict.Add(CommandId.SetTec2Temperature, new CommandInfo("stec2t", "Set TEC2 temperature.", true, true));
            cmdDict.Add(CommandId.SetTec3Temperature, new CommandInfo("stec3t", "Set TEC3 temperature.", true, true));
            cmdDict.Add(CommandId.SetTecldTemperature, new CommandInfo("stecldt", "Set TECLD temperature.", true, true));

            cmdDict.Add(CommandId.GetTec1Temperature, new CommandInfo("gtec1t?", "Get TEC1 temperature.", true, true));
            cmdDict.Add(CommandId.GetTec2Temperature, new CommandInfo("gtec2t?", "Get TEC2 temperature.", true, true));
            cmdDict.Add(CommandId.GetTec3Temperature, new CommandInfo("gtec3t?", "Get TEC3 temperature.", true, true));
            cmdDict.Add(CommandId.GetTecldTemperature, new CommandInfo("gtecldt?", "Get TECLD temperature.", true, true));

            cmdDict.Add(CommandId.ReadTec1Temperature, new CommandInfo("rtec1t?", "Read TEC1 temperature.", true, true));
            cmdDict.Add(CommandId.ReadTec2Temperature, new CommandInfo("rtec2t?", "Read TEC2 temperature.", true, true));
            cmdDict.Add(CommandId.ReadTec3Temperature, new CommandInfo("rtec3t?", "Read TEC3 temperature.", true, true));
            cmdDict.Add(CommandId.ReadTecldTemperature, new CommandInfo("rtecldt?", "Read TECLD temperature.", true, true));

            cmdDict.Add(CommandId.DisableTec1, new CommandInfo("dtec1", "Disable TEC1.", true, true));
            cmdDict.Add(CommandId.DisableTec2, new CommandInfo("dtec2", "Disable TEC2.", true, true));
            cmdDict.Add(CommandId.DisableTec3, new CommandInfo("dtec3", "Disable TEC3.", true, true));
            cmdDict.Add(CommandId.DisableTecld, new CommandInfo("dtecld", "Disable TECLD.", true, true));

            cmdDict.Add(CommandId.EnableTec1, new CommandInfo("etec1", "Enable TEC1.", true, true));
            cmdDict.Add(CommandId.EnableTec2, new CommandInfo("etec2", "Enable TEC2.", true, true));
            cmdDict.Add(CommandId.EnableTec3, new CommandInfo("etec3", "Enable TEC3.", true, true));
            cmdDict.Add(CommandId.EnableTecld, new CommandInfo("etecld", "Enable TECLD.", true, true));

            cmdDict.Add(CommandId.GetTec1EnabledState, new CommandInfo("gtec1es?", "Get TEC1 enable state.", true, true));
            cmdDict.Add(CommandId.GetTec2EnabledState, new CommandInfo("gtec2es?", "Get TEC2 enable state.", true, true));
            cmdDict.Add(CommandId.GetTec3EnabledState, new CommandInfo("gtec3es?", "Get TEC3 enable state.", true, true));
            cmdDict.Add(CommandId.GetTecldEnabledState, new CommandInfo("gtecldes?", "Get TECLD enable state.", true, true));

            cmdDict.Add(CommandId.RunTec1, new CommandInfo("rtec1", "Run TEC1.", true, true));
            cmdDict.Add(CommandId.RunTec2, new CommandInfo("rtec2", "Run TEC2.", true, true));
            cmdDict.Add(CommandId.RunTec3, new CommandInfo("rtec3", "Run TEC3.", true, true));
            cmdDict.Add(CommandId.RunTecld, new CommandInfo("rtecld", "Run TECLD.", true, true));

            cmdDict.Add(CommandId.StopTec1, new CommandInfo("stec1", "Stop TEC1.", true, true));
            cmdDict.Add(CommandId.StopTec2, new CommandInfo("stec2", "Stop TEC2.", true, true));
            cmdDict.Add(CommandId.StopTec3, new CommandInfo("stec3", "Stop TEC3.", true, true));
            cmdDict.Add(CommandId.StopTecld, new CommandInfo("stecld", "Stop TECLD.", true, true));

            cmdDict.Add(CommandId.GetTec1RunState, new CommandInfo("gtec1rs?", "Get TEC1 run state.", true, true));
            cmdDict.Add(CommandId.GetTec2RunState, new CommandInfo("gtec2rs?", "Get TEC2 run state.", true, true));
            cmdDict.Add(CommandId.GetTec3RunState, new CommandInfo("gtec3rs?", "Get TEC3 run state.", true, true));
            cmdDict.Add(CommandId.GetTecldRunState, new CommandInfo("gtecldrs?", "Get TECLD run state.", true, true));

            cmdDict.Add(CommandId.ReadTec1Drive, new CommandInfo("rtec1drv?", "Read TEC1 drive %.", true, true));
            cmdDict.Add(CommandId.ReadTec2Drive, new CommandInfo("rtec2drv?", "Read TEC2 drive %.", true, true));
            cmdDict.Add(CommandId.ReadTec3Drive, new CommandInfo("rtec3drv?", "Read TEC3 drive %.", true, true));
            cmdDict.Add(CommandId.ReadTecldDrive, new CommandInfo("rteclddrv?", "Read TECLD drive %.", true, true));

            cmdDict.Add(CommandId.SetTec1MaxDrive, new CommandInfo("st1md", "Set TEC1 max drive (0-100%).", true, true));
            cmdDict.Add(CommandId.SetTec2MaxDrive, new CommandInfo("st2md", "Set TEC2 max drive (0-100%).", true, true));
            cmdDict.Add(CommandId.SetTec3MaxDrive, new CommandInfo("st3md", "Set TEC3 max drive (0-100%).", true, true));
            cmdDict.Add(CommandId.SetTecldMaxDrive, new CommandInfo("stldmd", "Set TECLD max drive (0-100%).", true, true));

            cmdDict.Add(CommandId.GetTec1MaxDrive, new CommandInfo("gt1md?", "Get TEC1 max drive.", true, true));
            cmdDict.Add(CommandId.GetTec2MaxDrive, new CommandInfo("gt2md?", "Get TEC2 max drive.", true, true));
            cmdDict.Add(CommandId.GetTec3MaxDrive, new CommandInfo("gt3md?", "Get TEC3 max drive.", true, true));
            cmdDict.Add(CommandId.GetTecldMaxDrive, new CommandInfo("gtldmd?", "Get TECLD max drive.", true, true));

            cmdDict.Add(CommandId.SetWarmUpDelay, new CommandInfo("swud", "Set warm up delay.", true, true));
            cmdDict.Add(CommandId.GetWarmUpDelay, new CommandInfo("gwud?", "Get warm up delay.", true, true));
            cmdDict.Add(CommandId.SetAutostartWarmUpCurrent, new CommandInfo("sawuc", "Set autostart warm up current.", true, true));
            cmdDict.Add(CommandId.GetAutostartWarmUpCurrent, new CommandInfo("gawuc?", "Get autostart warm up current.", true, true));

            //MLD specific
            cmdDict.Add(CommandId.Set10VCtr, new CommandInfo("s10vctr", "Set 10V control value.", true, true));
            cmdDict.Add(CommandId.Get10VCtr, new CommandInfo("g10vctr?", "Get 10V control value.", true, true));
            cmdDict.Add(CommandId.Set10VEna, new CommandInfo("s10vena", "Set 10V enabled state.", true, true));
            cmdDict.Add(CommandId.Get10VEna, new CommandInfo("g10vena?", "Get 10V enabled state.", true, true));
 
            cmdDict.Add(CommandId.GetDac, new CommandInfo("gdac?", "Get laser drive (DAC) value.", true, true));

            cmdDict.Add(CommandId.SetPhotoDiodeDarkVal, new CommandInfo("spddv", "Set photo diode dark value.", true, true));
            cmdDict.Add(CommandId.GetPhotoDiodeDarkVal, new CommandInfo("gpddv?", "Get photo diode dark value.", true, true));

            cmdDict.Add(CommandId.EnterTestMode, new CommandInfo(":etm", "Enter test mode.", true, true));
            cmdDict.Add(CommandId.ExitTestMode, new CommandInfo(":xtm", "Exit test mode.", true, true));

            //0681 specific
            cmdDict.Add(CommandId.ReadTec1Ntc, new CommandInfo(":ntc1?", "Read Tec 1 Ntc.", true, true));
            cmdDict.Add(CommandId.ReadTec2Ntc, new CommandInfo(":ntc2?", "Read Tec 2 Ntc.", true, true));
            //cmdDict.Add(CommandId.ReadTec3Ntc, new CommandInfo(":ntc3?", "Read Tec 3 Ntc.", true, true));
            cmdDict.Add(CommandId.ReadTecldNtc, new CommandInfo(":ntcld?", "Read Tec Ld Ntc.", true, true));

            cmdDict.Add(CommandId.ReadTec1Curr, new CommandInfo(":itec1?", "Read Tec 1 current.", true, true));
            cmdDict.Add(CommandId.ReadTec2Curr, new CommandInfo(":itec2?", "Read Tec 2 current.", true, true));
            cmdDict.Add(CommandId.ReadTec3Curr, new CommandInfo(":itec3?", "Read Tec 3 current.", true, true));
            cmdDict.Add(CommandId.ReadTecldCurr, new CommandInfo(":itec´ld?", "Read Tec Ld current.", true, true));

            cmdDict.Add(CommandId.WriteTec1Ctrl, new CommandInfo(":tec1ctrl", "Set Tec 1 Control.", true, true));
            cmdDict.Add(CommandId.WriteTec2Ctrl, new CommandInfo(":tec2ctrl", "Set Tec 2 Control.", true, true));
            cmdDict.Add(CommandId.WriteTec3Ctrl, new CommandInfo(":tec3ctrl", "Set Tec 3 Control.", true, true));
            cmdDict.Add(CommandId.WriteTecldCtrl, new CommandInfo(":tecldctrl", "Set Tec Ld Control.", true, true));

            cmdDict.Add(CommandId.WriteTEC1ENA, new CommandInfo(":tec1ena", "Set Tec 1 enabled.", true, true));
            cmdDict.Add(CommandId.WriteTEC2ENA, new CommandInfo(":tec2ena", "Set Tec 2 enabled.", true, true));
            cmdDict.Add(CommandId.WriteTEC3ENA, new CommandInfo(":tec3ena", "Set TecC 3 enabled.", true, true));
            cmdDict.Add(CommandId.WriteTECldENA, new CommandInfo(":tecldena", "Set Tec Ld enabled.", true, true));

            cmdDict.Add(CommandId.Disable5VDirectInputStart, new CommandInfo("d5vdis", "Disable 5V direct input start.", true, true));
            cmdDict.Add(CommandId.Enable5VDirectInputStart, new CommandInfo("e5vdis", "Enable 5V direct input start.", true, true));

            cmdDict.Add(CommandId.EnableDisableDirectInput, new CommandInfo("@cobasdr", "Enable/disable direct control.", false, false));
            cmdDict.Add(CommandId.GetDirectInputEnableState, new CommandInfo("gdies?", "Get direct input enable state.", true, true));
            cmdDict.Add(CommandId.GetDirectInputEnableState2, new CommandInfo("@cobasdr?", "Get direct control enable state.", false, false));

            cmdDict.Add(CommandId.GetSerialNumber3, new CommandInfo("glsn?", "Get serial number", true, true));
            cmdDict.Add(CommandId.SetSerialNumber, new CommandInfo("slsn", "Set serial number", true, true));

            cmdDict.Add(CommandId.GetLaserModel, new CommandInfo("glm?", "Get laser model", true, true));
            cmdDict.Add(CommandId.SetLaserModel, new CommandInfo("slm", "Set laser model", true, true));

            cmdDict.Add(CommandId.GetDriverNumber, new CommandInfo("gdsn?", "Get driver number", true, true));
            cmdDict.Add(CommandId.SetDriverNumber, new CommandInfo("sdsn", "Set driver number", true, true));

            cmdDict.Add(CommandId.SetTec1Offset, new CommandInfo("o1", "Set Tec 1 offset.", true, true));
            cmdDict.Add(CommandId.SetTec2Offset, new CommandInfo("o2", "Set Tec 2 offset.", true, true));
            cmdDict.Add(CommandId.SetTec3Offset, new CommandInfo("o3", "Set Tec 3 offset.", true, true));
            cmdDict.Add(CommandId.SetTecLdOffset, new CommandInfo("old", "Set Tec ld offset.", true, true));

            cmdDict.Add(CommandId.ReadAllJSON, new CommandInfo("rajson?", "Read all JSON", true, true));
            cmdDict.Add(CommandId.ReadAllSetPointsJSON, new CommandInfo("raspjson?", "Read all setpoints JSON", true, true));

            cmdDict.Add(CommandId.SetBaud, new CommandInfo("sb", "Set baud rate", true, true));

            cmdDict.Add(CommandId.FirmwareUpdate, new CommandInfo("@FirmwareUpdate", "Update firmware", true, true));
            cmdDict.Add(CommandId.WriteFactoryDefaultLaserInfo, new CommandInfo("@wfdli", "Write factory default laser info", true, true));

            cmdDict.Add(CommandId.ReadBoardTemp, new CommandInfo("rbtmp?", "Read board temperature.", true, true));
            cmdDict.Add(CommandId.GetRemoteOnDetect, new CommandInfo("grod?", "Get remote on detect.", true, true));

            cmdDict.Add(CommandId.ReadDCSEN, new CommandInfo(":dcsen?", "Read laser diode current.", true, true));
            cmdDict.Add(CommandId.ReadPWRMON, new CommandInfo(":pwrmon?", "Read laser output power.", true, true));
            cmdDict.Add(CommandId.ReadREMON, new CommandInfo(":remon?", "Read REMON.", true, true));
            cmdDict.Add(CommandId.ReadBTMP, new CommandInfo(":btmp?", "Read BTMP.", true, true)); 
            cmdDict.Add(CommandId.ReadLDREM, new CommandInfo(":ldrem?", "Read LDREM.", true, true));

            cmdDict.Add(CommandId.Write10VENA, new CommandInfo(":10vena", "Write 10v ena.", true, true));
            cmdDict.Add(CommandId.WriteACLEV, new CommandInfo(":aclev", "Read aclev.", true, true));
            cmdDict.Add(CommandId.WriteCCLIM, new CommandInfo(":cclim", "Read cclim.", true, true));
            cmdDict.Add(CommandId.WriteBIAS, new CommandInfo(":bias", "Set BIAS.", true, true)); 
            cmdDict.Add(CommandId.Write10VCTR, new CommandInfo(":10vctr", "Read 10v ctr.", true, true));
        }

        /// <summary>
        /// Writes 0681 laser functionality to file.
        /// </summary>
        /// <param name="commandFilePath">Path to command file.</param>
        /// <param name="message">Eventual error message</param>
        /// <returns>true if successful else false</returns>
        public override bool SaveCommand(string commandFilePath, out string message)
        {
            try
            {
                using (StreamWriter writer = new StreamWriter(commandFilePath))
                {

                    // Some useful information.
                    Tec[] tecs = GetTecs();

                    writer.WriteLine("%----------------------------------------------------------------");
                    writer.WriteLine("%");
                    writer.WriteLine("%  Laser Driver Settings");
                    writer.WriteLine("%");
                    writer.WriteLine("%  Device Serial Number: {0}", SerialNumber);
                    writer.WriteLine("%");
                    writer.WriteLine("%  Date: {0}", DateTime.Now);
                    writer.WriteLine("%");
                    writer.WriteLine("%----------------------------------------------------------------");
                    writer.WriteLine();

                    // Set max laser diode current.
                    writer.WriteLine("% Set max laser diode current.");
                    writer.WriteLine("smlc {0}", MaxLaserCurrent);
                    writer.WriteLine();

                    // Laser PID settings.
                    writer.WriteLine("% Laser PID settings.");
                    writer.WriteLine("slpid {0} {1}", LdKp, LdKi);
                    writer.WriteLine();

                    // Maximum laser power (mW).
                    writer.WriteLine("% Maximum laser power (mW).");
                    writer.WriteLine("smlp {0}", MaxLaserPower);
                    writer.WriteLine();

                    // Laser Current set point.
                    writer.WriteLine("% Laser Current Set Point (A).");
                    writer.WriteLine("slc {0}", LaserCurrentSetPoint);
                    writer.WriteLine();

                    // Laser Power set point.
                    writer.WriteLine("% Laser Power Set Point (mW).");
                    writer.WriteLine("slp {0}", LaserPowerSetPoint);
                    writer.WriteLine();

                    // Laser autostart warm up current set point.
                    writer.WriteLine("% Laser Autostart Warm Up Current Set Point (A).");
                    writer.WriteLine("sawuc {0}", WarmUpCurrent);
                    writer.WriteLine();

                    // Laser Warm Up Delay setting.
                    writer.WriteLine("% Laser Warm Up Delay setting (s).");
                    writer.WriteLine("swud {0}", WarmUpTime);
                    writer.WriteLine();

                    // TEC delta setting.
                    writer.WriteLine("% TEC delta setting (C).");
                    writer.WriteLine("stdt {0}", TecsDeltaTemperature.ToString(new CultureInfo("en-US")));
                    writer.WriteLine();
                    // TEC PID settings.
                    writer.WriteLine("% TEC PID Settings.");
                    for (int i = 0; i < tecs.Length; i++)
                    {
                        writer.WriteLine("spidtec{0} {1} {2} {3}",
                            i + 1,
                            tecs[i].Kp.ToString(new CultureInfo("en-US")),
                            tecs[i].Ki.ToString(new CultureInfo("en-US")),
                            tecs[i].Kd.ToString(new CultureInfo("en-US")));
                    }
                    writer.WriteLine();

                    // TEC temperature set points.
                    writer.WriteLine("% TEC Temperature Set Points (C).");
                    for (int i = 0; i < tecs.Length; i++)
                    {
                        writer.WriteLine("stec" + (i + 1) + "t {0}", tecs[i].Temp);
                    }
                    writer.WriteLine();

                    // TEC max drive.
                    writer.WriteLine("% TEC Max Drive (%).");
                    for (int i = 0; i < tecs.Length; i++)
                    {
                        writer.WriteLine("st" + (i + 1) + "md {0}", tecs[i].MaxDrive);
                    }
                    writer.WriteLine();

                    // TEC Enable states.
                    writer.WriteLine("% TEC Enabled States.");
                    for (int i = 0; i < tecs.Length; i++)
                    {
                        writer.WriteLine("{0}tec" + (i + 1), tecs[i].Enabled ? "e" : "d");
                    }
                    writer.WriteLine();
                    // Autostart Configuration.
                    writer.WriteLine("% CDRH (Key Switch) Enabled State (dks = disabled, eks = enabled).");
                    writer.WriteLine(Cdrh ? "eks" : "dks");
                    writer.WriteLine();

                    writer.WriteLine("% Autostart Enabled State (da = disabled, ea = enabled).");
                    writer.WriteLine(AutoStart ? "ea" : "da");
                    writer.WriteLine();

                    writer.Close();
                }
            }
            catch (Exception ex)
            {
                message = ex.Message;
                return false;
            }
            message = String.Empty;
            return true;
        }
    }
}
