﻿using System;
using System.Collections.Generic;
using System.Text;

using System.Globalization;
using System.IO;

namespace CoboltMonitor.Laser
{
    
    /// <summary>
    /// Class handling generation4 lasers. Does not differ from lasercommon
    /// </summary>
    public class Gen4Laser:LaserCommon
    {
        #region Enumerations
        enum ReadAllOffset
        {
            TEC_TEMP = 0,
            TEC_DRIVE = 4,
            LASER_CURRENT = 8,
            LASER_POWER = 9,
            LASER_DIODE_V = 10,
            RMS_VOLTAGE = 11,
            RMS_PERCENTAGE = 12,
            LASER_PDIODE_V = 13,
            LED_BITS = 14,
            LASER_OP_MODE = 15,
            AUTOSTART_MODE = 16,
            TEC_LASER_FAULT = 17
        };

        enum ReadAllSetPointsOffset
        {
            TEC_TEMP_SETPOINT = 0,
            FLAGS = 4,
            TEC_PID = 5,
            LASER_CURRENT_SETPOINT = 17,
            LASER_POWER_SETPOINT = 18,
            LASER_WARMUP_CURR = 19,
            LASER_WARMUP_DELAY = 20,
            TEC_DELTA = 21
        };
        #endregion

        public Gen4Laser(SerialHandler sh)
            : base(sh,4)
        {
        }

        /// <summary>
        /// Parses the laser read all response.
        /// </summary>
        /// <param name="resp">String with readall data</param>
        protected override void ParseReadAllResponse(string resp)
        {
            String[] readings = resp.Split(new char[] { ' ' });

            if (readings.Length == 18)
            {
                // TEC temperature readings.
                for (int i = 0; i < tecs.Length; i++)
                {
                    ParseReadTempResponse(readings[(int)ReadAllOffset.TEC_TEMP + i], i);
                }

                // TEC drive readings.
                for (int i = 0; i < tecs.Length; i++)
                {
                    ParseReadDriveResponse(readings[(int)ReadAllOffset.TEC_DRIVE + i], i);
                }

                ParseReadLaserCurrentResponse(readings[(int)ReadAllOffset.LASER_CURRENT]);
                ParseReadLaserPowerResponse(readings[(int)ReadAllOffset.LASER_POWER]);
                ParseReadLaserDiodeVoltageResponse(readings[(int)ReadAllOffset.LASER_DIODE_V]);
                ParseReadRmsVoltageResponse(readings[(int)ReadAllOffset.RMS_VOLTAGE]);
                ParseReadRmsPercentResponse(readings[(int)ReadAllOffset.RMS_PERCENTAGE]);
                ParseReadPhotoDiodeVoltageResponse(readings[(int)ReadAllOffset.LASER_PDIODE_V]);

                int leds;
                if (int.TryParse(readings[(int)ReadAllOffset.LED_BITS], out leds))
                    LedBits = leds;

                // Laser Operating mode.
                int om;
                if (int.TryParse(readings[(int)ReadAllOffset.LASER_OP_MODE], out om))
                {
                    OperatingMode = om;
                }

                // Constant Current or Constant Power.
                if (readings[(int)ReadAllOffset.AUTOSTART_MODE].Equals("0"))
                {
                    ConstantCurrent = true;
                    ConstantPower = false;
                }
                else if (readings[(int)ReadAllOffset.AUTOSTART_MODE].Equals("1"))
                {
                    ConstantCurrent = false;
                    ConstantPower = true;
                }

                // TEC & Laser Faults.
                int bm;
                if (int.TryParse(readings[(int)ReadAllOffset.TEC_LASER_FAULT], out bm))
                {
                    int fault;

                    lock (tecs.SyncRoot)
                    {
                        for (int i = 0; i < tecs.Length; ++i)
                        {
                            fault = bm & 0x0f;
                            tecs[i].Fault = fault;
                            bm >>= 4;
                        }

                        // Laser Fault?
                        fault = bm & 0x0f;
                        LaserFault = fault;
                        bm >>= 4;
                        for (int i = 0; i < tecs.Length; ++i)
                            tecs[i].Running = (bm & 1 << i) != 0;
                    }
                }
            }
        }

        /// <summary>
        /// Parses the laser read all setpoints response.
        /// </summary>
        /// <param name="resp">String with readallsetpoints data</param>
        protected override void ParseReadAllSetPointsResponse(string resp)
        {
            String[] readings = resp.Split(new char[] { ' ' });
            if (readings.Length == 22)
            {
                // TEC temperature set points.
                for (int i = 0; i < tecs.Length; i++)
                {
                    ParseGetTempResponse(readings[(int)ReadAllSetPointsOffset.TEC_TEMP_SETPOINT + i], i);
                }
                // TEC Enable states.
                int flags;
                int.TryParse(readings[(int)ReadAllSetPointsOffset.FLAGS], out flags);

                lock (tecs.SyncRoot)
                {
                    tecs[0].Enabled = (flags & 1 << 0) != 0;
                    tecs[1].Enabled = (flags & 1 << 1) != 0;
                    tecs[2].Enabled = (flags & 1 << 2) != 0;
                    tecs[3].Enabled = (flags & 1 << 3) != 0;
                }
                // Autostart Enable state.
                AutoStart = (flags & 1 << 4) != 0;
                Cdrh = ((flags & 1 << 5) != 0);
                DirectInput = ((flags & 1 << 6) != 0);

                // TEC PID settings.
                for (int i = 0; i < tecs.Length; i++)
                {
                    ParseGetPidResponse(readings, (int)ReadAllSetPointsOffset.TEC_PID + i * 3, i);
                }

                // Laser Current set point.
                ParseGetLaserCurrentResponse(readings[(int)ReadAllSetPointsOffset.LASER_CURRENT_SETPOINT]);

                // Laser Power set point.
                ParseGetLaserPowerResponse(readings[(int)ReadAllSetPointsOffset.LASER_POWER_SETPOINT]);

                // Laser autostart warm up current set point.
                ParseGetWarmUpCurrentResponse(readings[(int)ReadAllSetPointsOffset.LASER_WARMUP_CURR]);

                // Laser Warm Up Delay setting.
                ParseGetWarmUpTimeResponse(readings[(int)ReadAllSetPointsOffset.LASER_WARMUP_DELAY]);

                // TEC delta setting.
                ParseGetDeltaTempResponse(readings[(int)ReadAllSetPointsOffset.TEC_DELTA]);
            }
        }
        
        /// <summary>
        /// Handles the Tick event of the timerComm control. Sends poll messages
        /// to the Cobolt Laser Driver.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        protected override void timerComm_Tick(object sender, EventArgs e)
        {
            if (timerComm.Enabled == false)
                return;

            if (GetValues == true && serialHandler.Resends == 0)
            {
                SendCommand(CommandId.GetTec1MaxDrive, "");
                SendCommand(CommandId.GetTec2MaxDrive, "");
                SendCommand(CommandId.GetTec3MaxDrive, "");
                SendCommand(CommandId.GetTec4MaxDrive, "");
                SendCommand(CommandId.GetCalibrateLdCurr, "");
            }
            SendCommand(CommandId.ReadNeVolt, "");
            base.timerComm_Tick(sender, e);
        }

        /// <summary>
        /// Adds gen4 specific commands to command dictionary.
        /// </summary>
        protected override void BuildDictionary()
        {
            base.BuildDictionary();
            cmdDict.Add(CommandId.GetPidForTec1, new CommandInfo("gpidtec1?", "Get PID parameters for TEC1.", true, true));
            cmdDict.Add(CommandId.GetPidForTec2, new CommandInfo("gpidtec2?", "Get PID parameters for TEC2.", true, true));
            cmdDict.Add(CommandId.GetPidForTec3, new CommandInfo("gpidtec3?", "Get PID parameters for TEC3.", true, true));
            cmdDict.Add(CommandId.GetPidForTec4, new CommandInfo("gpidtec4?", "Get PID parameters for TEC4.", true, true));

            cmdDict.Add(CommandId.SetPidForTec1, new CommandInfo("spidtec1", "Set PID parameters for TEC1.", true, true));
            cmdDict.Add(CommandId.SetPidForTec2, new CommandInfo("spidtec2", "Set PID parameters for TEC2.", true, true));
            cmdDict.Add(CommandId.SetPidForTec3, new CommandInfo("spidtec3", "Set PID parameters for TEC3.", true, true));
            cmdDict.Add(CommandId.SetPidForTec4, new CommandInfo("spidtec4", "Set PID parameters for TEC4.", true, true));

            cmdDict.Add(CommandId.SetTec1Temperature, new CommandInfo("stec1t", "Set TEC1 temperature.", true, true));
            cmdDict.Add(CommandId.SetTec2Temperature, new CommandInfo("stec2t", "Set TEC2 temperature.", true, true));
            cmdDict.Add(CommandId.SetTec3Temperature, new CommandInfo("stec3t", "Set TEC3 temperature.", true, true));
            cmdDict.Add(CommandId.SetTec4Temperature, new CommandInfo("stec4t", "Set TEC4 temperature.", true, true));

            cmdDict.Add(CommandId.GetTec1Temperature, new CommandInfo("gtec1t?", "Get TEC1 temperature.", true, true));
            cmdDict.Add(CommandId.GetTec2Temperature, new CommandInfo("gtec2t?", "Get TEC2 temperature.", true, true));
            cmdDict.Add(CommandId.GetTec3Temperature, new CommandInfo("gtec3t?", "Get TEC3 temperature.", true, true));
            cmdDict.Add(CommandId.GetTec4Temperature, new CommandInfo("gtec4t?", "Get TEC4 temperature.", true, true));

            cmdDict.Add(CommandId.ReadTec1Temperature, new CommandInfo("rtec1t?", "Read TEC1 temperature.", true, true));
            cmdDict.Add(CommandId.ReadTec2Temperature, new CommandInfo("rtec2t?", "Read TEC2 temperature.", true, true));
            cmdDict.Add(CommandId.ReadTec3Temperature, new CommandInfo("rtec3t?", "Read TEC3 temperature.", true, true));
            cmdDict.Add(CommandId.ReadTec4Temperature, new CommandInfo("rtec4t?", "Read TEC4 temperature.", true, true));

            cmdDict.Add(CommandId.DisableTec1, new CommandInfo("dtec1", "Disable TEC1.", true, true));
            cmdDict.Add(CommandId.DisableTec2, new CommandInfo("dtec2", "Disable TEC2.", true, true));
            cmdDict.Add(CommandId.DisableTec3, new CommandInfo("dtec3", "Disable TEC3.", true, true));
            cmdDict.Add(CommandId.DisableTec4, new CommandInfo("dtec4", "Disable TEC4.", true, true));

            cmdDict.Add(CommandId.EnableTec1, new CommandInfo("etec1", "Enable TEC1.", true, true));
            cmdDict.Add(CommandId.EnableTec2, new CommandInfo("etec2", "Enable TEC2.", true, true));
            cmdDict.Add(CommandId.EnableTec3, new CommandInfo("etec3", "Enable TEC3.", true, true));
            cmdDict.Add(CommandId.EnableTec4, new CommandInfo("etec4", "Enable TEC4.", true, true));

            cmdDict.Add(CommandId.GetTec1EnabledState, new CommandInfo("gtec1es?", "Get TEC1 enable state.", true, true));
            cmdDict.Add(CommandId.GetTec2EnabledState, new CommandInfo("gtec2es?", "Get TEC2 enable state.", true, true));
            cmdDict.Add(CommandId.GetTec3EnabledState, new CommandInfo("gtec3es?", "Get TEC3 enable state.", true, true));
            cmdDict.Add(CommandId.GetTec4EnabledState, new CommandInfo("gtec4es?", "Get TEC4 enable state.", true, true));

            cmdDict.Add(CommandId.RunTec1, new CommandInfo("rtec1", "Run TEC1.", true, true));
            cmdDict.Add(CommandId.RunTec2, new CommandInfo("rtec2", "Run TEC2.", true, true));
            cmdDict.Add(CommandId.RunTec3, new CommandInfo("rtec3", "Run TEC3.", true, true));
            cmdDict.Add(CommandId.RunTec4, new CommandInfo("rtec4", "Run TEC4.", true, true));

            cmdDict.Add(CommandId.StopTec1, new CommandInfo("stec1", "Stop TEC1.", true, true));
            cmdDict.Add(CommandId.StopTec2, new CommandInfo("stec2", "Stop TEC2.", true, true));
            cmdDict.Add(CommandId.StopTec3, new CommandInfo("stec3", "Stop TEC3.", true, true));
            cmdDict.Add(CommandId.StopTec4, new CommandInfo("stec4", "Stop TEC4.", true, true));

            cmdDict.Add(CommandId.RunStopTec2, new CommandInfo("xten", "Run/Stop TEC2.", false, true));
            cmdDict.Add(CommandId.RunStopTec3, new CommandInfo("t2en", "Run/Stop TEC3.", false, true));
            cmdDict.Add(CommandId.RunStopTec4, new CommandInfo("lten", "Run/Stop TEC4.", false, true));

            cmdDict.Add(CommandId.GetTec1RunState, new CommandInfo("gtec1rs?", "Get TEC1 run state.", true, true));
            cmdDict.Add(CommandId.GetTec2RunState, new CommandInfo("gtec2rs?", "Get TEC2 run state.", true, true));
            cmdDict.Add(CommandId.GetTec3RunState, new CommandInfo("gtec3rs?", "Get TEC3 run state.", true, true));
            cmdDict.Add(CommandId.GetTec4RunState, new CommandInfo("gtec4rs?", "Get TEC4 run state.", true, true));

            cmdDict.Add(CommandId.ReadTec1Drive, new CommandInfo("rtec1drv?", "Read TEC1 drive %.", true, true));
            cmdDict.Add(CommandId.ReadTec2Drive, new CommandInfo("rtec2drv?", "Read TEC2 drive %.", true, true));
            cmdDict.Add(CommandId.ReadTec3Drive, new CommandInfo("rtec3drv?", "Read TEC3 drive %.", true, true));
            cmdDict.Add(CommandId.ReadTec4Drive, new CommandInfo("rtec4drv?", "Read TEC4 drive %.", true, true));

			cmdDict.Add(CommandId.SetTec1MaxDrive, new CommandInfo("st1md", "Set TEC1 max drive (0-100%).", true, true));
			cmdDict.Add(CommandId.SetTec2MaxDrive, new CommandInfo("st2md", "Set TEC2 max drive (0-100%).", true, true));
			cmdDict.Add(CommandId.SetTec3MaxDrive, new CommandInfo("st3md", "Set TEC3 max drive (0-100%).", true, true));
            cmdDict.Add(CommandId.SetTec4MaxDrive, new CommandInfo("st4md", "Set TEC4 max drive (0-100%).", true, true));

			cmdDict.Add(CommandId.GetTec1MaxDrive, new CommandInfo("gt1md?", "Get TEC1 max drive.", true, true));
			cmdDict.Add(CommandId.GetTec2MaxDrive, new CommandInfo("gt2md?", "Get TEC2 max drive.", true, true));
			cmdDict.Add(CommandId.GetTec3MaxDrive, new CommandInfo("gt3md?", "Get TEC3 max drive.", true, true));
            cmdDict.Add(CommandId.GetTec4MaxDrive, new CommandInfo("gt4md?", "Get TEC4 max drive.", true, true));

            cmdDict.Add(CommandId.SetWarmUpDelay, new CommandInfo("swud", "Set warm up delay.", true, true));
            cmdDict.Add(CommandId.GetWarmUpDelay, new CommandInfo("gwud?", "Get warm up delay.", true, true));
            cmdDict.Add(CommandId.Disable5VDirectInputStart, new CommandInfo("d5vdis", "Disable 5V direct input start.", true, true));
            cmdDict.Add(CommandId.Enable5VDirectInputStart, new CommandInfo("e5vdis", "Enable 5V direct input start.", true, true));
            cmdDict.Add(CommandId.EnableDisableDirectInput, new CommandInfo("@cobasdr", "Enable/disable direct control.", false, false));
            cmdDict.Add(CommandId.GetDirectInputEnableState, new CommandInfo("gdies?", "Get direct input enable state.", true, true));
            cmdDict.Add(CommandId.GetDirectInputEnableState2, new CommandInfo("@cobasdr?", "Get direct input enable state.", false, false));
            cmdDict.Add(CommandId.SetAutostartWarmUpCurrent, new CommandInfo("sawuc", "Set autostart warm up current.", true, true));
            cmdDict.Add(CommandId.GetAutostartWarmUpCurrent, new CommandInfo("gawuc?", "Get autostart warm up current.", true, true));

            cmdDict.Add(CommandId.SetAutostartFinalCurrent, new CommandInfo("safc", "Set autostart final up current.", true, true));
            cmdDict.Add(CommandId.GetAutostartFinalCurrent, new CommandInfo("gafc?", "Get autostart final current.", true, true));
            cmdDict.Add(CommandId.ReadLaserDiodeVoltage, new CommandInfo("rldv?", "Read laser diode voltage.", true, true));
            cmdDict.Add(CommandId.ReadRmsVoltage, new CommandInfo("rrv?", "Read RMS voltage.", true, true));
            cmdDict.Add(CommandId.ReadRmsPercent, new CommandInfo("rrp?", "Read RMS percent.", true, true));
            cmdDict.Add(CommandId.ReadNeVolt, new CommandInfo("rnev?", "Read noise eater monitor voltage.", true, true));

            cmdDict.Add(CommandId.EnableS3, new CommandInfo("es3", "Enable S3 algorithm.", true, true));
            cmdDict.Add(CommandId.DisableS3, new CommandInfo("ds3", "Disable S3 algorithm.", true, true));
            cmdDict.Add(CommandId.GetS3, new CommandInfo("gs3?", "Get S3 enabled status.", true, true));
            cmdDict.Add(CommandId.TestCmd, new CommandInfo("test", "Test command.", true, true));
            cmdDict.Add(CommandId.GetIo, new CommandInfo("io?", "Get some info.", true, true));

            cmdDict.Add(CommandId.SysRes, new CommandInfo("SysRes", "Reset Target", true, true));
            cmdDict.Add(CommandId.EnableDebugMessages, new CommandInfo("edm", "Enable/disable debug messages", true, true));
        }

        /// <summary>
        /// Writes gen4 laser functionality to file.
        /// </summary>
        /// <param name="commandFilePath">Path to command file.</param>
        /// <param name="message">Eventual error message</param>
        /// <returns>true if successful else false</returns>
        public override bool SaveCommand(string commandFilePath, out string message)
        {
            try
            {
                using (StreamWriter writer = new StreamWriter(commandFilePath))
                {
                    // Some useful information.
                    Tec[] tecs = GetTecs();

                    writer.WriteLine("%----------------------------------------------------------------");
                    writer.WriteLine("%");
                    writer.WriteLine("%  Laser Driver Settings");
                    writer.WriteLine("%");
                    writer.WriteLine("%  Device Serial Number: {0}", SerialNumber);
                    writer.WriteLine("%");
                    writer.WriteLine("%  Date: {0}", DateTime.Now);
                    writer.WriteLine("%");
                    writer.WriteLine("%----------------------------------------------------------------");
                    writer.WriteLine();

                    // Set max laser diode current.
                    writer.WriteLine("% Set max laser diode current.");
                    writer.WriteLine("smlc {0}", MaxLaserCurrent);
                    writer.WriteLine();

                    // Laser PID settings.
                    writer.WriteLine("% Laser PID settings.");
                    writer.WriteLine("slpid {0} {1}", LdKp, LdKi);
                    writer.WriteLine();

                    // Maximum laser power (mW).
                    writer.WriteLine("% Maximum laser power (mW).");
                    writer.WriteLine("smlp {0}", MaxLaserPower);
                    writer.WriteLine();

                    // Laser Current set point.
                    writer.WriteLine("% Laser Current Set Point (A).");
                    writer.WriteLine("slc {0}", LaserCurrentSetPoint);
                    writer.WriteLine();

                    // Laser Power set point.
                    writer.WriteLine("% Laser Power Set Point (mW).");
                    writer.WriteLine("slp {0}", LaserPowerSetPoint);
                    writer.WriteLine();

                    // Laser autostart warm up current set point.
                    writer.WriteLine("% Laser Autostart Warm Up Current Set Point (A).");
                    writer.WriteLine("sawuc {0}", WarmUpCurrent);
                    writer.WriteLine();

                    // Laser Warm Up Delay setting.
                    writer.WriteLine("% Laser Warm Up Delay setting (s).");
                    writer.WriteLine("swud {0}", WarmUpTime);
                    writer.WriteLine();

                    // TEC delta setting.
                    writer.WriteLine("% TEC delta setting (C).");
                    writer.WriteLine("stdt {0}", TecsDeltaTemperature.ToString(new CultureInfo("en-US")));
                    writer.WriteLine();
                    // TEC PID settings.
                    writer.WriteLine("% TEC PID Settings.");
                    for (int i = 0; i < tecs.Length; i++)
                    {
                        writer.WriteLine("spidtec{0} {1} {2} {3}",
                            i + 1,
                            tecs[i].Kp.ToString(new CultureInfo("en-US")),
                            tecs[i].Ki.ToString(new CultureInfo("en-US")),
                            tecs[i].Kd.ToString(new CultureInfo("en-US")));
                    }
                    writer.WriteLine();

                    // TEC temperature set points.
                    writer.WriteLine("% TEC Temperature Set Points (C).");
                    for (int i = 0; i < tecs.Length; i++)
                    {
                        writer.WriteLine("stec" + (i + 1) + "t {0}", tecs[i].Temp);
                    }
                    writer.WriteLine();

                    // TEC max drive.
                    writer.WriteLine("% TEC Max Drive (%).");
                    for (int i = 0; i < tecs.Length; i++)
                    {
                        writer.WriteLine("st" + (i + 1) + "md {0}", tecs[i].MaxDrive);
                    }
                    writer.WriteLine();

                    // TEC Enable states.
                    writer.WriteLine("% TEC Enabled States.");
                    for (int i = 0; i < tecs.Length; i++)
                    {
                        writer.WriteLine("{0}tec" + (i + 1), tecs[i].Enabled ? "e" : "d");
                    }
                    writer.WriteLine();
                    // Autostart Configuration.
                    writer.WriteLine("% CDRH (Key Switch) Enabled State (dks = disabled, eks = enabled).");
                    writer.WriteLine(Cdrh ? "eks" : "dks");
                    writer.WriteLine();

                    writer.WriteLine("% Direct Input Enabled State (d5vdis = disabled, e5vdis = enabled).");
                    writer.WriteLine(DirectInput ? "e5vdis" : "d5vdis");
                    writer.WriteLine();

                    writer.WriteLine("% Autostart Enabled State (da = disabled, ea = enabled).");
                    writer.WriteLine(AutoStart ? "ea" : "da");
                    writer.WriteLine();

                    writer.Close();
                }
            }
            catch (Exception ex)
            {
                message = ex.Message;
                return false;
            }
            message = String.Empty;
            return true;
        }
    }
}
