﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using System.Globalization;

namespace CoboltMonitor.Laser
{
    /// <summary>
    /// Class handling generation06_0x lasers
    /// </summary>
    public class Gen6Laser : LaserCommon
    {
        #region Enumerations
        enum ReadAllOffset
        {
            BASEPLATE_TEMP = 0,
            BOARD_TEMP = 1,
            TEC_TEMP = 2,
            TEC_DRIVE = 3,
            LASER_CURRENT = 4,
            LASER_POWER = 5,
            LASER_OP_MODE = 6,
            AUTOSTART_MODE = 7,
            TEC_FAULT = 8,
            LASER_FAULT = 9,
            TEC_RUNNING = 10,
            MOD_SETTINGS = 11
        };

        enum ReadAllSetPointsOffset
        {
            TEC_TEMP_SETPOINT = 0,
            FLAGS = 1,
            TEC_PID = 2,
            LASER_CURRENT_SETPOINT = 5,
            LASER_POWER_SETPOINT = 6,
            TEC_DELTA = 7
        };
        #endregion
        
        public Gen6Laser(SerialHandler sh)
            : base(sh, 1)
        {
            powerRangeMax = 500.0f;
            currentRangeMin = 0.20f;
            currentRangeMax = 5.0f;
        }

        /// <summary>
        /// Parses the laser read all response.
        /// </summary>
        /// <param name="resp">String with readall data</param>
        protected override void ParseReadAllResponse(string resp)
        {
            String[] readings = resp.Split(new char[] { ' ' });

            if (readings.Length == 12)
            {
                ParseFloatResponse(readings[(int)ReadAllOffset.BASEPLATE_TEMP], out basePlateTemp, 0);
                ParseFloatResponse(readings[(int)ReadAllOffset.BOARD_TEMP], out boardTemp, 0);

                // TEC temperature readings.
                ParseReadTempResponse(readings[(int)ReadAllOffset.TEC_TEMP], 0);

                // TEC drive readings.
                ParseReadDriveResponse(readings[(int)ReadAllOffset.TEC_DRIVE], 0);

                // Laser Current set point.
                ParseReadLaserCurrentResponse(readings[(int)ReadAllOffset.LASER_CURRENT]);

                // Laser Power set point.
                ParseReadLaserPowerResponse(readings[(int)ReadAllOffset.LASER_POWER]);

                // Laser Operating mode.
                int om;
                if (int.TryParse(readings[(int)ReadAllOffset.LASER_OP_MODE], out om))
                {
                    OperatingMode = om;
                }

                // Constant Current or Constant Power.
                int auto;
                int.TryParse(readings[(int)ReadAllOffset.AUTOSTART_MODE], out auto);
                if (auto == 0)
                {
                    ConstantCurrent = true;
                    ConstantPower = false;
                    Modulation = false;
                }
                else if (auto == 1)
                {
                    ConstantCurrent = false;
                    ConstantPower = true;
                    Modulation = false;
                }
                else if (auto == 2)
                {
                    ConstantCurrent = false;
                    ConstantPower = false;
                    Modulation = true;
                }

                //TEC Faults
                int fault = 0;
                lock (valueLock)
                {
                    if (int.TryParse(readings[(int)ReadAllOffset.TEC_FAULT], out fault))
                        tecs[0].Fault = fault;
                }

                if (int.TryParse(readings[(int)ReadAllOffset.LASER_FAULT], out fault))
                    LaserFault = fault;

                // TEC running.
                ParseReadTecRunningResponse(readings[(int)ReadAllOffset.TEC_RUNNING], 0);

                // Modulation settings
                int modType = 0;
                if (int.TryParse(readings[(int)ReadAllOffset.MOD_SETTINGS], out modType))
                    ModulationSettings = modType;
            }
        }

        /// <summary>
        /// Parses the laser read all setpoints response.
        /// </summary>
        /// <param name="resp">String with readallsetpoints data</param>
        protected override void ParseReadAllSetPointsResponse(string resp)
        {
            String[] readings = resp.Split(new char[] { ' ' });
            if (readings.Length == 8)
            {
                // TEC temperature set points.
                ParseGetTempResponse(readings[(int)ReadAllSetPointsOffset.TEC_TEMP_SETPOINT], 0);
                // TEC Enable states.
                int flags;
                int.TryParse(readings[(int)ReadAllSetPointsOffset.FLAGS], out flags);

                lock (tecs.SyncRoot)
                {
                    tecs[0].Enabled = (flags & 1 << 0) != 0;
                }
                // Autostart Enable state.
                AutoStart = (flags & 1 << 1) != 0;
                Cdrh = ((flags & 1 << 2) != 0);

                // TEC PID settings.
                ParseGetPidResponse(readings, (int)ReadAllSetPointsOffset.TEC_PID, 0);

                // Laser Current set point.
                ParseGetLaserCurrentResponse(readings[(int)ReadAllSetPointsOffset.LASER_CURRENT_SETPOINT]);

                // Laser Power set point.
                ParseGetLaserPowerResponse(readings[(int)ReadAllSetPointsOffset.LASER_POWER_SETPOINT]);

                // TEC delta setting.
                ParseGetDeltaTempResponse(readings[(int)ReadAllSetPointsOffset.TEC_DELTA]);
            }
        }

        /// <summary>
        /// Handles the Tick event of the timerComm control.
        /// Handles Gen0681 specific functionality.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        protected override void timerComm_Tick(object sender, EventArgs e)
        {
            if (timerComm.Enabled == false)
                return;
            if (GetValues == true && serialHandler.Resends == 0)
            {
                SendCommand(CommandId.GetTecMaxDrive, "");
                SendCommand(CommandId.GetAnalogLowImpedanceState, "");
                SendCommand(CommandId.GetLaserModPower, "");
            }
            base.timerComm_Tick(sender, e); 
        }

        /// <summary>
        /// Adds gen0681 specific commands to command dictionary.
        /// </summary>
        protected override void BuildDictionary()
		{
            base.BuildDictionary();

            //MLD specific

            cmdDict.Add(CommandId.GetPidForTec, new CommandInfo("gpidtec?", "Get PID parameters for TEC.", true, true));
            cmdDict.Add(CommandId.SetPidForTec, new CommandInfo("spidtec", "Set PID parameters for TEC.", true, true));
            cmdDict.Add(CommandId.SetTecTemperature, new CommandInfo("stect", "Set TEC temperature.", true, true));
            cmdDict.Add(CommandId.GetTecTemperature, new CommandInfo("gtect?", "Get TEC temperature.", true, true));
            cmdDict.Add(CommandId.ReadTecTemperature, new CommandInfo("rtect?", "Read TEC temperature.", true, true));
            cmdDict.Add(CommandId.DisableTec, new CommandInfo("dtec", "Disable TEC.", true, true));
            cmdDict.Add(CommandId.EnableTec, new CommandInfo("etec", "Enable TEC.", true, true));
            cmdDict.Add(CommandId.GetTecEnabledState, new CommandInfo("gteces?", "Get TEC enable state.", true, true));
            cmdDict.Add(CommandId.RunTec, new CommandInfo("rtec", "Run TEC.", true, true));
            cmdDict.Add(CommandId.StopTec, new CommandInfo("stec", "Stop TEC.", true, true));
            cmdDict.Add(CommandId.GetTecRunState, new CommandInfo("gtecrs?", "Get TEC run state.", true, true));
            cmdDict.Add(CommandId.ReadTecDrive, new CommandInfo("rtecdrv?", "Read TEC drive %.", true, true));
            cmdDict.Add(CommandId.GetTecMaxDrive, new CommandInfo("gtmd?", "Get TEC max drive.", true, true));

            cmdDict.Add(CommandId.ReadIpLogCurr1, new CommandInfo("ripl1?", "Read I/P Log current for channel 1.", true, true));
            cmdDict.Add(CommandId.ReadIpLogCurr2, new CommandInfo("ripl2?", "Read I/P Log current for channel 2.", true, true));

            cmdDict.Add(CommandId.TestCmd, new CommandInfo("test", "Test command.", true, true));
            cmdDict.Add(CommandId.GetIo, new CommandInfo("io?", "Get some info.", true, true));
            
            cmdDict.Add(CommandId.EnterModMode, new CommandInfo("em", "Enter modulation mode.", false, false));
            cmdDict.Add(CommandId.EnterOnOffMode, new CommandInfo("eoom", "Enter on/off modulation mode.", false, false));
            cmdDict.Add(CommandId.ExitOnOffMode, new CommandInfo("xoom", "Exit on/off modulation mode.", false, false));

            cmdDict.Add(CommandId.SetLaserModPower, new CommandInfo("slmp", "Set laser modulated power (mW).", true, true));
            cmdDict.Add(CommandId.GetLaserModPower, new CommandInfo("glmp?", "Get laser modulated power (mW).", true, true));
            
            cmdDict.Add(CommandId.ReadBoardTemp, new CommandInfo("rbtmp?", "Read board temperature.", true, true));
            cmdDict.Add(CommandId.ReadBasePlateTemp, new CommandInfo("rbpt?", "Read baseplate temperature.", false, false));
            cmdDict.Add(CommandId.SetModulationEnabledState, new CommandInfo("smes", "Set modulation enabled state.", true, true));
            cmdDict.Add(CommandId.GetModulationEnabledState, new CommandInfo("gmes?", "Get modulation enabled state.", true, true));
            cmdDict.Add(CommandId.SetDigitalModulationEnabledState, new CommandInfo("sdmes", "Set digital modulation enabled state.", false, false));
            cmdDict.Add(CommandId.GetDigitalModulationEnabledState, new CommandInfo("gdmes?", "Get digital modulation enabled state.", false, false));
            cmdDict.Add(CommandId.SetAnalogModulationEnabledState, new CommandInfo("sames", "Set analog modulation enabled state.", false, false));
            cmdDict.Add(CommandId.GetAnalogModulationEnabledState, new CommandInfo("games?", "Get analog modulation enabled state.", false, false));
            cmdDict.Add(CommandId.SetAnalogLowImpedanceState, new CommandInfo("salis", "Set analog low imedance state.", false, false));
            cmdDict.Add(CommandId.GetAnalogLowImpedanceState, new CommandInfo("galis?", "Get analog low impedance state.", false, false));
            cmdDict.Add(CommandId.GetRemoteOnDetect, new CommandInfo("grod?", "Get remote on detect.", false, false));

            cmdDict.Add(CommandId.ReadSpare, new CommandInfo("@rsp?", "Read spare.", true, true));

            cmdDict.Add(CommandId.Set10VCtr, new CommandInfo("s10vctr", "Set 10V control value.", true, true));
            cmdDict.Add(CommandId.Get10VCtr, new CommandInfo("g10vctr?", "Get 10V control value.", true, true));
            cmdDict.Add(CommandId.Set10VEna, new CommandInfo("s10vena", "Set 10V enabled state.", true, true));
            cmdDict.Add(CommandId.Get10VEna, new CommandInfo("g10vena?", "Get 10V enabled state.", true, true));
            cmdDict.Add(CommandId.Set2ChDrv, new CommandInfo("s2chdrv", "Set two channel drive.", true, true));
            cmdDict.Add(CommandId.Get2ChDrv, new CommandInfo("g2chdrv?", "Get two channel drive setting.", true, true));
            cmdDict.Add(CommandId.GetCcLim1, new CommandInfo("gcclim1?", "Get CCLIM 1.", true, true));
            cmdDict.Add(CommandId.GetCcLim2, new CommandInfo("gcclim2?", "Get CCLIM 2.", true, true));

            cmdDict.Add(CommandId.SetAcLevAmod1, new CommandInfo("sala1", "Set ACLEV for analog modulation on channel (one channel drive).", true, true));
            cmdDict.Add(CommandId.GetAcLevAmod1, new CommandInfo("gala1?", "Get ACLEV for analog modulation on channel (one channel drive).", true, true));
            cmdDict.Add(CommandId.SetAcLevAmod2, new CommandInfo("sala2", "Set ACLEV for analog modulation on channel (two channel drive).", true, true));
            cmdDict.Add(CommandId.GetAcLevAmod2, new CommandInfo("gala2?", "Get ACLEV for analog modulation on channel (two channel drive).", true, true));
            cmdDict.Add(CommandId.GetAcLevDmod, new CommandInfo("gald?", "Get ACLEV for digital modulation.", true, true));

            cmdDict.Add(CommandId.GetDac, new CommandInfo("gdac?", "Get laser drive (DAC) value.", true, true));

            cmdDict.Add(CommandId.SetPhotoDiodeDarkVal, new CommandInfo("spddv", "Set photo diode dark value.", true, true));
            cmdDict.Add(CommandId.GetPhotoDiodeDarkVal, new CommandInfo("gpddv?", "Get photo diode dark value.", true, true));

            cmdDict.Add(CommandId.EnterTestMode, new CommandInfo(":etm", "Enter test mode.", true, true));
            cmdDict.Add(CommandId.ExitTestMode, new CommandInfo(":xtm", "Exit test mode.", true, true));

            cmdDict.Add(CommandId.ReadDCSEN, new CommandInfo(":dcsen?", "Read DCSEN.", true, true));
            cmdDict.Add(CommandId.ReadPWRMON, new CommandInfo(":pwrmon?", "Read PWRMON.", true, true));
            cmdDict.Add(CommandId.ReadHFDET, new CommandInfo(":hfdet?", "Read HFDET.", true, true));
            cmdDict.Add(CommandId.ReadREMON, new CommandInfo(":remon?", "Read REMON.", true, true));
            cmdDict.Add(CommandId.ReadBTMP, new CommandInfo(":btmp?", "Read BTMP.", true, true));
            cmdDict.Add(CommandId.Read5VIN, new CommandInfo(":5vin?", "Read 5V input.", true, true));
            cmdDict.Add(CommandId.ReadKEYINT, new CommandInfo(":keyint?", "Read Key/Interlock input.", true, true));
            cmdDict.Add(CommandId.ReadNTCTE, new CommandInfo(":ntcte?", "Read NTCTE.", true, true));
            cmdDict.Add(CommandId.ReadNTCSP, new CommandInfo(":ntcsp?", "Read NTCSP.", true, true));
            cmdDict.Add(CommandId.ReadNTCBP, new CommandInfo(":ntcbp?", "Read NTCBP.", true, true));
            cmdDict.Add(CommandId.ReadLDREM, new CommandInfo(":ldrem?", "Read LDREM.", true, true));
            cmdDict.Add(CommandId.ReadTECFAULT, new CommandInfo(":tecfault?", "Read TECFAULT.", true, true));

            cmdDict.Add(CommandId.Write10VENA, new CommandInfo(":10vena", "Set 10VENA.", true, true));
            cmdDict.Add(CommandId.WriteTECENA, new CommandInfo(":tecena", "Set TECENA.", true, true));
            cmdDict.Add(CommandId.WriteAMODDIS, new CommandInfo(":amoddis", "Set AMODDIS.", true, true));
            cmdDict.Add(CommandId.WriteAMOD50R, new CommandInfo(":amod50r", "Set AMOD50R.", true, true));
            cmdDict.Add(CommandId.WriteACLEV, new CommandInfo(":aclev", "Set ACLEV.", true, true));
            cmdDict.Add(CommandId.WriteCCLIM, new CommandInfo(":cclim", "Set CCLIM.", true, true));
            cmdDict.Add(CommandId.WriteBIAS, new CommandInfo(":bias", "Set BIAS.", true, true));
            cmdDict.Add(CommandId.Write10VCTR, new CommandInfo(":10vctr", "Set 10VCTR.", true, true));
            cmdDict.Add(CommandId.WriteTECCTRL, new CommandInfo(":tecctrl", "Set TECCTRL.", true, true));
            cmdDict.Add(CommandId.WriteSPARE, new CommandInfo(":spare", "Set SPARE.", true, true));
        }

        /// <summary>
        /// Writes gen6 laser functionality to file.
        /// </summary>
        /// <param name="commandFilePath">Path to command file.</param>
        /// <param name="message">Eventual error message</param>
        /// <returns>true if successful else false</returns>
        public override bool SaveCommand(string commandFilePath, out string message)
        {
            try
            {
                using (StreamWriter writer = new StreamWriter(commandFilePath))
                {
                    Tec[] tecs = GetTecs();

                    // Some useful information.
					writer.WriteLine("%----------------------------------------------------------------");
					writer.WriteLine("%");
					writer.WriteLine("%  Laser Driver Settings");
					writer.WriteLine("%");
					writer.WriteLine("%  Device Serial Number: {0}", SerialNumber);
					writer.WriteLine("%");
					writer.WriteLine("%  Date: {0}", DateTime.Now);
					writer.WriteLine("%");
					writer.WriteLine("%----------------------------------------------------------------");
					writer.WriteLine();

					// Set max laser diode current.
					writer.WriteLine("% Set max laser diode current.");
					writer.WriteLine("smlc {0}", MaxLaserCurrent);
					writer.WriteLine();

					// Laser PID settings.
					writer.WriteLine("% Laser PID settings.");
					writer.WriteLine("slpid {0} {1}", LdKp, LdKi);
					writer.WriteLine();

					// Maximum laser power (mW).
					writer.WriteLine("% Maximum laser power (mW).");
					writer.WriteLine("smlp {0}", MaxLaserPower);
					writer.WriteLine();

					// Laser Current set point.
					writer.WriteLine("% Laser Current Set Point (A).");
					writer.WriteLine("slc {0}", LaserCurrentSetPoint);
					writer.WriteLine();

					// Laser Power set point.
					writer.WriteLine("% Laser Power Set Point (mW).");
					writer.WriteLine("slp {0}", LaserPowerSetPoint);
					writer.WriteLine();

					// TEC delta setting.
					writer.WriteLine("% TEC delta setting (C).");
					writer.WriteLine("stdt {0}", TecsDeltaTemperature.ToString(new CultureInfo("en-US")));
					writer.WriteLine();

					// TEC PID settings.
					writer.WriteLine("% TEC PID Settings.");
					writer.WriteLine("spidtec {0} {1} {2}",
					    tecs[0].Kp.ToString(new CultureInfo("en-US")),
					    tecs[0].Ki.ToString(new CultureInfo("en-US")),
					    tecs[0].Kd.ToString(new CultureInfo("en-US")));
					writer.WriteLine();

					// TEC temperature set point.
					writer.WriteLine("% TEC Temperature Set Point (C).");
					writer.WriteLine("stect {0}", tecs[0].Temp);
					writer.WriteLine();

					// TEC max drive.
					writer.WriteLine("% TEC Max Drive (%).");
					writer.WriteLine("stmd {0}", tecs[0].MaxDrive);
					writer.WriteLine();

					// TEC Enable states.
					writer.WriteLine("% TEC Enabled States.");
					writer.WriteLine("{0}tec", tecs[0].Enabled ? "e" : "d");
					writer.WriteLine();

					// Autostart Configuration.
					writer.WriteLine("% CDRH (Key Switch) Enabled State (dks = disabled, eks = enabled).");
					writer.WriteLine(Cdrh ? "eks" : "dks");
					writer.WriteLine();

					writer.WriteLine("% Autostart Enabled State (da = disabled, ea = enabled).");
					writer.WriteLine(AutoStart ? "ea" : "da");
					writer.WriteLine();

					writer.Close();
				}
			}
            catch (Exception ex)
            {
                message = ex.Message;
                return false;
            }
            message = String.Empty;
            return true;
        }

    }
}
