﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Globalization;
using System.IO;

namespace CoboltMonitor.Laser
{
    /// <summary>
    /// Class defining a LaserUpdatedEvent on event LaserUpdatedEvent
    /// </summary>
    public class LaserUpdatedEvent : EventArgs
    {
        private CommandResponse commandResponse;

        public LaserUpdatedEvent(CommandResponse commandresponse)
        {
            commandResponse = commandresponse;
        }

        public CommandResponse CommandResponse
        {
            get { return commandResponse; }
        }
    }
    
    /// <summary>
    /// Class describing common laser functionallity. To be inherited by specific laser class.
    /// </summary>
    public abstract class LaserCommon
    {
        protected int opMode;       // The lasers operating mode

        public event EventHandler<LaserUpdatedEvent> LaserUpdatedEvent; // Event raised when data has been received from serial handler
        protected SerialHandler serialHandler;                          // Handles all the serial communications

        private DateTime timeOfNextGoh;                                // Contains the time to send the next GetOperatingHours commando
        protected System.Timers.Timer timerComm;                          // Timer for periodic commands to be sent to the laser.
        protected bool getValues = true;                                 // Tells if polling of setpoints shall be made.

        protected static Dictionary<CommandId, CommandInfo> cmdDict;   // Dictionary with all valid commands for current laser

        protected Tec[] tecs;                                          // Contains data for the lasers tec(s)

        protected int serialNumber = -1;
        private string firmwareVersion = String.Empty;
        private string firmwareDate = String.Empty;
        private string firmwareBuildDate = String.Empty;
        
        protected float tecsDeltaTemperature = 0.2f;
        protected float laserCurrentSetPoint = 0.0f;
        protected float laserPowerSetPoint = 0.0f;
        protected float laserCurrent = 0.0f;
        protected float laserPower = 0.0f;
        protected float warmUpCurrent = 0.0f;
        protected float warmUpTime = 0.0f;
        protected float photoDiodeVolt = 0.0f;
        protected float laserDiodeVolt = 0.0f;
        protected float rmsVolt = 0.0f;
        protected float rmsPercent = 0.0f;
        protected float maxLaserCurrent = 0.0f;
        protected float ldKp = 0.0f;
        protected float ldKi = 0.0f;
        protected float maxLaserPower = 0.0f;

        protected bool cdrh = false;
        protected bool autoStart = false;
        protected bool laserEnabled = false;
        protected bool directInput = false;
        protected int ledBits = 0;

        protected float currentRangeMin = 0.25f;
        protected float currentRangeMax = 6.5f;
        protected float powerRangeMin = 0.0f;
        protected float powerRangeMax = 3000.0f;

        protected bool constantCurrent = false;
        protected bool constantPower = true;

        protected float neVolt = 0.0f;
        protected string cmdName = String.Empty;
        protected string operatingHours = String.Empty;
        protected int laserFault = 0;
        protected bool ldCalibrated = false;
        protected bool tecLdCalibrated = true;

        protected float basePlateTemp = 0;
        protected float boardTemp = 0;

        protected float modPowerRangeMin = 0.0f;
        protected float modPowerRangeMax = 500.0f;
        protected float laserModPower = 0.0f;
        protected int modbm = 0;
        protected bool mod50ohm = false;
        protected bool mod1kohm = true;
        protected bool modulation = false;

        protected object valueLock;

        public LaserCommon(SerialHandler sh, int nTecs)
        {
            tecs = new Tec[nTecs];
            for (int i = 0; i < nTecs; ++i)
            {
                tecs[i] = new Tec();
            }

            serialHandler = sh;
            //Listen on data received events from serialhandler
            serialHandler.ResponseReceived += new EventHandler<SerialEvent>(SerialPort_DataReceived);

            //Get operating hours every 20 sec
            timeOfNextGoh = DateTime.Now.AddSeconds(20);

            //Periodic transmission to laser every 200 ms
            timerComm = new System.Timers.Timer();
            timerComm.Interval = 200;
            timerComm.Elapsed += new System.Timers.ElapsedEventHandler(timerComm_Tick);

            firmwareVersion = "Unknown";
            firmwareDate = "Unknown";
            firmwareBuildDate = "Unknown";

            cmdDict = new Dictionary<CommandId, CommandInfo>();
            BuildDictionary();

            valueLock = new object();
        }
        
        #region Properties
        /// <summary>
        /// Gets or sets whether laser values should be read from the laser.
        /// </summary>
        public bool GetValues
        {
            get
            {
                lock (valueLock)
                {
                    return getValues;
                }
            }
            set
            {
                lock (valueLock)
                {
                    getValues = value;
                }
            }
        }

        /// <summary>
        /// Gets the operating mode.
        /// </summary>
        public int OperatingMode
        {
            get
            {
                lock (valueLock)
                {
                    return opMode;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    opMode = value;
                }
            }
        }
        
        /// <summary>
        /// Gets max power.
        /// </summary>
        public float PowerRangeMax
        {
            get
            {
                lock (valueLock)
                {
                    return powerRangeMax;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    powerRangeMax = value;
                }
            }
        }

        /// <summary>
        /// Gets min mod current.
        /// </summary>
        public float CurrentRangeMin
        {
            get
            {
                lock (valueLock)
                {
                    return currentRangeMin;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    currentRangeMin = value;
                }
            }
        }

        /// <summary>
        /// Gets max current.
        /// </summary>
        public float CurrentRangeMax
        {
            get
            {
                lock (valueLock)
                {
                    return currentRangeMax;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    currentRangeMax = value;
                }
            }
        }

        /// <summary>
        /// Gets min mod power.
        /// </summary>
        public float PowerRangeMin
        {
            get
            {
                lock (valueLock)
                {
                    return powerRangeMin;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    powerRangeMin = value;
                }
            }
        }
        

        /// <summary>
        /// Gets the TEC collection.
        /// </summary>
        public Tec[] GetTecs()
        {
            lock(tecs.SyncRoot)
            {
                return (Tec[])tecs.Clone();
            }
        }

        /// <summary>
        /// Gets the lasers serial number.
        /// </summary>
        public int SerialNumber
        {
            get
            {
                lock (valueLock)
                {
                    return serialNumber;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    serialNumber = value;
                }
            }
        }

        /// <summary>
        /// Gets the lasers firmware number.
        /// </summary>
        public string FirmwareVersion
        {
            get
            {
                lock (valueLock)
                {
                    return firmwareVersion;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    firmwareVersion = value;
                }
            }
        }

        /// <summary>
        /// Gets firmware date.
        /// </summary>
        public string FirmwareDate
        {
            get
            {
                lock (valueLock)
                {
                    return firmwareDate;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    firmwareDate = value;
                }
            }
        }

        /// <summary>
        /// Gets firmware build date.
        /// </summary>
        public string FirmwareBuildDate
        {
            get
            {
                lock (valueLock)
                {
                    return firmwareBuildDate;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    firmwareBuildDate = value;
                }
            }
        }

        /// <summary>
        /// Gets the Operating Hourst.
        /// </summary>
        public string CmdName
        {
            get
            {
                lock (valueLock)
                {
                    return cmdName;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    cmdName = value;
                }
            }
        }

        /// <summary>
        /// Gets the Operating Hourst.
        /// </summary>
        public string OperatingHours
        {
            get
            {
                lock (valueLock)
                {
                    return operatingHours;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    operatingHours = value;
                }
            }
        }

        /// <summary>
        /// Gets the number of tecs in the laser.
        /// </summary>
        public int NumberOfTecs
        {
            get
            {
                lock (valueLock)
                {
                    return tecs.Length;
                }
            }
        }

        /// <summary>
        /// Gets whether the laser is enabled or not.
        /// </summary>
        public bool LaserEnabled
        {
            get
            {
                lock (valueLock)
                {
                    return laserEnabled;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    laserEnabled = value;
                }
            }
        }

        public bool LdCalibrated
        {
            get
            {
                lock (valueLock)
                {
                    return ldCalibrated;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    ldCalibrated = value;
                }
            }
        }

        /// <summary>
        /// Gets whether autostart mode is on or not.
        /// </summary>
        public bool AutoStart
        {
            get
            {
                lock (valueLock)
                {
                    return autoStart;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    autoStart = value;
                }
            }
        }

        /// <summary>
        /// Gets whether direct input is on or not.
        /// </summary>
        public bool DirectInput
        {
            get
            {
                lock (valueLock)
                {
                    return directInput;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    directInput = value;
                }
            }
        }

        /// <summary>
        /// Gets whether CDRH is on or not.
        /// </summary>
        public bool Cdrh
        {
            get
            {
                lock (valueLock)
                {
                    return cdrh;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    cdrh = value;
                }
            }
        }

        /// <summary>
        /// Gets the lasers current setpoint.
        /// </summary>
        public float LaserCurrentSetPoint
        {
            get
            {
                lock (valueLock)
                {
                    return laserCurrentSetPoint;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    laserCurrentSetPoint = value;
                }
            }
        }

        /// <summary>
        /// Gets the lasers power setpoint.
        /// </summary>
        public float LaserPowerSetPoint
        {
            get
            {
                lock (valueLock)
                {
                    return laserPowerSetPoint;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    laserPowerSetPoint = value;
                }
            }
        }

        /// <summary>
        /// Gets lasers current.
        /// </summary>
        public float LaserCurrent
        {
            get
            {
                lock (valueLock)
                {
                    return laserCurrent;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    laserCurrent = value;
                }
            }
        }

        /// <summary>
        /// Gets the lasers power.
        /// </summary>
        public float LaserPower
        {
            get
            {
                lock (valueLock)
                {
                    return laserPower;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    laserPower = value;
                }
            }
        }

        /// <summary>
        /// Gets the lasers warmup current.
        /// </summary>
        public float WarmUpCurrent
        {
            get
            {
                lock (valueLock)
                {
                    return warmUpCurrent;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    warmUpCurrent = value;
                }
            }
        }

        /// <summary>
        /// Gets the lasers warmup time.
        /// </summary>
        public float WarmUpTime
        {
            get
            {
                lock (valueLock)
                {
                    return warmUpTime;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    warmUpTime = value;
                }
            }
        }

        /// <summary>
        /// Gets the voltage on the photo diode.
        /// </summary>
        public float PhotoDiodeVoltage
        {
            get
            {
                lock (valueLock)
                {
                    return photoDiodeVolt;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    photoDiodeVolt = value;
                }
            }
        }

        /// <summary>
        /// Gets the voltage on the laser diode.
        /// </summary>
        public float LaserDiodeVoltage
        {
            get
            {
                lock (valueLock)
                {
                    return laserDiodeVolt;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    laserDiodeVolt = value;
                }
            }
        }

        /// <summary>
        /// Gets the RMS voltage.
        /// </summary>
        public float RmsVoltage
        {
            get
            {
                lock (valueLock)
                {
                    return rmsVolt;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    rmsVolt = value;
                }
            }
        }

        /// <summary>
        /// Gets the RMS percent.
        /// </summary>
        public float RmsPercent
        {
            get
            {
                lock (valueLock)
                {
                    return rmsPercent;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    rmsPercent = value;
                }
            }
        }

        /// <summary>
        /// Gets ledbits.
        /// </summary>
        public int LedBits
        {
            get
            {
                lock (valueLock)
                {
                    return ledBits;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    ledBits = value;
                }
            }
        }

        /// <summary>
        /// Gets whether constant current is on or not.
        /// </summary>
        public bool ConstantCurrent
        {
            get
            {
                lock (valueLock)
                {
                    return constantCurrent;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    constantCurrent = value;
                }
            }
        }

        /// <summary>
        /// Gets whether constant power is on or not.
        /// </summary>
        public bool ConstantPower
        {
            get
            {
                lock (valueLock)
                {
                    return constantPower;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    constantPower = value;
                }
            }
        }

        /// <summary>
        /// Gets the lasers fault value.
        /// </summary>
        public int LaserFault
        {
            get
            {
                lock (valueLock)
                {
                    return laserFault;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    laserFault = value;
                }
            }
        }

        /// <summary>
        /// Gets the lasers current NeVolt.
        /// </summary>
        public float NeVolt
        {
            get
            {
                lock (valueLock)
                {
                    return neVolt;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    neVolt = value;
                }
            }
        }

        /// <summary>
        /// Gets whether tec has been calibrated or not.
        /// </summary>
        public bool TecLdCalibrated
        {
            get
            {
                lock (valueLock)
                {
                    return tecLdCalibrated;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    tecLdCalibrated = value;
                }
            }
        }

        /// <summary>
        /// Gets the modulation settings.
        /// </summary>
        public int ModulationSettings
        {
            get
            {
                lock (valueLock)
                {
                    return modbm;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    modbm = value;
                }
            }
        }

        /// <summary>
        /// Gets the lasers power setpoint.
        /// </summary>
        public float LaserModPower
        {
            get
            {
                lock (valueLock)
                {
                    return laserModPower;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    laserModPower = value;
                }
            }
        }

        /// <summary>
        /// Gets the lasers baseplate temperature.
        /// </summary>
        public float BasePlateTemp
        {
            get
            {
                lock (valueLock)
                {
                    return basePlateTemp;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    basePlateTemp = value;
                }
            }
        }

        /// <summary>
        /// Gets the lasers board temperature.
        /// </summary>
        public float BoardTemp
        {
            get
            {
                lock (valueLock)
                {
                    return boardTemp;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    boardTemp = value;
                }
            }
        }

        /// <summary>
        /// Gets whether 50 ohm is enabled or not.
        /// </summary>
        public bool Analog50Ohm
        {
            get
            {
                lock (valueLock)
                {
                    return mod50ohm;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    mod50ohm = value;
                }
            }
        }

        /// <summary>
        /// Gets whether 1k ohm is enabled or not.
        /// </summary>
        public bool Analog1kOhm
        {
            get
            {
                lock (valueLock)
                {
                    return mod1kohm;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    mod1kohm = value;
                }
            }
        }

        /// <summary>
        /// Gets whether modulation is on or not.
        /// </summary>
        public bool Modulation
        {
            get
            {
                lock (valueLock)
                {
                    return modulation;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    modulation = value;
                }
            }
        }

        /// <summary>
        /// Gets max mod power.
        /// </summary>
        public float ModPowerRangeMax
        {
            get
            {
                lock (valueLock)
                {
                    return modPowerRangeMax;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    modPowerRangeMax = value;
                }
            }
        }

        /// <summary>
        /// Gets min mod power.
        /// </summary>
        public float ModPowerRangeMin
        {
            get
            {
                lock (valueLock)
                {
                    return modPowerRangeMin;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    modPowerRangeMin = value;
                }
            }
        }

        /// <summary>
        /// Gets max laser current.
        /// </summary>
        public float MaxLaserCurrent
        {
            get
            {
                lock (valueLock)
                {
                    return maxLaserCurrent;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    maxLaserCurrent = value;
                }
            }
        }

        /// <summary>
        /// Gets max laser power.
        /// </summary>
        public float MaxLaserPower
        {
            get
            {
                lock (valueLock)
                {
                    return maxLaserPower;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    maxLaserPower = value;
                }
            }
        }

        /// <summary>
        /// Gets LdKi
        /// </summary>
        public float LdKi
        {
            get
            {
                lock (valueLock)
                {
                    return ldKi;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    ldKi = value;
                }
            }
        }

        /// <summary>
        /// Gets LdKp
        /// </summary>
        public float LdKp
        {
            get
            {
                lock (valueLock)
                {
                    return ldKp;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    ldKp = value;
                }
            }
        }

        /// <summary>
        /// Gets tecs delta temp
        /// </summary>
        public float TecsDeltaTemperature
        {
            get
            {
                lock (valueLock)
                {
                    return tecsDeltaTemperature;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    tecsDeltaTemperature = value;
                }
            }
        }

        /// <summary>
        /// Gets the laser diode voltage
        /// </summary>
        public float LaserDiodeVolt
        {
            get
            {
                lock (valueLock)
                {
                    return laserDiodeVolt;
                }
            }
            protected set
            {
                lock (valueLock)
                {
                    laserDiodeVolt = value;
                }
            }
        }

        #endregion

        #region Laser commands
        /// <summary>
        /// Gets whether connection to laser is established.
        /// </summary>
        public virtual bool IsConnected()
        {
            return serialHandler.IsSerialPortOpen;
        }

        /// <summary>
        /// Sends set power command to the laser.
        /// </summary>
        /// <param name="power">The laser power to be set.</param>
        public virtual bool SetPower(float power)
        {
            // Check that new set-point is within range.
            if (power >= powerRangeMin && power <= powerRangeMax)
            {
                SendCommand(CommandId.SetLaserPower, power.ToString(new CultureInfo("en-US")));
                return true;
            }
            return false;
        }

        /// <summary>
        /// Sends get power command to the laser.
        /// </summary>
        public virtual void GetPower()
        {
            SendCommand(CommandId.GetLaserPower, "");
        }

        /// <summary>
        /// Sends set laser current command to the laser.
        /// </summary>
        /// <param name="power">The laser current to be set.</param>
        public virtual bool SetCurrent(float current)
        {
            // Check that new set-point is within range.
            if (current >= currentRangeMin && current <= currentRangeMax)
            {
                SendCommand(CommandId.SetLaserCurrent, current.ToString(new CultureInfo("en-US")));
                return true;
            }
            return false;
        }

        /// <summary>
        /// Sends get laser current command to the laser.
        /// </summary>
        public virtual void GetCurrent()
        {
            SendCommand(CommandId.GetLaserCurrent, "");
        }

        /// <summary>
        /// Sends set laser warmup current command to the laser.
        /// </summary>
        /// <param name="current">The laser warmup current to be set.</param>
        /// <returns>true if current is within correct range else false</returns>
        public virtual bool SetWarmUpCurrent(float current)
        {
            // Check that new set-point is within range.
            if (current >= currentRangeMin && current <= currentRangeMax)
            {
                SendCommand(CommandId.SetAutostartWarmUpCurrent, current.ToString(new CultureInfo("en-US")));
                return true;
            }
            return false;
        }

        /// <summary>
        /// Sends get laser warmup current command to the laser.
        /// </summary>
        public virtual void GetWarmUpCurrent()
        {
            SendCommand(CommandId.GetAutostartWarmUpCurrent, "");
        }

        /// <summary>
        /// Sends set laser warmup delay command to the laser.
        /// </summary>
        /// <param name="current">The laser warmup delay to be set.</param>
        /// <returns>true if delay is within correct range else false</returns>
        public virtual bool SetWarmUpDelay(int current)
        {
            // Check that new set-point is within range.
            if (current >= 0)
            {
                SendCommand(CommandId.SetWarmUpDelay, current.ToString(new CultureInfo("en-US")));
                return true;
            }
            return false;
        }

        /// <summary>
        /// Sends command to enable the laser.
        /// </summary>
        public virtual void SetLaserOn()
        {
            SendCommand(CommandId.LaserDiodeEnable, "");
        }

        /// <summary>
        /// Sends command to disable the laser.
        /// </summary>        
        public virtual void SetLaserOff()
        {
            SendCommand(CommandId.LaserDisable, "");
        }

        /// <summary>
        /// Sends a get warmup delay command to the laser.
        /// </summary>
        public virtual void GetWarmUpDelay()
        {
            SendCommand(CommandId.GetWarmUpDelay, "");
        }

        /// <summary>
        /// Sends a enable tec1 command to the laser.
        /// </summary>
        public virtual void EnableTec1()
        {
            SendCommand(CommandId.EnableTec1, "");
        }

        /// <summary>
        /// Sends a disable tec1 command to the laser.
        /// </summary>
        public virtual void DisableTec1()
        {
            SendCommand(CommandId.DisableTec1, "");
        }

        /// <summary>
        /// Sends a enable tec2 command to the laser.
        /// </summary>
        public virtual void EnableTec2()
        {
            SendCommand(CommandId.EnableTec2, "");
        }

        /// <summary>
        /// Sends a disable tec2 command to the laser.
        /// </summary>
        public virtual void DisableTec2()
        {
            SendCommand(CommandId.DisableTec2, "");
        }

        /// <summary>
        /// Sends a enable tec3 command to the laser.
        /// </summary>
        public virtual void EnableTec3()
        {
            SendCommand(CommandId.EnableTec3, "");
        }

        /// <summary>
        /// Sends a disable tec3 command to the laser.
        /// </summary>
        public virtual void DisableTec3()
        {
            SendCommand(CommandId.DisableTec3, "");
        }

        /// <summary>
        /// Sends a enable tec4 command to the laser.
        /// </summary>
        public virtual void EnableTec4()
        {
            SendCommand(CommandId.EnableTec4, "");
        }

        /// <summary>
        /// Sends a disable tec4 command to the laser.
        /// </summary>
        public virtual void DisableTec4()
        {
            SendCommand(CommandId.DisableTec4, "");
        }

        /// <summary>
        /// Sends a get tec1 enabled state command to the laser.
        /// </summary>
        public virtual void GetTec1EnabledState()
        {
            SendCommand(CommandId.GetTec1EnabledState, "");
        }

        /// <summary>
        /// Sends a get tec2 enabled state command to the laser.
        /// </summary>
        public virtual void GetTec2EnabledState()
        {
            SendCommand(CommandId.GetTec2EnabledState, "");
        }

        /// <summary>
        /// Sends a get tec3 enabled state command to the laser.
        /// </summary>
        public virtual void GetTec3EnabledState()
        {
            SendCommand(CommandId.GetTec3EnabledState, "");
        }

        /// <summary>
        /// Sends a get tec4 enabled state command to the laser.
        /// </summary>
        public virtual void GetTec4EnabledState()
        {
            SendCommand(CommandId.GetTec4EnabledState, "");
        }

        /// <summary>
        /// Sends command to get tec1 running to the laser.
        /// </summary>
        public virtual void RunTec()
        {
            SendCommand(CommandId.RunTec, "");
        }

        /// <summary>
        /// Sends command to get tec to stop to the laser.
        /// </summary>
        public virtual void StopTec()
        {
            SendCommand(CommandId.StopTec, "");
        }

        /// <summary>
        /// Sends command to get tec1 running to the laser.
        /// </summary>
        public virtual void RunTec1()
        {
            SendCommand(CommandId.RunTec1, "");
        }

        /// <summary>
        /// Sends command to get tec1 to stop to the laser.
        /// </summary>
        public virtual void StopTec1()
        {
            SendCommand(CommandId.StopTec1, "");
        }

        /// <summary>
        /// Sends command to get tec2 running to the laser.
        /// </summary>
        public virtual void RunTec2()
        {
            SendCommand(CommandId.RunTec2, "");
        }

        /// <summary>
        /// Sends command to get tec2 to stop to the laser.
        /// </summary>
        public virtual void StopTec2()
        {
            SendCommand(CommandId.StopTec2, "");
        }

        /// <summary>
        /// Sends command to get tec3 running to the laser.
        /// </summary>
        public virtual void RunTec3()
        {
            SendCommand(CommandId.RunTec3, "");
        }

        /// <summary>
        /// Sends command to get tec3 to stop to the laser.
        /// </summary>
        public virtual void StopTec3()
        {
            SendCommand(CommandId.StopTec3, "");
        }

        /// <summary>
        /// Sends command to get tec4 running to the laser.
        /// </summary>
        public virtual void RunTec4()
        {
            SendCommand(CommandId.RunTec4, "");
        }

        /// <summary>
        /// Sends command to get tec4 to stop to the laser.
        /// </summary>
        public virtual void StopTec4()
        {
            SendCommand(CommandId.StopTec4, "");
        }

        /// <summary>
        /// Sends command to set tec temperature to the laser.
        /// <param name="temp">The temperature to be set on tec 1</param>
        /// </summary>
        public virtual void SetTecTemp(string temp)
        {
            SendCommand(CommandId.SetTecTemperature, temp);
        }

        /// <summary>
        /// Sends command to get tec temperature to the laser.
        /// </summary>
        public virtual void GetTecTemp()
        {
            SendCommand(CommandId.GetTecTemperature, "");
        }

        /// <summary>
        /// Sends command to set tec1 temperature to the laser.
        /// <param name="temp">The temperature to be set on tec 1</param>
        /// </summary>
        public virtual void SetTec1Temp(string temp)
        {
            SendCommand(CommandId.SetTec1Temperature, temp);
        }

        /// <summary>
        /// Sends command to get tec1 temperature to the laser.
        /// </summary>
        public virtual void GetTec1Temp()
        {
            SendCommand(CommandId.GetTec1Temperature, "");
        }

        /// <summary>
        /// Sends command to set tec2 temperature to the laser.
        /// <param name="temp">The temperature to be set on tec 2</param>
        /// </summary>
        public virtual void SetTec2Temp(string temp)
        {
            SendCommand(CommandId.SetTec2Temperature, temp);
        }

        /// <summary>
        /// Sends command to get tec2 temperature to the laser.
        /// </summary>
        public virtual void GetTec2Temp()
        {
            SendCommand(CommandId.GetTec2Temperature, "");
        }

        /// <summary>
        /// Sends command to set tec3 temperature to the laser.
        /// <param name="temp">The temperature to be set on tec 3</param>
        /// </summary>
        public virtual void SetTec3Temp(string temp)
        {
            SendCommand(CommandId.SetTec3Temperature, temp);
        }

        /// <summary>
        /// Sends command to get tec3 temperature to the laser.
        /// </summary>
        public virtual void GetTec3Temp()
        {
            SendCommand(CommandId.GetTec3Temperature, "");
        }

        /// <summary>
        /// Sends command to set tec4 temperature to the laser.
        /// <param name="temp">The temperature to be set on tec 4</param>
        /// </summary>
        public virtual void SetTec4Temp(string temp)
        {
            SendCommand(CommandId.SetTec4Temperature, temp);
        }

        /// <summary>
        /// Sends command to get tec4 temperature to the laser.
        /// </summary>
        public virtual void GetTec4Temp()
        {
            SendCommand(CommandId.GetTec4Temperature, "");
        }

        /// <summary>
        /// Sends command to calibrate power to the laser.
        /// <param name="power">The power to calibrate to.</param>
        /// </summary>
        public virtual bool SetCalibratePower(float power)
        {
            // Check that new set-point is within range.
            if (power >= powerRangeMin && power <= powerRangeMax)
            {
                SendCommand(CommandId.CalibrateWattsPerVolt, power.ToString(new CultureInfo("en-US")));
                return true;
            }
            return false;
        }

        /// <summary>
        /// Sends restart command to the laser.
        /// </summary>
        public virtual void Restart()
        {
            SendCommand(CommandId.Restart, "");
        }

        /// <summary>
        /// Sends abort command to the laser.
        /// </summary>
        public virtual void Abort()
        {
            SendCommand(CommandId.Abort, "");
        }

        /// <summary>
        /// Sends set command name command to the laser.
        /// <param name="cmdName">The command name</param>
        /// </summary>
        public virtual void SetCmdName(string cmdName)
        {
            SendCommand(CommandId.SetCmdName, cmdName);
            lock (valueLock)
            {
                getValues = true;
            }       
        }

        /// <summary>
        /// Sends command to set the laser in cp mode.
        /// </summary>
        public virtual void EnterCpMode()
        {
            SendCommand(CommandId.EnterCpMode, "");
            lock (valueLock)
            {
                getValues = true;
            }
        }

        /// <summary>
        /// Sends command to set the laser in cc mode.
        /// </summary>
        public virtual void EnterCcMode()
        {
            SendCommand(CommandId.EnterCcMode, "");
            lock (valueLock)
            {
                getValues = true;
            }
        }

        /// <summary>
        /// Sends set pid command to the laser for specified tec.
        /// <param name="pid">The pid value to set</param>
        /// <returns>true if command was sent, false if tec idx was wrong</returns>
        /// </summary>
        public virtual void SetPidForTec(string pid)
        {
            SendCommand(CommandId.SetPidForTec, pid);
        }

        /// <summary>
        /// Sends get pid command to the laser for specified tec.
        /// <returns>true if command was sent, false if tec idx was wrong</returns>
        /// </summary>
        public virtual void GetPidForTec()
        {
            SendCommand(CommandId.GetPidForTec, "");
        }

        /// <summary>
        /// Sends set pid command to the laser for specified tec.
        /// <param name="tec">The tec to set pid for</param>
        /// <param name="pid">The pid value to set</param>
        /// <returns>true if command was sent, false if tec idx was wrong</returns>
        /// </summary>
        public virtual bool SetPidForTec(int tec, string pid)
        {
            switch (tec)
            {
                case 0:
                    SendCommand(CommandId.SetPidForTec1, pid);
                    break;
                case 1:
                    SendCommand(CommandId.SetPidForTec2, pid);
                    break;
                case 2:
                    SendCommand(CommandId.SetPidForTec3, pid);
                    break;
                case 3:
                    SendCommand(CommandId.SetPidForTec4, pid);
                    break;
                default:
                    return false;
            }
            return true;
        }

        /// <summary>
        /// Sends get pid command to the laser for specified tec.
        /// <param name="tec">The tec to get pid for</param>
        /// <returns>true if command was sent, false if tec idx was wrong</returns>
        /// </summary>
        public virtual bool GetPidForTec(int tec)
        {
            switch (tec)
            {
                case 0:
                    SendCommand(CommandId.GetPidForTec1, "");
                    break;
                case 1:
                    SendCommand(CommandId.GetPidForTec2, "");
                    break;
                case 2:
                    SendCommand(CommandId.GetPidForTec3, "");
                    break;
                case 3:
                    SendCommand(CommandId.GetPidForTec4, "");
                    break;
                default:
                    return false;
            }
            return true;
        }

        /// <summary>
        /// Sends ip log curernt command to set the laser.
        /// </summary>
        public virtual void SetIPLogCurrent(float inext)
        {
            SendCommand(CommandId.SetIpLogCurrent, inext.ToString(new CultureInfo("en-US")));
        }

        /// <summary>
        /// Sends a enable autostart command to the laser.
        /// </summary>
        public virtual void EnableAutoStart()
        {
            SendCommand(CommandId.EnableAutostart, "");
            lock (valueLock)
            {
                getValues = true;
            }
        }

        /// <summary>
        /// Sends a disable autostart command to the laser.
        /// </summary>
        public virtual void DisableAutoStart()
        {
            SendCommand(CommandId.DisableAutostart, "");
            lock (valueLock)
            {
                getValues = true;
            }
        }

        /// <summary>
        /// Sends a enable key switch command to the laser.
        /// </summary>
        public virtual void EnableKeySwitch()
        {
            SendCommand(CommandId.EnableKeySwitch, "");
            lock (valueLock)
            {
                getValues = true;
            }
        }

        /// <summary>
        /// Sends a disable key switch command to the laser.
        /// </summary>
        public virtual void DisableKeySwitch()
        {
            SendCommand(CommandId.DisableKeySwitch, "");
            lock (valueLock)
            {
                getValues = true;
            }
        }

        /// <summary>
        /// Sends command to enable 5V direct input to the laser.
        /// </summary>
        public virtual void Enable5VDirectInputStart()
        {
            SendCommand(CommandId.Enable5VDirectInputStart, "");
            lock (valueLock)
            {
                getValues = true;
            }
        }

        /// <summary>
        /// Sends command to disable 5V direct input to the laser.
        /// </summary>
        public virtual void Disable5VDirectInputStart()
        {
            SendCommand(CommandId.Disable5VDirectInputStart, "");
            lock (valueLock)
            {
                getValues = true;
            }
        }

        /// <summary>
        /// Sends clear fault command to the laser.
        /// </summary>
        public virtual void ClearFault()
        {
            SendCommand(CommandId.ClearFault, "");
            lock (valueLock)
            {
                getValues = true;
            }

        }
        /// <summary>
        /// Sends a enable tec command to the laser.
        /// </summary>
        public virtual void EnableTec()
        {
            SendCommand(CommandId.EnableTec, "");
        }

        /// <summary>
        /// Sends a disable tec command to the laser.
        /// </summary>
        public virtual void DisableTec()
        {
            SendCommand(CommandId.DisableTec, "");
        }

        /// <summary>
        /// Sends a get tec enabled state command to the laser.
        /// </summary>
        public virtual void GetTecEnabledState()
        {
            SendCommand(CommandId.GetTecEnabledState, "");
        }

        public bool SetLaserModPower(float modPower)
        {
            if (modPower >= modPowerRangeMin && modPower <= modPowerRangeMax)
            {
                SendCommand(CommandId.SetLaserModPower, modPower.ToString(new CultureInfo("en-US")));
                return true;
            }
            return false;
        }

        public void GetLaserModPower()
        {
            SendCommand(CommandId.GetLaserModPower, "");
        }

        public void EnterModMode()
        {
            SendCommand(CommandId.EnterModMode, "");
        }

        public void SetDigitalModulationEnabledState(string state)
        {
            SendCommand(CommandId.SetDigitalModulationEnabledState, state);
        }

        public void SetAnalogModulationEnabledState(string state)
        {
            SendCommand(CommandId.SetAnalogModulationEnabledState, state);
        }

        public void SetAnalogLowImpedanceState(string state)
        {
            SendCommand(CommandId.SetAnalogLowImpedanceState, state);
        }

        public void EnterOnOffMode()
        {
            SendCommand(CommandId.EnterOnOffMode, "");
            lock (valueLock)
            {
                getValues = true;
            }
        }

        public void ExitOnOffMode()
        {
            SendCommand(CommandId.ExitOnOffMode, "");
            lock (valueLock)
            {
                getValues = true;
            }
        }
        #endregion

        #region Laser communication
        /// <summary>
        /// Open a serial connection to the laser driver.
        /// </summary>
        public bool Connect(string serialPortName, int serialBaudRate)
        {
            bool retVal = false;
            retVal = serialHandler.OpenSerialPort(serialPortName, serialBaudRate);
            
            if (retVal)
            {
                timerComm.Enabled = true;
                GetValues = true;
            }
            return retVal;

        }

        /// <summary>
        /// Close the connection to the laser driver.
        /// </summary>
        public void Disconnect()
        {
            serialHandler.CloseSerialPort();
            timerComm.Enabled = false;
        }

        /// <summary>
        /// Sends a command to the laser driver.
        /// </summary>
        /// <param name="id">The command id.</param>
        /// <param name="arg">Data to be sent to the laser.</param>
        protected virtual void SendCommand(CommandId commandId, string arg)
        {
            serialHandler.SendCommand(new CommandResponse(commandId, arg, AsString(commandId, arg)));
        }
        
        /// <summary>
        /// Sends a command to the laser driver.
        /// </summary>
        /// <param name="cmd">Raw string command to be sent.</param>
        public void SendCommand(String cmd)
        {
            SendCommand(null, cmd);
        }
        
        /// <summary>
        /// Sends a command to the laser driver.
        /// </summary>
        /// <param name="cmd">String command to be sent.</param>
        public virtual void SendCommand(Object sender, String cmd)
        {
            string key;
            string args;

            ParseToKeyAndArgs(cmd, out key, out args);
            CommandId commandId;

            commandId = FindKey(key);
            serialHandler.SendCommand(new CommandResponse(sender, commandId, args, AsString(commandId, args)));
        }

        ///// <summary>
        ///// Sends a command to the laser driver.
        ///// </summary>
        ///// <param name="cmd">Raw string command to be sent.</param>
        public virtual void SendRawCommand(Object sender, String cmd)
        {
            string key;
            string args;

            ParseToKeyAndArgs(cmd, out key, out args);
            CommandId commandId;

            commandId = FindKey(key);
            serialHandler.SendCommand(new CommandResponse(sender, commandId, args, cmd));
        }

        /// <summary>
        /// Splits a laser command string into a string with key and one with arguments
        /// </summary>
        /// <param name="cmd">command string</param>
        /// <param name="key">string with key value</param>
        /// <param name="args">string with arguments</param>
        private void ParseToKeyAndArgs(string cmd, out string key, out string args)
        {
            string command = cmd;
            command = command.TrimEnd(new char[] { '\r' });
            int idx = command.IndexOf(' ');

            if (idx >= 0)
            {
                key = command.Substring(0, idx);
                args = command.Substring(idx + 1);
            }
            else
            {
                key = command;
                args = "";
            }
        }
        #endregion

        #region Response parsing functions
        /// <summary>
        /// Parses the laser command name response.
        /// </summary>
        /// <param name="resp">String with the command name</param>
        protected void ParseGetCommandNameResponse(String resp)
        {
            CmdName = resp;
        }

        /// <summary>
        /// Parses the laser operating hours response.
        /// </summary>
        /// <param name="resp">String with the operating hours</param>
        protected void GetOperatingHoursResponse(String resp)
        {
            OperatingHours = resp;
        }

        /// <summary>
        /// Parses a float value.
        /// </summary>
        /// <param name="resp">String to parse to a float</param>
        protected void ParseFloatResponse(String resp, out float outVar, float defaultValue)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
                outVar = f1;
            else
                outVar = defaultValue;
        }

        /// <summary>
        /// Parses the laser current setpoint data response.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        protected void ParseGetLaserCurrentResponse(String resp)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
            {
                LaserCurrentSetPoint = f1;
            }
        }

        /// <summary>
        /// Parses the laser warmup current data response.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        protected void ParseGetWarmUpCurrentResponse(String resp)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
            {
                WarmUpCurrent = f1;
            }
        }

        /// <summary>
        /// Parses the laser warmup time data response.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        protected void ParseGetWarmUpTimeResponse(String resp)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
            {
                WarmUpTime = f1;
            }
        }

        /// <summary>
        /// Parses the delta temp response.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        protected void ParseGetDeltaTempResponse(String resp)
        {
            float f1;

            ParseFloatResponse(resp, out f1, 0.2f);
            TecsDeltaTemperature = f1;

            lock (tecs.SyncRoot)
            {
                foreach (Tec t in tecs)
                    t.Delta = TecsDeltaTemperature;
            }
        }                 

        /// <summary>
        /// Parses the laser current data response.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        protected void ParseReadLaserCurrentResponse(String resp)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
            {
                LaserCurrent = f1;
            }
        }

        /// <summary>
        /// Parses the laser diode voltage data response.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        protected void ParseReadLaserDiodeVoltageResponse(String resp)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
            {
                LaserDiodeVolt = f1;
            }
        }

        /// <summary>
        /// Parses the Rms voltage data response.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        protected void ParseReadRmsVoltageResponse(String resp)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
            {
                RmsVoltage = f1;
            }
        }

        /// <summary>
        /// Parses the Rms percent data response.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        protected void ParseReadRmsPercentResponse(String resp)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
            {
                RmsPercent = f1;
            }
        }

        /// <summary>
        /// Parses the photo diode voltage data response.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        protected void ParseReadPhotoDiodeVoltageResponse(String resp)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
            {
                PhotoDiodeVoltage = f1;
            }
        }

        /// <summary>
        /// Parses the laser power data response.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        protected void ParseReadLaserPowerResponse(String resp)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
            {
                LaserPower = f1;
            }
        }

        /// <summary>
        /// Parses the laser power setpoint data response.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        protected void ParseGetLaserPowerResponse(String resp)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
            {
                LaserPowerSetPoint = f1;
            }
        }

        /// <summary>
        /// Parses the pid data response to a 'GetPidForTecn' command.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        /// <param name="tecIdx">TEC index</param>
        protected void ParseGetPidResponse(String[] args, int startIdx, int tecIdx)
        {
            float f1, f2, f3;

            if (!float.TryParse(args[startIdx], NumberStyles.Float, new CultureInfo("en-US"), out f1))
                return;

            if (!float.TryParse(args[startIdx + 1], NumberStyles.Float, new CultureInfo("en-US"), out f2))
                return;

            if (!float.TryParse(args[startIdx + 2], NumberStyles.Float, new CultureInfo("en-US"), out f3))
                return;

            lock (tecs.SyncRoot)
            {
                tecs[tecIdx].PidData(f1, f2, f3);
            }
        }

        /// <summary>
        /// Parses the pid data response to a 'GetPidForTecn' command.
        /// </summary>
        /// <param name="tecIdx">The tec index.</param>
        protected void ParseGetPidResponse(CommandResponse resp, int tecIdx)
        {
            string[] args;
            //float f1, f2, f3;

            args = resp.Resp.Split(new char[] { ' ' }, 3);
            ParseGetPidResponse(args, 0, tecIdx);
        }

        /// <summary>
        /// Parses the temp data response to a 'GetTecnTemperature' command.
        /// </summary>
        /// <param name="tecIdx">The tec index.</param>
        protected void ParseGetTempResponse(String resp, int tecIdx)
        {
            float f1 = 0.0f;

            lock (tecs.SyncRoot)
            {
                if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
                    tecs[tecIdx].SetPoint = f1;
            }
        }

        /// <summary>
        /// Parses the temp data response to a 'ReadTecnTemperature' command.
        /// </summary>
        /// <param name="tecIdx">The tec index.</param>
        protected void ParseReadTempResponse(String resp, int tecIdx)
        {
            float f1;

            lock (tecs.SyncRoot)
            {
                if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
                    tecs[tecIdx].Temp = f1;
            }
        }

        /// <summary>
        /// Parses the tec running data response to a 'ReadTecnRunning' command.
        /// </summary>
        /// <param name="tecIdx">The tec index.</param>
        protected void ParseReadTecRunningResponse(String resp, int tecIdx)
        {
            int f1;

            lock (tecs.SyncRoot)
            {
                if (int.TryParse(resp, out f1))
                    tecs[tecIdx].Running = f1 == 1;
            }
        }

        /// <summary>
        /// Parses the read nevolt response.
        /// </summary>
        /// <param name="tecIdx">The tec index.</param>
        private void ParseReadNeVoltResponse(String resp)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
            {
                NeVolt = f1;
            }
        }

        /// <summary>
        /// Parse laser data to tec drive.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        /// <param name="tecIdx">TEC index</param>
        protected void ParseReadDriveResponse(String resp, int tecIdx)
        {
            float f1;

            lock (tecs.SyncRoot)
            {
                if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
                    tecs[tecIdx].Drive = f1;
            }
        }

        /// <summary>
        /// Parse laser data to tec max drive.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        /// <param name="tecIdx">TEC index</param>
        private void ParseGetMaxDriveResponse(String resp, int tecIdx)
        {
            float f1;

            lock (tecs.SyncRoot)
            {
                if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
                    tecs[tecIdx].MaxDrive = f1;
            }
        }

        /// <summary>
        /// Parse laser data to max laser current.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        private void ParseGetMaxLaserCurrentResponse(String resp)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
                MaxLaserCurrent = f1;
        }

        /// <summary>
        /// Parse laser data to max laser power.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        private void ParseGetMaxLaserPowerResponse(String resp)
        {
            float f1;

            if (float.TryParse(resp, NumberStyles.Float, new CultureInfo("en-US"), out f1))
                MaxLaserPower = f1;
        }

        /// <summary>
        /// Parse laser data to laser pid.
        /// </summary>
        /// <param name="resp">Response from laser.</param>
        private void ParseGetLaserPidResponse(String resp)
        {
            if (resp.Split(new char[] { ' ' }).Length == 2)
            {
                String[] sa = resp.Split(new char[] { ' ' });
                lock (valueLock)
                {
                    float f1;
                    ParseFloatResponse(sa[0], out f1, 0.0f);
                    ldKp = f1;
                    ParseFloatResponse(sa[1], out f1, 0.0f);
                    ldKi = f1;
                }
            }
        }

        protected abstract void ParseReadAllSetPointsResponse(String resp);
        protected abstract void ParseReadAllResponse(String resp);

                    
        #endregion

        #region Response processing methods
        /// <summary>
        /// Processes the response. Updates laser information with data received from laser. When done an event is raised
        /// so that those interested is notified that the laser has been updated.
        /// </summary>
        /// <param name="resp">The resp.</param>
        protected virtual void ProcessResponse(CommandResponse resp)
        {
            // Just log syntax error responses, don't process further.
            if (resp.Resp.StartsWith("Syntax error:"))
            {
                //Raise event for listners to be aware that the laser has been updated
                if (LaserUpdatedEvent != null)
                    LaserUpdatedEvent(null, new LaserUpdatedEvent(resp));
                return;
            }

            switch (resp.CommandId)
            {
                case CommandId.ReadAll:
                    ParseReadAllResponse(resp.Resp);
                    break;

                case CommandId.ReadAllSetPoints:
                    ParseReadAllSetPointsResponse(resp.Resp);
                    break;

                case CommandId.GetCmdName:
                    ParseGetCommandNameResponse(resp.Resp);
                    break;

                case CommandId.GetOperatingHours:
                    GetOperatingHoursResponse(resp.Resp);
                    break;
    
                case CommandId.LaserDiodeEnable:
                    LaserEnabled = true;
                    break;

                case CommandId.LaserDisable:
                    LaserEnabled = false; 
                    break;

                case CommandId.GetAutostartState:
                    AutoStart = resp.Resp.Contains("1"); 
                    break;

                case CommandId.GetAutostartWarmUpCurrent:
                    ParseGetWarmUpCurrentResponse(resp.Resp);
                    break;

                case CommandId.GetWarmUpDelay:
                    ParseGetWarmUpTimeResponse(resp.Resp);
                    break;

                case CommandId.GetLaserCurrent:			// Laser current set-point
                    ParseGetLaserCurrentResponse(resp.Resp);
                    break;

                case CommandId.ReadLaserCurrent:		// Laser current value
                    ParseReadLaserCurrentResponse(resp.Resp);
                    break;

                case CommandId.GetPidForTec:
                case CommandId.GetPidForTec1:
                    ParseGetPidResponse(resp, 0);
                    break;

                case CommandId.GetPidForTec2:
                    ParseGetPidResponse(resp, 1);
                    break;

                case CommandId.GetPidForTec3:
                    ParseGetPidResponse(resp, 2);
                    break;

                case CommandId.GetPidForTec4:
                    ParseGetPidResponse(resp, 3);
                    break;

                case CommandId.GetTecTemperature:
                case CommandId.GetTec1Temperature:
                    ParseGetTempResponse(resp.Resp, 0);
                    break;

                case CommandId.GetTec2Temperature:
                    ParseGetTempResponse(resp.Resp, 1);
                    break;

                case CommandId.GetTec3Temperature:
                    ParseGetTempResponse(resp.Resp, 2);
                    break;

                case CommandId.GetTec4Temperature:
                    ParseGetTempResponse(resp.Resp, 3);
                    break;

                case CommandId.ReadTecTemperature:
                case CommandId.ReadTec1Temperature:
                    ParseReadTempResponse(resp.Resp, 0);
                    break;

                case CommandId.ReadTec2Temperature:
                    ParseReadTempResponse(resp.Resp, 1);
                    break;

                case CommandId.ReadTec3Temperature:
                    ParseReadTempResponse(resp.Resp, 2);
                    break;

                case CommandId.ReadTec4Temperature:
                    ParseReadTempResponse(resp.Resp, 3);
                    break;

                case CommandId.GetTecEnabledState:
                case CommandId.GetTec1EnabledState:
                    lock (tecs.SyncRoot)
                    {
                        tecs[0].Enabled = resp.Resp.Contains("1");
                    }
                    break;

                case CommandId.GetTec2EnabledState:
                    lock (tecs.SyncRoot)
                    {
                        tecs[1].Enabled = resp.Resp.Contains("1");
                    }
                    break;

                case CommandId.GetTec3EnabledState:
                    lock (tecs.SyncRoot)
                    {
                        tecs[2].Enabled = resp.Resp.Contains("1");
                    }
                    break;

                case CommandId.GetTec4EnabledState:
                    lock (tecs.SyncRoot)
                    {
                        tecs[3].Enabled = resp.Resp.Contains("1");
                    }
                    break;

                case CommandId.ReadTecDrive:
                case CommandId.ReadTec1Drive:
                    ParseReadDriveResponse(resp.Resp, 0);
                    break;

                case CommandId.ReadTec2Drive:
                    ParseReadDriveResponse(resp.Resp, 1);
                    break;

                case CommandId.ReadTec3Drive:
                    ParseReadDriveResponse(resp.Resp, 2);
                    break;

                case CommandId.ReadTec4Drive:
                    ParseReadDriveResponse(resp.Resp, 3);
                    break;
                    
                case CommandId.GetTecMaxDrive:
                case CommandId.GetTec1MaxDrive:
                    ParseGetMaxDriveResponse(resp.Resp, 0);
                    break;

                case CommandId.GetTec2MaxDrive:
                    ParseGetMaxDriveResponse(resp.Resp, 1);
                    break;

                case CommandId.GetTec3MaxDrive:
                    ParseGetMaxDriveResponse(resp.Resp, 2);
                    break;

                case CommandId.GetTec4MaxDrive:
                    ParseGetMaxDriveResponse(resp.Resp, 3);
                    break;

                case CommandId.GetTecDeltaTemp:
                    ParseGetDeltaTempResponse(resp.Resp);
                    break;

                case CommandId.GetLaserPower:
                    ParseGetLaserPowerResponse(resp.Resp);
                    break;

                case CommandId.GetCalibrateLdCurr:
                    LdCalibrated = resp.Resp.Equals("1");
                    break;

                case CommandId.GetLaserPid:
                    ParseGetLaserPidResponse(resp.Resp);
                    break;

                case CommandId.GetMaxLaserCurrent:
                    ParseGetMaxLaserCurrentResponse(resp.Resp);
                    break;

                case CommandId.GetMaxLaserPower:
                    ParseGetMaxLaserPowerResponse(resp.Resp);
                    break;

                case CommandId.GetSerialNumber:
                    int sn; 
                    int.TryParse(resp.Resp, out sn);
                    SerialNumber = sn;
                    break;

                case CommandId.GetFirmwareVersion:
                    FirmwareVersion = resp.Resp;
                    break;

                case CommandId.GetFirmwareDate:
                    FirmwareDate = resp.Resp;
                    break;

                case CommandId.GetFirmwareBuildDate:
                    FirmwareBuildDate = resp.Resp;
                    break;

                case CommandId.ReadNeVolt:
                    ParseReadNeVoltResponse(resp.Resp); 
                    break;

                case CommandId.GetLaserModPower:
                    ParseFloatResponse(resp.Resp, out laserModPower, 0.2f);
                    break;

                case CommandId.GetAnalogLowImpedanceState:
                    if (resp.Resp.Contains("1"))
                    {
                        Analog50Ohm = true;
                        Analog1kOhm = false;
                    }
                    else
                    {
                        Analog50Ohm = false;
                        Analog1kOhm = true;
                    }
                    break;
            }

            //Raise event for listners to be aware that the laser has been updated
            if (LaserUpdatedEvent != null)
                LaserUpdatedEvent(null, new LaserUpdatedEvent(resp));
        }
        #endregion

        /// <summary>
        /// Causes all the set points to be requested from the laser driver.
        /// </summary>
        public void GetSetPoints()
        {
            GetValues = true;
        }

        /// <summary>
        /// Opens a script file and sends the contained commands to laser driver.
        /// </summary>
        /// <param name="commandFilePath">Path to command file.</param>
        /// <param name="message">Eventual error message if</param>
        /// <returns>true if successful else false</returns>
        public bool LoadCommand(string commandFilePath, out string message)
        {
            List<string> script = new List<string>();
            using (StreamReader reader = new StreamReader(commandFilePath))
            {
                // List of RS232 command strings.
                int lineNo = 0;
                bool scriptOk = true;
                while (!reader.EndOfStream)
                {
                    lineNo++;

                    // Read next line from script file.
                    string line = reader.ReadLine();

                    // Trim leading and trailing white-space characters.
                    string cmd = line.Trim();

                    // Skip empty lines and comment lines starting with '%'.
                    if (cmd.Length == 0 || cmd[0] == '%')
                        continue;


                    // Syntax checking.
                    int idx = cmd.IndexOf(' ');
                    string key;
                    string args;

                    if (idx >= 0)
                    {
                        key = cmd.Substring(0, idx);
                        args = cmd.Substring(idx + 1);
                    }
                    else
                    {
                        key = cmd;
                        args = "";
                    }

                    CommandId commandId;
                    if ((commandId = FindKey(key)) != CommandId.UnknownCommand)
                    {
                        script.Add(cmd);
                    }
                    else
                    {
                        message = string.Format("Unknown command.\n\nIn file: {0}\nAt line {1:D}\n\n{2}", Path.GetFileName(commandFilePath), lineNo, line);
                        return false;
                    }
                }

                // Send script commands to laser driver.
                if (scriptOk)
                {
                    foreach (string command in script)
                    {
                        SendCommand(command);
                    }

                    // Show filename.
                    cmdName = Path.GetFileNameWithoutExtension(commandFilePath);

                    // Send command name to laser driver.
                    SendCommand(CommandId.SetCmdName, Path.GetFileName(commandFilePath));

                    // Ensure everything is refreshed in the GUI.
                    GetValues = true;
                }

            }
            message = String.Empty;
            return true;
        }

        /// <summary>
        /// Writes common laser functionality to file.
        /// </summary>
        /// <param name="commandFilePath">Path to command file.</param>
        /// <param name="message">Eventual error message</param>
        /// <returns>true if successful else false</returns>
        public abstract bool SaveCommand(string commandFilePath, out string message);
      
        /// <summary>
        /// Finds the key in command dictionary.
        /// </summary>
        /// <param name="str">Laser command</param>
        /// <returns>CommandId for the command.</returns>
        public CommandId FindKey(string str)
        {
            CommandId commandId = CommandId.UnknownCommand;

            foreach (KeyValuePair<CommandId, CommandInfo> kvp in cmdDict)
            {
                if (kvp.Value.Name.Equals(str))
                {
                    commandId = kvp.Key;
                    break;
                }
            }

            return commandId;
        }

        /// <summary>
        /// Build string to send to laser.
        /// </summary>
        /// <param name="cmd">The command to send to the laser.</param>
        /// <param name="args">Arguments to send with the command to the laser.</param>
        /// <returns>True if authication is needed for this command, else false</returns>
        public string AsString(CommandId commandId, string args)
        {
            CommandInfo cmdInfo;
            string cs;

            if (cmdDict.TryGetValue(commandId, out cmdInfo))
            {
                cs = cmdInfo.Name;
                if (args.Length > 0)
                    cs += " " + args;
            }
            else
                cs = "";

            return cs;
        }

        /// <summary>
        /// Adds common laser commands to command dictionary.
        /// </summary>
		protected virtual void BuildDictionary()
		{
			cmdDict.Add(CommandId.AreYouThere, new CommandInfo("?", "Are you there. Returns OK", false, false));
			cmdDict.Add(CommandId.LaserDisable, new CommandInfo("ld", "Laser OFF.", true, true));
			cmdDict.Add(CommandId.LaserDisable2, new CommandInfo("l0", "Laser OFF.", false, false));
            cmdDict.Add(CommandId.LaserDisable3, new CommandInfo("@cob0", "Laser OFF.", true, true));
            cmdDict.Add(CommandId.LaserEnable, new CommandInfo("le", "Laser enable.", true, true));
			cmdDict.Add(CommandId.LaserDiodeEnable, new CommandInfo("lde", "Laser ON", true, true));
			cmdDict.Add(CommandId.LaserDiodeEnable2, new CommandInfo("l1", "Laser ON", false, false));
			cmdDict.Add(CommandId.DisableAutostart, new CommandInfo("da", "Disable autostart.", true, true));
			cmdDict.Add(CommandId.EnableAutostart, new CommandInfo("ea", "Enable autostart.", true, true));
			cmdDict.Add(CommandId.ForcedAutostart, new CommandInfo("@cob1", "Force laser into autostart mode.", false, false));
			cmdDict.Add(CommandId.SetAutostartState, new CommandInfo("@cobas", "Enable/disable autostart.", false, false));
			cmdDict.Add(CommandId.GetAutostartState2, new CommandInfo("@cobas?", "Get autostart enable state.", false, false));
			cmdDict.Add(CommandId.GetAutostartState, new CommandInfo("gas?", "Get autostart enable state.", true, true));
			
			cmdDict.Add(CommandId.GetKeySwitchState, new CommandInfo("gkss?", "Get key switch state.", true, true));
            cmdDict.Add(CommandId.GetKeySwitchState2, new CommandInfo("@cobasks?", "Get key switch state.", true, true));
            cmdDict.Add(CommandId.DisableKeySwitch, new CommandInfo("dks", "Disable key switch.", true, true));
			cmdDict.Add(CommandId.EnableKeySwitch, new CommandInfo("eks", "Enable key switch.", true, true));
			cmdDict.Add(CommandId.SetKeySwitchEnableState, new CommandInfo("@cobasky", "Enable/disable key switch.", false, false));
			cmdDict.Add(CommandId.GetKeySwitchEnableState, new CommandInfo("gkses?", "Get key switch enable state.", true, true));
			cmdDict.Add(CommandId.GetKeySwitchEnableState2, new CommandInfo("@cobasky?", "Get key switch enable state.", false, false));
			cmdDict.Add(CommandId.SetAutostartMode, new CommandInfo("sam", "Set autostart mode.", true, true));
			cmdDict.Add(CommandId.GetAutostartMode, new CommandInfo("gam?", "Get autostart mode.", true, true));
			cmdDict.Add(CommandId.SetAutostartPower, new CommandInfo("sap", "Set autostart power.", true, true));
			cmdDict.Add(CommandId.GetAutostartPower, new CommandInfo("gap?", "Get autostart power.", true, true));
			cmdDict.Add(CommandId.GetOperatingMode, new CommandInfo("gom?", "Get operating mode.", true, true));
            cmdDict.Add(CommandId.GetAutostartProgramState, new CommandInfo("@cobast?", "Get autostart program state.", true, true));

			cmdDict.Add(CommandId.ClearFault, new CommandInfo("cf", "Clear fault.", false, false));
			cmdDict.Add(CommandId.EnterCcMode, new CommandInfo("ecc", "Enter constant current mode.", true, true));
            cmdDict.Add(CommandId.EnterCcMode2, new CommandInfo("ci", "Enter constant current mode.", false, true));
            cmdDict.Add(CommandId.EnterCpMode, new CommandInfo("ecp", "Enter constant power mode.", true, true));
            cmdDict.Add(CommandId.EnterCpMode2, new CommandInfo("cp", "Enter constant power mode.", false, true));
            cmdDict.Add(CommandId.GetOperatingFault, new CommandInfo("gof?", "Get operating fault.", true, true));
			cmdDict.Add(CommandId.GetErrorStatus, new CommandInfo("f?", "Get error status.", false, false));
			cmdDict.Add(CommandId.GetOperatingHours, new CommandInfo("goh?", "Get operating hours.", true, true));
			cmdDict.Add(CommandId.GetOperatingHours2, new CommandInfo("hrs?", "Get operating hours.", false, false));
			cmdDict.Add(CommandId.GetOperatingSeconds, new CommandInfo("gos?", "Get operating seconds.", true, true));
			cmdDict.Add(CommandId.ClearOperatingHours, new CommandInfo("coh", "Clear operating hours.", true, true));

			cmdDict.Add(CommandId.SetLaserCurrent, new CommandInfo("slc", "Set laser current (A).", false, false));
			cmdDict.Add(CommandId.GetLaserCurrent, new CommandInfo("glc?", "Get laser current.", true, true));
			cmdDict.Add(CommandId.ReadLaserCurrent, new CommandInfo("rlc?", "Read laser current.", true, true));
			cmdDict.Add(CommandId.ReadLaserCurrent2, new CommandInfo("i?", "Read laser current.", false, false));
			cmdDict.Add(CommandId.ReadInterlockState, new CommandInfo("rils?", "Read interlock state.", true, true));
			cmdDict.Add(CommandId.ReadInterlockStateInv, new CommandInfo("ilk?", "Read interlock state (inverted).", false, false));
			cmdDict.Add(CommandId.GetLaserEnableState, new CommandInfo("gles?", "Get laser ON/OFF state.", true, true));
			cmdDict.Add(CommandId.GetLaserEnableState2, new CommandInfo("l?", "Get laser ON/OFF state.", false, false));
			cmdDict.Add(CommandId.GetLedStatus, new CommandInfo("gls?", "Get LED status as 4 integers.", true, true));
			cmdDict.Add(CommandId.GetLedStatus2, new CommandInfo("leds?", "Get LED status as integer.", false, false));
			cmdDict.Add(CommandId.SetMaxLaserCurrent, new CommandInfo("smlc", "Set max laser current.", true, true));
			cmdDict.Add(CommandId.GetMaxLaserCurrent, new CommandInfo("gmlc?", "Get max laser current.", true, true));
			cmdDict.Add(CommandId.SetMaxLaserPower, new CommandInfo("smlp", "Set max laser power.", true, true));
			cmdDict.Add(CommandId.GetMaxLaserPower, new CommandInfo("gmlp?", "Get max laser power.", true, true));

			cmdDict.Add(CommandId.SetLaserPower, new CommandInfo("slp", "Set laser power (mW).", true, true));
			cmdDict.Add(CommandId.GetLaserPower, new CommandInfo("glp?", "Get laser power (mW).", true, true));
			cmdDict.Add(CommandId.ReadLaserPower, new CommandInfo("rlp?", "Read laser power (mW).", true, true));

			cmdDict.Add(CommandId.SetLaserPowerW, new CommandInfo("p", "Set laser power (W).", false, false));
			cmdDict.Add(CommandId.GetLaserPowerW, new CommandInfo("p?", "Get laser power (W).", false, false));
			cmdDict.Add(CommandId.ReadLaserPowerW, new CommandInfo("pa?", "Read laser power (W).", false, false));
            cmdDict.Add(CommandId.SetLaserPowerW2, new CommandInfo("@cobasp", "Set laser power (W).", true, true));
            cmdDict.Add(CommandId.GetLaserPowerW2, new CommandInfo("ps?", "Get laser power (W).", true, true));

			cmdDict.Add(CommandId.GetSerialNumber, new CommandInfo("gsn?", "Get serial number", true, true));
			cmdDict.Add(CommandId.GetSerialNumber2, new CommandInfo("sn?", "Get serial number", false, false));

			cmdDict.Add(CommandId.ReadPhotoDiodeVoltage, new CommandInfo("rpdv?", "Read photo diode voltage.", true, true));

			cmdDict.Add(CommandId.GetFirmwareVersion, new CommandInfo("gfv?", "Get firmware version.", true, true));
			cmdDict.Add(CommandId.GetFirmwareDate, new CommandInfo("gfd?", "Get firmware date.", true, true));
			cmdDict.Add(CommandId.GetFirmwareBuildDate, new CommandInfo("gfbd?", "Get firmware build date.", true, true));
            cmdDict.Add(CommandId.GetFirmwareVersion2, new CommandInfo("ver?", "Get firmware version.", true, true));

			cmdDict.Add(CommandId.SetCalPointWattVolt, new CommandInfo("scpwv", "Set calibration point W/V for light loop.", true, true));
			cmdDict.Add(CommandId.GetCalPointWattVolt, new CommandInfo("gcpwv?", "Get calibration point W/V for light loop.", true, true));
			cmdDict.Add(CommandId.CalibrateWattsPerVolt, new CommandInfo("cwpv", "Calibrate Watts per Volt.", true, true));

			cmdDict.Add(CommandId.SetCalPointAmpsVolt, new CommandInfo("scpav", "Set calibration point A/V.", true, true));
			cmdDict.Add(CommandId.GetCalPointAmpsVolt, new CommandInfo("gcpav?", "Get calibration point A/V.", true, true));
			cmdDict.Add(CommandId.CalibrateAmpsPerVolt, new CommandInfo("capv", "Calibrate Amps per Volt.", true, true));

			cmdDict.Add(CommandId.SetTecTimeout, new CommandInfo("stto", "Set TEC regulation timeout.", true, true));
			cmdDict.Add(CommandId.GetTecTimeout, new CommandInfo("gtto?", "Get TEC regulation timeout.", true, true));
			cmdDict.Add(CommandId.GetTecDeltaTemp, new CommandInfo("gtdt?", "Get TEC delta temperature.", true, true));
			cmdDict.Add(CommandId.SetTecDeltaTemp, new CommandInfo("stdt", "Set TEC delta temperature.", true, true));

			cmdDict.Add(CommandId.ReadAll, new CommandInfo("ra?", "Read all.", true, true));
			cmdDict.Add(CommandId.ReadAllSetPoints, new CommandInfo("rasp?", "Read all set points.", true, true));

			cmdDict.Add(CommandId.SetIpLogCurrent, new CommandInfo("siplc", "Set I-P Logging Current.", true, true));

			cmdDict.Add(CommandId.Abort, new CommandInfo("abort", "Abort autostart sequence.", true, true));
			cmdDict.Add(CommandId.Restart, new CommandInfo("restart", "Restart autostart sequence.", true, true));

			cmdDict.Add(CommandId.SetLaserPid, new CommandInfo("slpid", "Set Laser PID Constants", true, true));
			cmdDict.Add(CommandId.GetLaserPid, new CommandInfo("glpid?", "Get Laser PID Constants", true, true));

			cmdDict.Add(CommandId.CalibrateLdCurr, new CommandInfo("cldct", "Calibrate Laser Diode Current.", true, true));
			cmdDict.Add(CommandId.EnableCalibrateLdCurr, new CommandInfo("ecldct", "Enable Calibrate Laser Diode Current.", true, true));
			cmdDict.Add(CommandId.GetCalibrateLdCurr, new CommandInfo("gcldct?", "Get Calibrate Laser Diode Current State.", true, true));

			cmdDict.Add(CommandId.SetCmdName, new CommandInfo("scn", "Set Command Name.", true, true));
			cmdDict.Add(CommandId.GetCmdName, new CommandInfo("gcn?", "Get Command Name.", true, true));

			cmdDict.Add(CommandId.SpecialCommand, new CommandInfo(":", "Special Command.", true, true));
		}

        /// <summary>
        /// Looks up if the command needs authentication to be sent to the laser.
        /// </summary>
        /// <param name="cmd">The command to send to the laser.</param>
        /// <returns>True if authication is needed for this command, else false</returns>
        public bool NeedsAuth(CommandId commandId)
        {
            CommandInfo info;
            if (cmdDict.TryGetValue(commandId, out info))
                return info.NeedsAuth;
            else
                return false;
        }

        /// <summary>
        /// Looks up if the command needs authentication to be sent to the laser.
        /// </summary>
        /// <param name="cmd">The command to send to the laser.</param>
        /// <returns>True if authication is needed for this command, else false</returns>
        public bool NeedsAuth(string cmd)
        {
            string key;
            string args;
            ParseToKeyAndArgs(cmd, out key, out args);
            CommandId commandId = FindKey(key);
            return NeedsAuth(commandId);
        }

        /// <summary>
        /// Gets all valid commands for the laser.
        /// </summary>
        /// <returns>A ICollection with alla valid laser commands.</returns>
        public ICollection<CommandInfo> GetCommandList()
        {
            return cmdDict.Values;
        }
        
        #region Event handlers

        /// <summary>
        /// Handles the DataReceived event of the serialPort object.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="SerialEvent"/> instance containing the event data.</param>
        private void SerialPort_DataReceived(object sender, SerialEvent e)
        {
            ProcessResponse(e.Response);
        }
        
        /// <summary>
        /// Handles the Tick event of the timerComm control. Sends poll messages
        /// to the Cobolt Laser Driver.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        protected virtual void timerComm_Tick(object sender, EventArgs e)
        {
            if (timerComm.Enabled == false)
                return;

            if (GetValues == true && serialHandler.Resends == 0)
            {
                GetValues = false;

                SendCommand(CommandId.AreYouThere, "");
                SendCommand(CommandId.ReadAllSetPoints, "");
                SendCommand(CommandId.GetSerialNumber, "");
                SendCommand(CommandId.GetCmdName, "");
                SendCommand(CommandId.GetFirmwareVersion, "");
                SendCommand(CommandId.GetFirmwareDate, "");
                SendCommand(CommandId.GetFirmwareBuildDate, "");
                SendCommand(CommandId.GetOperatingHours, "");
                SendCommand(CommandId.GetMaxLaserCurrent, "");
                SendCommand(CommandId.GetLaserPid, "");
                SendCommand(CommandId.GetMaxLaserPower, "");

                //wait 20 sec until sending next GetOperatingHours command
                timeOfNextGoh = DateTime.Now.AddSeconds(20);
            }

            SendCommand(CommandId.ReadAll, "");

            if (DateTime.Now >= timeOfNextGoh)
            {
                timeOfNextGoh = DateTime.Now.AddSeconds(20);
                SendCommand(CommandId.GetOperatingHours, "");
            }
        }
        #endregion
    }
}
