﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO.Ports;
using System.IO;
using System.Threading;
using System.Windows.Forms;

namespace CoboltMonitor.Laser
{
    /// <summary>
    /// Class defining a SerialEvent on event SerialData_Received
    /// </summary>
    public class SerialEvent : EventArgs
    {
        private CommandResponse reponse;

        public SerialEvent(CommandResponse resp)
        {
            reponse = resp;
        }

        public CommandResponse Response
        {
            get { return reponse; }
        }
    }

    /// <summary>
    /// SerialHandler takes care of communication with the serial port. 
    /// Commands to be sent is added to a queue and is sent when possible.
    /// Data received from the serial port is sent to listerners on SerialData_Received event for handling.
    /// </summary>
    public class SerialHandler
    {
        public event EventHandler<SerialEvent> ResponseReceived;
        private System.IO.Ports.SerialPort serialPort;
        private object _sendQLock;
        private object _serialLock;
        Queue<CommandResponse> _sendQ;
        private int _cmdRespTimeout = 0;
        private int _resends = 0;
        private string _response;
        private System.Timers.Timer timerComm;
        private int _ticksSinceLastResponse = 0;
        private bool timeout = false;
        private string synchronousResponse = "";

        public SerialHandler()
        {
            _sendQ = new Queue<CommandResponse>();
            _sendQLock = new object();

            serialPort = new SerialPort();
            serialPort.DataReceived += new System.IO.Ports.SerialDataReceivedEventHandler(SerialPort_DataReceived);
            _serialLock = new object();

            timerComm = new System.Timers.Timer();
            timerComm.Interval = 200;
            timerComm.Elapsed += new System.Timers.ElapsedEventHandler(timerComm_Tick);
        }

        /// <summary>
        /// Gets whether serial port is open or not.
        /// </summary>
        public bool IsSerialPortOpen
        {
            get
            {
                lock (_serialLock)
                {
                    return serialPort.IsOpen;
                }
            }
        }

        /// <summary>
        /// Gets the number of resends made of last command.
        /// </summary>
        public int Resends
        {
            get
            {
                return _resends;
            }
        }

        /// <summary>
        /// Closes connection on the serial port.
        /// </summary>
        public bool CloseSerialPort()
        {
            try
            {
                timerComm.Enabled = false;
                lock (_sendQLock)
                {
                    _sendQ.Clear();
                }
                Thread.Sleep(500);
                lock (_serialLock)
                {
                    serialPort.Close();
                }
            }
            catch
            {
                return false;
            }
            finally
            {
                _cmdRespTimeout = 0;
            }
            return true;
        }

        /// <summary>
        /// Opens connection on the serial port.
        /// </summary>
        public bool OpenSerialPort(string serialPortName, int serialBaudRate)
        {
            try
            {
                if (serialPort.IsOpen)
                    return true;

                _response = "";
                lock (_serialLock)
                {
                    serialPort.NewLine = "\r";
                    serialPort.PortName = serialPortName;
                    serialPort.BaudRate = serialBaudRate;
                    serialPort.Open();
                    serialPort.DiscardInBuffer();
                }
                _cmdRespTimeout = 0;
                _resends = 0;
                timerComm.Enabled = true;
            }
            catch (UnauthorizedAccessException)
            {
                MessageBox.Show("Not able to open selected port, maybe used by another program?");
                return false;
            }
            catch
            {
                return false;
            }
            return true;
        }

        /// <summary>
        /// Sends a string to the serial port.
        /// </summary>
        /// <param name="str">The string to be sent.</param>
        private void SendString(string str)
        {
            try
            {
                _ticksSinceLastResponse = 0;
                _cmdRespTimeout = 5000 / (int)timerComm.Interval;
                lock (_serialLock)
                {
                    serialPort.DiscardInBuffer();
                    serialPort.WriteLine(str);
                }
            }
            catch
            {
                CloseSerialPort();
            }
        }

        /// <summary>
        /// Sends a command to the laser driver.
        /// </summary>
        /// <param name="command">The command to be sent.</param>
        public void SendCommand(CommandResponse cmd)
        {
            lock (_serialLock)
            {
                if (!serialPort.IsOpen)
                    return;
            }

            lock(_sendQLock)
            {
                _sendQ.Enqueue(cmd);
            }
            if (_sendQ.Count == 1)
                SendString(cmd.Sendstring);
        }

        #region Serial Port Events
        /// <summary>
        /// Handles the DataReceived event of the serialPort control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.IO.Ports.SerialDataReceivedEventArgs"/> instance containing the event data.</param>
        private void SerialPort_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            // Far reaching code here. Trying to get rid of an annoying exception
            // occuring sometimes when disconnecting.
            string response = String.Empty;
            lock (_serialLock)
            {
                if (!serialPort.IsOpen)
                    return;

                response = serialPort.ReadExisting();
            }

            int i = response.IndexOfAny(new char[] { '\r', '\n' });

            _response += response;

            //Process all responses in the receiced data
            while ((i = _response.IndexOf("\r\n")) >= 0)
            {
                string resp = _response.Substring(0, i);

                if (_response.Length > (resp.Length + 2))
                    _response = _response.Substring(i + 2);
                else
                    _response = "";

                resp = resp.Trim();

                //Raise serialevent if data is expected
                if (response.Length > 0 && _sendQ.Count > 0)
                {
                    _cmdRespTimeout = 0;
                    _resends = 0;
                    lock (_sendQLock)
                    {
                        _sendQ.Peek().Resp = resp;

                        if (ResponseReceived != null)
                            ResponseReceived(null, new SerialEvent(_sendQ.Dequeue()));

                        if (_sendQ.Count > 0)
                        {
                            SendString(_sendQ.Peek().Sendstring);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Handles the DataReceived event of the serialPort control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.IO.Ports.SerialDataReceivedEventArgs"/> instance containing the event data.</param>
        private void Synchronous_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            // Far reaching code here. Trying to get rid of an annoying exception
            // occuring sometimes when disconnecting.
            if (!serialPort.IsOpen)
                return;

            synchronousResponse = serialPort.ReadExisting();
        }
        #endregion

        #region Timer Control Events
        /// <summary>
        /// Handles the Tick event of the timerComm control. Sends a poll message
        /// to the Cobolt Laser Driver if _sendQ is empty.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void timerComm_Tick(object sender, EventArgs e)
        {
            if (timerComm.Enabled == false)
                return;
            _ticksSinceLastResponse++;

            // Check for response timeout. If timeout, resend last command if still in send queue.
            if (_cmdRespTimeout > 0 && _ticksSinceLastResponse > _cmdRespTimeout)
            {
                lock (_sendQLock)
                {
                    if (_sendQ.Count > 0)
                    {
                        string s;
                        SendString(s = _sendQ.Peek().Sendstring);
                        _resends++;
                    }
                }
            }
        }

        private void SynchronSendReceiveTimeout(object sender, EventArgs e)
        {
            timeout = true;
        }
        #endregion

        /// <summary>
        /// Sends a command to laser and waits for response.
        /// </summary>
        /// <param name="cmd">command to be sent</param>
        /// <returns>Response from laser.</returns>
        public string SendCommandAndWaitForResponse(string cmd)
        {
            timeout = false;
            System.Timers.Timer timerComm;
            timerComm = new System.Timers.Timer();
            timerComm.Interval = 5000;
            timerComm.Elapsed += new System.Timers.ElapsedEventHandler(SynchronSendReceiveTimeout);

            //Disable serialPort_DataReceived so that it doesn't interfere with call
            serialPort.DataReceived -= new System.IO.Ports.SerialDataReceivedEventHandler(SerialPort_DataReceived);
            serialPort.DataReceived += new System.IO.Ports.SerialDataReceivedEventHandler(Synchronous_DataReceived);

            synchronousResponse = String.Empty;

            try
            {
                lock (_serialLock)
                {
                    serialPort.DiscardInBuffer();
                    serialPort.WriteLine(cmd);
                }
                timerComm.Start();
                while (synchronousResponse == String.Empty && !timeout)
                    Thread.Sleep(10);
            }
            finally
            {
                //Enable serialPort_DataReceived again
                serialPort.DataReceived -= new System.IO.Ports.SerialDataReceivedEventHandler(Synchronous_DataReceived);
                serialPort.DataReceived += new System.IO.Ports.SerialDataReceivedEventHandler(SerialPort_DataReceived);
                timerComm.Stop();
            }

            return synchronousResponse;
        }
    }
}
