using System;
using System.Collections.Generic;
using System.Text;

namespace CoboltMonitor.Laser
{
	public class Tec
	{
		#region Instance Variables.
		// Temperatures.
        float _setpoint = 25.0f;		// setpoint temperature
        float _temp = 0.0f;			// actual temperature
        float _delta = 0.02f;			// acceptable difference between temperature and setpoint
		float _drive;
		float _maxdrive;		// max allowed drive in %

		// PID constants.
        float _Kp = 20.0f;				// proportional constant
        float _Ki = 0.007f;				// integral constant
        float _Kd = 0.0f;				// differential constant

		bool _enabled;			// true if PID controller active for this TEC
        bool _running;			// Is TEC running?
        int _fault = 0;
        #endregion

		#region Properties
        /// <summary>
        /// Gets or sets if the tec is running.
        /// </summary>
        /// <value>true->running, false->not running.</value>
        public bool Running
        {
            get { return _running; }
            set { _running = value; }
        }
        /// <summary>
		/// Gets or sets the set point.
		/// </summary>
		/// <value>The set point.</value>
		public float SetPoint
		{
			get { return _setpoint; }
			set { _setpoint = value; }
		}

		/// <summary>
		/// Gets or sets the temp.
		/// </summary>
		/// <value>The temp.</value>
		public float Temp
		{
			get { return _temp; }
			set { _temp = value; }
		}

		/// <summary>
		/// Gets or sets the drive in percent.
		/// </summary>
		/// <value>The drive.</value>
		public float Drive
		{
			get { return _drive; }
			set { _drive = value; }
		}

		/// <summary>
		/// Gets or sets the max drive.
		/// </summary>
		/// <value>The max drive.</value>
		public float MaxDrive
		{
			get { return _maxdrive; }
			set { _maxdrive = value; }
		}

		/// <summary>
		/// Gets or sets the allowed delta.
		/// </summary>
		/// <value>The delta.</value>
		public float Delta
		{
			get { return _delta; }
			set { _delta = value; }
		}

		/// <summary>
		/// Gets the delta as string.
		/// </summary>
		/// <value>The delta.</value>
		public string StrDelta
		{
			get { return Delta.ToString(); }
		}

		/// <summary>
		/// Gets the diff between actual value and setpoint.
		/// </summary>
		/// <value>The diff.</value>
		public float Diff
		{
			get { return (float)(Math.Round(Math.Abs(_temp - _setpoint) * 100.0) / 100.0); }
		}

		/// <summary>
		/// Gets the diff in string format.
		/// </summary>
		/// <value>The diff.</value>
		public string StrDiff
		{
			get { return Diff.ToString(); }
		}

		/// <summary>
		/// Gets a value indicating whether this <see cref="Tec"/> is locked.
		/// </summary>
		/// <value><c>true</c> if locked; otherwise, <c>false</c>.</value>
		public bool Locked
		{
			get { return Diff <= Delta; }
		}

		/// <summary>
		/// Gets or sets the kp.
		/// </summary>
		/// <value>The kp.</value>
		public float Kp
		{
			get { return _Kp; }
			set { _Kp = value; }
		}

		/// <summary>
		/// Gets or sets the ki.
		/// </summary>
		/// <value>The ki.</value>
		public float Ki
		{
			get { return _Ki; }
			set { _Ki = value; }
		}

		/// <summary>
		/// Gets or sets the kd.
		/// </summary>
		/// <value>The kd.</value>
		public float Kd
		{
			get { return _Kd; }
			set { _Kd = value; }
		}

		/// <summary>
		/// Gets or sets a value indicating whether this <see cref="Tec"/> is enabled.
		/// </summary>
		/// <value><c>true</c> if enabled; otherwise, <c>false</c>.</value>
		public bool Enabled
		{
			get { return _enabled; }
			set { _enabled = value; }
		}

        public int Fault
        {
            get { return _fault; }
            set { _fault = value; }
        }
        #endregion

		#region Public member functions
		/// <summary>
		/// Sets the PID parameters.
		/// </summary>
		/// <param name="Kp">The kp.</param>
		/// <param name="Ki">The ki.</param>
		/// <param name="Kd">The kd.</param>
		public void PidData(float Kp, float Ki, float Kd)
		{
			_Kp = Kp;
			_Ki = Ki;
			_Kd = Kd;
		}
		#endregion
	}
}
