﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Globalization;

using CoboltMonitor.Laser;
using CoboltMonitor.Gui;
using System.Threading;
using System.Reflection;

namespace CoboltMonitor4_5.Gui
{
    public partial class MainForm : Form
    {
        #region Enumerations
        public enum OperatingMode
        {
            mode_off,
            mode_wait_tec,
            mode_wait_key,
            mode_warm_up,
            mode_final,
            mode_fault,
            mode_aborted,
            mode_count
        };

        public enum TecFaults
        {
            fault_none,
            fault_upper_rail,
            fault_lower_rail,
            fault_no_conv,
            fault_NTC_bad_conn,
            fault_NTC_short,
            fault_driver_overcurr,
            fault_driver_undervolt,
            fault_driver_overtemp,
            fault_count
        };

        public enum LaserFaults
        {
            fault_none,
            fault_no_conv_warm_up,
            fault_no_completion,
            fault_interlock,
            fault_current_clamp,
            fault_count
        };
        #endregion

        #region Delegates
        /// <summary>
        /// Helper delegate to update UI and be thread safe.
        /// </summary>
        public delegate void UpdateSerialData(Object objSender, LaserUpdatedEvent dataReceivedEventArgs);
        #endregion

        CoboltMonitor.Laser.CoboltMonitor coboltMonitor;
        LaserCommon laser;

        LogTerm _log;
        PIDMonitor _PidMon;
        IPLog2 _ipLog;
        ParLog _parLog;
        NEMon _neMon;

        Image _ledGray;
        Image _ledGreen;
        Image _ledOrange;
        Image _ledRed;

        Image _led1;
        Image _led1Off;
        Image _led2;
        Image _led2Off;
        Image _led3;
        Image _led3Off;
        Image _led4;
        Image _led4Off; 
        
        bool _led1Blinks = false;
        bool _led2Blinks = false;
        bool _led3Blinks = false;
        bool _led4Blinks = false;
        private int _blinkTimer = 0;
        private String serialPortName;
        private int serialBaudRate;

        string[] _strTecFault;
        string[] _strLaserFault;

        public String SerialPortName
        {
            get
            {
                return serialPortName;
            }
            set 
            {
                serialPortName = value;
            }
        }

        public int SerialBaudRate
        {
            get
            {
                return serialBaudRate;
            }
            set
            {
                serialBaudRate = value;
            }
        }
        public MainForm()
        {
            Thread.CurrentThread.CurrentUICulture = Thread.CurrentThread.CurrentCulture = new CultureInfo("en-US");
            InitializeComponent();
            coboltMonitor = new CoboltMonitor.Laser.CoboltMonitor();
            _ledGray = Image.FromStream(GetEmbeddedFile("CoboltMonitor4_5.Resources.LedGray.bmp"));
            _led1 = _ledGreen = Image.FromStream(GetEmbeddedFile("CoboltMonitor4_5.Resources.LedGreen.bmp"));
            _led2 = _led3 = _ledOrange = Image.FromStream(GetEmbeddedFile("CoboltMonitor4_5.Resources.LedOrange.bmp"));
            _led4 = _ledRed = Image.FromStream(GetEmbeddedFile("CoboltMonitor4_5.Resources.LedRed.bmp"));
            _led1Off = _ledGray;
            _led2Off = _ledGray;
            _led3Off = _ledGray;
            _led4Off = _ledGray;
            pbLed1.Image = _led1Off;
            pbLed2.Image = _led2Off;
            pbLed3.Image = _led3Off;
            pbLed4.Image = _led4Off;
            Disconnect();
            SerialBaudRate = Properties.Settings.Default.SerialBaudRate;
            SerialPortName = Properties.Settings.Default.SerialPortName;
            UpdateStatusBar();

            _strTecFault = new string[(int)TecFaults.fault_count];
            _strTecFault[(int)TecFaults.fault_none] = "None";
            _strTecFault[(int)TecFaults.fault_upper_rail] = "Upper rail";
            _strTecFault[(int)TecFaults.fault_lower_rail] = "Lower rail";
            _strTecFault[(int)TecFaults.fault_no_conv] = "No convergence";
            _strTecFault[(int)TecFaults.fault_NTC_bad_conn] = "NTC bad conn";
            _strTecFault[(int)TecFaults.fault_NTC_short] = "NTC shorted";
            _strTecFault[(int)TecFaults.fault_driver_overcurr] = "Over-current";
            _strTecFault[(int)TecFaults.fault_driver_undervolt] = "Under-voltage";
            _strTecFault[(int)TecFaults.fault_driver_overtemp] = "Over-temp";

            _strLaserFault = new string[(int)LaserFaults.fault_count];
            _strLaserFault[(int)LaserFaults.fault_none] = "None";
            _strLaserFault[(int)LaserFaults.fault_no_conv_warm_up] = "No convergence";
            _strLaserFault[(int)LaserFaults.fault_no_completion] = "Unstable";
            _strLaserFault[(int)LaserFaults.fault_interlock] = "Interlock open";
            _strLaserFault[(int)LaserFaults.fault_current_clamp] = "Current Clamp";

        }
        
        #region Form Events
        /// <summary>
        /// Handles the FormClosing event of the MainForm control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.Windows.Forms.FormClosingEventArgs"/> instance containing the event data.</param>
        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if(laser != null)
                laser.Disconnect();
        }

        /// <summary>
        /// Handles the Load event of the MainForm control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void MainForm_Load(object sender, EventArgs e)
        {
            this.Text += " - " + Application.ProductVersion.ToString();
        }
        #endregion

        private void UpdateGui(object sender, LaserUpdatedEvent e)
        {
            if (this.InvokeRequired == true)
            {
                UpdateSerialData updateHandler = new UpdateSerialData(UpdateGui);
                this.BeginInvoke(updateHandler, new Object[] { sender, e });
                return;
            }

            Tec[] tecs = laser.GetTecs();
            pbLed1.Image = _led1;
            switch (e.CommandResponse.CommandId)
            {
                case CommandId.GetCmdName:
                    tbLastCmdFile.Text = laser.CmdName;
                    break;

                case CommandId.GetOperatingHours:
                    toolStripStatusLabelHours.Text = laser.OperatingHours;
                    break;

                case CommandId.GetCalibrateLdCurr:
                    if (!laser.LdCalibrated)
                    {
                        toolStripStatusLabelCalibrated.Text = "Calibrate Laser Driver";
                    }

                    if (!(laser.LdCalibrated && laser.TecLdCalibrated))
                        toolStripStatusLabelCalibrated.Visible = true;
                    else
                        toolStripStatusLabelCalibrated.Visible = false;
                    break;

                case CommandId.GetTecLdThreshold:
                    if (!laser.TecLdCalibrated && laser.LdCalibrated)
                    {
                        toolStripStatusLabelCalibrated.Text = "Calibrate TEC-LD Driver";
                    }

                    if (!(laser.LdCalibrated && laser.TecLdCalibrated))
                        toolStripStatusLabelCalibrated.Visible = true;
                    else
                        toolStripStatusLabelCalibrated.Visible = false;
                    break;

                case CommandId.LaserDiodeEnable:
                    buttonLaserOn.Enabled = false;
                    buttonLaserOff.Enabled = true;
                    break;

                case CommandId.LaserDisable:
                    buttonLaserOn.Enabled = true;
                    buttonLaserOff.Enabled = false;
                    break;

                case CommandId.GetAutostartState:
                    cbAutostartEnable.Checked = laser.AutoStart;
                    buttonLaserOn.Enabled = !laser.AutoStart;
                    buttonLaserOff.Enabled = !laser.AutoStart;
                    cbCdrh.Enabled = laser.AutoStart && coboltMonitor.LoggedIn;
                    buttonAbort.Enabled = laser.AutoStart;
                    buttonRestart.Enabled = laser.AutoStart;
                    break;

                case CommandId.DisableAutostart:
                    buttonLaserOn.Enabled = true;
                    buttonLaserOff.Enabled = false;
                    break;

                case CommandId.GetAutostartWarmUpCurrent:
                    tbSetWarmUpCurrent.Text = laser.WarmUpCurrent.ToString("0.000", new CultureInfo("en-US"));
                    break;

                case CommandId.GetWarmUpDelay:
                    tbSetWarmUpTime.Text = laser.WarmUpTime.ToString("0", new CultureInfo("en-US"));
                    break;

                case CommandId.GetLaserCurrent:			// Laser current set-point
                    tbSetCurrent.Text = laser.LaserCurrentSetPoint.ToString("0.000", new CultureInfo("en-US"));
                    break;

                case CommandId.ReadLaserCurrent:		// Laser current value
                    tbCurrent.Text = laser.LaserCurrent.ToString("0.000", new CultureInfo("en-US"));
                    break;

                case CommandId.ReadTec1Temperature:
                    tbTec1Temp.Text = tecs[0].Temp.ToString("0.00", new CultureInfo("en-US"));
                    SetReadTempToolTip(tbTec1Temp, 0,tecs);
                    break;

                case CommandId.ReadTec2Temperature:
                    tbTec2Temp.Text = tecs[1].Temp.ToString("0.00", new CultureInfo("en-US"));
                    SetReadTempToolTip(tbTec2Temp, 1, tecs);
                    break;

                case CommandId.ReadTec3Temperature:
                    tbTec3Temp.Text = tecs[2].Temp.ToString("0.00", new CultureInfo("en-US"));
                    SetReadTempToolTip(tbTec3Temp, 2, tecs);
                    break;

                case CommandId.ReadTec4Temperature:
                    tbTec4Temp.Text = tecs[3].Temp.ToString("0.00", new CultureInfo("en-US"));
                    SetReadTempToolTip(tbTec4Temp, 3, tecs);
                    break;

                case CommandId.GetTec1EnabledState:
                    cbTec1Ena.Checked = tecs[0].Enabled;
                    break;

                case CommandId.GetTec2EnabledState:
                    cbTec2Ena.Checked = tecs[1].Enabled;
                    break;

                case CommandId.GetTec3EnabledState:
                    cbTec3Ena.Checked = tecs[2].Enabled;
                    break;

                case CommandId.GetTec4EnabledState:
                    cbTec4Ena.Checked = tecs[3].Enabled;
                    break;

                case CommandId.ReadTec1Drive:
                    tbTec1Drive.Text = tecs[0].Drive.ToString("0.00", new CultureInfo("en-US"));
                    break;

                case CommandId.ReadTec2Drive:
                    tbTec2Drive.Text = tecs[1].Drive.ToString("0.00", new CultureInfo("en-US"));
                    break;

                case CommandId.ReadTec3Drive:
                    tbTec3Drive.Text = tecs[2].Drive.ToString("0.00", new CultureInfo("en-US"));
                    break;

                case CommandId.ReadTec4Drive:
                    tbTec4Drive.Text = tecs[3].Drive.ToString("0.00", new CultureInfo("en-US"));
                    break;

                case CommandId.GetLaserPower:
                    tbSetPower.Text = laser.LaserPowerSetPoint.ToString("0.0", new CultureInfo("en-US"));
                    break;

                case CommandId.ReadAll:
                    tbTec1Temp.Text = tecs[0].Temp.ToString("0.00", new CultureInfo("en-US"));
                    tbTec2Temp.Text = tecs[1].Temp.ToString("0.00", new CultureInfo("en-US"));
                    tbTec3Temp.Text = tecs[2].Temp.ToString("0.00", new CultureInfo("en-US"));
                    tbTec4Temp.Text = tecs[3].Temp.ToString("0.00", new CultureInfo("en-US"));
                    SetReadTempToolTip(tbTec1Temp, 0, tecs);
                    SetReadTempToolTip(tbTec2Temp, 1, tecs);
                    SetReadTempToolTip(tbTec3Temp, 2, tecs);
                    SetReadTempToolTip(tbTec4Temp, 3, tecs);

                    // TEC drive readings.
                    tbTec1Drive.Text = tecs[0].Drive.ToString("0.00", new CultureInfo("en-US"));
                    tbTec2Drive.Text = tecs[1].Drive.ToString("0.00", new CultureInfo("en-US"));
                    tbTec3Drive.Text = tecs[2].Drive.ToString("0.00", new CultureInfo("en-US"));
                    tbTec4Drive.Text = tecs[3].Drive.ToString("0.00", new CultureInfo("en-US"));

                    // Laser Current, Power, Diode Volt, RMS %, RMS Volt, Photo Diode Vol readings.
                    tbCurrent.Text = laser.LaserCurrent.ToString("0.000", new CultureInfo("en-US"));
                    tbPower.Text = laser.LaserPower.ToString("0.0", new CultureInfo("en-US"));
                    tbLaserDiodeVoltage.Text = laser.LaserDiodeVoltage.ToString("0.000", new CultureInfo("en-US"));
                    tbRmsVoltage.Text = laser.RmsVoltage.ToString("0.000", new CultureInfo("en-US"));
                    tbRmsPercent.Text = laser.RmsPercent.ToString("0.000", new CultureInfo("en-US"));
                    tbPhotoDiodeVoltage.Text = laser.PhotoDiodeVoltage.ToString("0.000", new CultureInfo("en-US"));

                    // Keep text color upated (laser current).
                    if (!rbSetCurrent.Checked || buttonLaserOn.Enabled)
                        tbCurrent.ForeColor = Color.Yellow;
                    else if (Math.Abs(laser.LaserCurrentSetPoint - laser.LaserCurrent) <= 0.003f)
                        tbCurrent.ForeColor = Color.LightGreen;
                    else
                        tbCurrent.ForeColor = Color.HotPink;

                    // Keep text color upated (laser power).
                    if (!rbSetPower.Checked || buttonLaserOn.Enabled)
                        tbPower.ForeColor = Color.Yellow;
                    else if (Math.Abs(laser.LaserPowerSetPoint - laser.LaserPower) / laser.LaserPowerSetPoint <= 0.03f)
                        tbPower.ForeColor = Color.LightGreen;
                    else
                        tbPower.ForeColor = Color.HotPink;

                    ParseLedBits(laser.LedBits);

                    // Laser Operating mode.
                    if (laser.OperatingMode >= (int)OperatingMode.mode_off && laser.OperatingMode < (int)OperatingMode.mode_count)
                    {
                        ShowOpMode();
                        if (laser.OperatingMode == (int)OperatingMode.mode_final || laser.OperatingMode == (int)OperatingMode.mode_warm_up)
                        {
                            buttonLaserOff.Enabled = true;
                            buttonLaserOn.Enabled = false;
                        }
                        else
                        {
                            buttonLaserOff.Enabled = false;
                            buttonLaserOn.Enabled = !laser.AutoStart; // true;
                        }
                    }

                    // Constant Current or Constant Power.
                    if (laser.ConstantCurrent)
                    {
                        rbSetCurrent.Checked = true;
                    }
                    else if (laser.ConstantPower)
                    {
                        rbSetPower.Checked = true;
                    }

                    if (tecs[0].Fault >= (int)TecFaults.fault_none && tecs[0].Fault < (int)TecFaults.fault_count)
                    {
                        lbTec1Fault.Text = GetTecFault(tecs[0].Fault);
                        lbTec1Fault.ForeColor = (tecs[0].Fault == (int)TecFaults.fault_none) ? Color.Green : Color.Red;
                    }
                    else
                    {
                        lbTec1Fault.Text = "Unknown";
                        lbTec1Fault.ForeColor = Color.Red;
                    }

                    if (tecs[1].Fault >= (int)TecFaults.fault_none && tecs[1].Fault < (int)TecFaults.fault_count)
                    {
                        lbTec2Fault.Text = GetTecFault(tecs[1].Fault);
                        lbTec2Fault.ForeColor = (tecs[1].Fault == (int)TecFaults.fault_none) ? Color.Green : Color.Red;
                    }
                    else
                    {
                        lbTec2Fault.Text = "Unknown";
                        lbTec2Fault.ForeColor = Color.Red;
                    }

                    if (tecs[2].Fault >= (int)TecFaults.fault_none && tecs[2].Fault < (int)TecFaults.fault_count)
                    {
                        lbTec3Fault.Text = GetTecFault(tecs[2].Fault);
                        lbTec3Fault.ForeColor = (tecs[2].Fault == (int)TecFaults.fault_none) ? Color.Green : Color.Red;
                    }
                    else
                    {
                        lbTec3Fault.Text = "Unknown";
                        lbTec3Fault.ForeColor = Color.Red;
                    }

                    if (tecs[3].Fault >= (int)TecFaults.fault_none && tecs[3].Fault < (int)TecFaults.fault_count)
                    {
                        lbTec4Fault.Text = GetTecFault(tecs[3].Fault);
                        lbTec4Fault.ForeColor = (tecs[3].Fault == (int)TecFaults.fault_none) ? Color.Green : Color.Red;
                    }
                    else
                    {
                        lbTec4Fault.Text = "Unknown";
                        lbTec4Fault.ForeColor = Color.Red;
                    }

                    if (laser.LaserFault >= (int)LaserFaults.fault_none && laser.LaserFault < (int)LaserFaults.fault_count)
                    {
                        lbLaserFault.Text = GetLaserFault(laser.LaserFault);
                        lbLaserFault.ForeColor = (laser.LaserFault == (int)LaserFaults.fault_none || laser.LaserFault == (int)LaserFaults.fault_current_clamp) ? Color.Green : Color.Red;
                    }
                    else
                    {
                        lbLaserFault.Text = "Unknown";
                        lbLaserFault.ForeColor = Color.Red;
                    }


                    int faultsOred = tecs[0].Fault | tecs[1].Fault | tecs[2].Fault | tecs[3].Fault | laser.LaserFault;
                    if (laser.LaserFault > (int)LaserFaults.fault_none && laser.LaserFault != (int)LaserFaults.fault_current_clamp)
                        buttonClearFault.Enabled = (faultsOred != 0);

                    cbTec1Running.Checked = tecs[0].Running;
                    cbTec2Running.Checked = tecs[1].Running;
                    cbTec3Running.Checked = tecs[2].Running;
                    cbTec4Running.Checked = tecs[3].Running;

                    if ((_blinkTimer++ & 2) == 2)
                    {
                        if (_led1Blinks) pbLed1.Image = _led1;
                        if (_led2Blinks) pbLed2.Image = _led2;
                        if (_led3Blinks) pbLed3.Image = _led3;
                        if (_led4Blinks) pbLed4.Image = _led4;
                    }
                    else
                    {
                        if (_led1Blinks) pbLed1.Image = _led1Off;
                        if (_led2Blinks) pbLed2.Image = _led2Off;
                        if (_led3Blinks) pbLed3.Image = _led3Off;
                        if (_led4Blinks) pbLed4.Image = _led4Off;
                    }
                    break;

                case CommandId.ReadAllSetPoints:
                    nudTec1SetTemp.Value = (decimal)tecs[0].SetPoint;
                    nudTec2SetTemp.Value = (decimal)tecs[1].SetPoint;
                    nudTec3SetTemp.Value = (decimal)tecs[2].SetPoint;
                    nudTec4SetTemp.Value = (decimal)tecs[3].SetPoint;
                    cbTec1Ena.Checked = tecs[0].Enabled;
                    cbTec2Ena.Checked = tecs[1].Enabled;
                    cbTec3Ena.Checked = tecs[2].Enabled;
                    cbTec4Ena.Checked = tecs[3].Enabled;

                    cbAutostartEnable.Checked = laser.AutoStart;

                    if (laser.AutoStart)
                    {
                        buttonLaserOn.Enabled = false;
                        buttonLaserOff.Enabled = false;
                    }
                    buttonAbort.Enabled = laser.AutoStart;
                    buttonRestart.Enabled = laser.AutoStart;
                    cbCdrh.Enabled = laser.AutoStart && coboltMonitor.LoggedIn;
                    cbCdrh.Checked = laser.Cdrh;
                    cbDirectInput.Enabled = laser.AutoStart && !cbCdrh.Checked && coboltMonitor.LoggedIn;
                    cbDirectInput.Checked = laser.DirectInput;

                    if (laser.AutoStart)
                    {
                        lbAutostartState.Items[1] = "Waiting for Temp";
                        if (cbCdrh.Checked)
                            lbAutostartState.Items[2] = "Waiting for Key";
                        else if (cbDirectInput.Checked)
                            lbAutostartState.Items[2] = "Waiting for 5V Input";
                        else
                            lbAutostartState.Items[2] = "---";
                        lbAutostartState.Items[3] = "Warming Up";
                    }
                    else
                    {
                        lbAutostartState.Items[1] = "---";
                        lbAutostartState.Items[2] = "---";
                        lbAutostartState.Items[3] = "---";
                    }

                    lblMode.Text = laser.AutoStart ? (cbCdrh.Checked ? "CDRH Mode" : "OEM Mode") : "Manual Mode";

                    cbTec1Ena.Enabled = coboltMonitor.LoggedIn && !laser.AutoStart;
                    cbTec2Ena.Enabled = coboltMonitor.LoggedIn && !laser.AutoStart;
                    cbTec3Ena.Enabled = coboltMonitor.LoggedIn && !laser.AutoStart;
                    cbTec4Ena.Enabled = coboltMonitor.LoggedIn && !laser.AutoStart;
                    cbTec1Running.Enabled = coboltMonitor.LoggedIn && !laser.AutoStart;
                    cbTec2Running.Enabled = coboltMonitor.LoggedIn && !laser.AutoStart;
                    cbTec3Running.Enabled = coboltMonitor.LoggedIn && !laser.AutoStart;
                    cbTec4Running.Enabled = coboltMonitor.LoggedIn && !laser.AutoStart;

                    // Laser Current set point.
                    tbSetCurrent.Text = laser.LaserCurrentSetPoint.ToString("0.000",new CultureInfo("en-US"));

                    // Laser Power set point.
                    tbSetPower.Text = laser.LaserPowerSetPoint.ToString("0.0", new CultureInfo("en-US"));

                    // Laser autostart warm up current set point.
                    tbSetWarmUpCurrent.Text = laser.WarmUpCurrent.ToString("0.000", new CultureInfo("en-US"));

                    // Laser Warm Up Delay setting.
                    tbSetWarmUpTime.Text = laser.WarmUpTime.ToString("0", new CultureInfo("en-US"));
                    break;

                case CommandId.GetSerialNumber:
                    tsSerNo.Text = laser.SerialNumber.ToString();
                    break;

                case CommandId.GetFirmwareVersion:
                    tsFwVer.Text = laser.FirmwareVersion;
                    break;

                case CommandId.GetFirmwareBuildDate:
                    toolTip1.SetToolTip(statusStrip1, "Firmware Date: " + laser.FirmwareDate + "\nFirmware Build Date: " + laser.FirmwareBuildDate);
                    break;
            }
        }

        private void ParseLedBits(int bits)
        {
            switch (bits & 3)
            {
                case 0: _led1Blinks = false; pbLed1.Image = _led1Off; break;
                case 1: _led1Blinks = false; pbLed1.Image = _led1; break;
                case 2: _led1Blinks = true; break;
            }
            bits >>= 2;

            switch (bits & 3)
            {
                case 0: _led2Blinks = false; pbLed2.Image = _led2Off; break;
                case 1: _led2Blinks = false; pbLed2.Image = _led2; break;
                case 2: _led2Blinks = true; break;
            }
            bits >>= 2;

            switch (bits & 3)
            {
                case 0: _led3Blinks = false; pbLed3.Image = _led3Off; break;
                case 1: _led3Blinks = false; pbLed3.Image = _led3; break;
                case 2: _led3Blinks = true; break;
            }
            bits >>= 2;

            switch (bits & 3)
            {
                case 0: _led4Blinks = false; pbLed4.Image = _led4Off; break;
                case 1: _led4Blinks = false; pbLed4.Image = _led4; break;
                case 2: _led4Blinks = true; break;
            }
            bits >>= 2;
        }
        
        private void SetReadTempToolTip(TextBox tb, int idx, Tec[] tecs)
        {
            toolTip1.SetToolTip(tb, "Actual TEC" + (idx + 1) + " temperature\rDiff: " + tecs[idx].StrDiff + "\rDelta: " + tecs[idx].StrDelta);
            if (tecs[idx].Locked)
                tb.ForeColor = Color.LightGreen;
            else
                tb.ForeColor = Color.HotPink;
        }

        private void ShowOpMode()
        {
            lbAutostartState.SelectedIndex = (int)laser.OperatingMode;
        }

        #region Menu Events
        private void connectToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Connect();
        }

        private void disconnectToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Disconnect();
            laser.Disconnect();
            connectToolStripMenuItem.Enabled = true;
            disconnectToolStripMenuItem.Enabled = false;
            _log.Connected = laser.IsConnected();
        }

        private void loginToolStripMenuItem_Click(object sender, EventArgs e)
        {
            LogInDlg _logIn = new LogInDlg();
            if (_logIn.ShowDialog(this) == DialogResult.OK)
            {
                if (_logIn.InputString.Equals("1960"))
                {
                    SetAuthorizedState(true);
                }
                else
                {
                    MessageBox.Show("Incorrect password");
                    SetAuthorizedState(false);
                    coboltMonitor.LoggedIn = false;
                }
            }
        }

        private void logoutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetAuthorizedState(false);
        }

        private void aboutCoboltMonitorToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AboutDlg about = new AboutDlg();
            about.ShowDialog(this);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void communicationsLogToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (!_log.Visible)
            {
                _log.Show();
            }
            _log.BringToFront();
        }

        private void optionsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OptionsDlg _opts = new OptionsDlg();
            _opts.SerialBaudRate = SerialBaudRate;
            _opts.SerialPortName = SerialPortName;
            if (_opts.ShowDialog(this) == DialogResult.OK)
            {
                SerialPortName = _opts.SerialPortName;
                SerialBaudRate = _opts.SerialBaudRate;
                UpdateStatusBar();
            }
            Properties.Settings.Default.SerialBaudRate = SerialBaudRate;
            Properties.Settings.Default.SerialPortName = SerialPortName;
            Properties.Settings.Default.Save();
        }
        #endregion

        private void Connect()
        {
            laser = coboltMonitor.ConnectLaser(SerialPortName, SerialBaudRate);
            if (laser != null)
            {
                laser.LaserUpdatedEvent += new EventHandler<LaserUpdatedEvent>(UpdateGui);

                _ipLog = new IPLog2(laser,false);
                _PidMon = new PIDMonitor(laser);
                _parLog = new ParLog(laser);
                _log = new LogTerm(laser);
                _neMon = new NEMon(laser);

                coboltMonitor.LoggedInEvent += new EventHandler<LoggedIn>(_log.LoggedIn_Received);
                SetAuthorizedState(coboltMonitor.LoggedIn);

                if (laser.Connect(SerialPortName, SerialBaudRate))
                {
                    rbSetPower.Enabled = true;
                    tbSetPower.Enabled = true;
                    buttonClearFault.Enabled = true;
                    btnSetpoints.Enabled = true;
                    communicationsLogToolStripMenuItem.Enabled = true; 
                    connectToolStripMenuItem.Enabled = false;
                    disconnectToolStripMenuItem.Enabled = true;
                    _log.Connected = laser.IsConnected();
                }
            }
            else
            {
                MessageBox.Show("Could not connect with laser");
            }
        }
        private void Disconnect()
        {
            communicationsLogToolStripMenuItem.Enabled = false;
            connectToolStripMenuItem.Enabled = true;
            disconnectToolStripMenuItem.Enabled = false;

            // Enable some Tools menu choices.
            pIDMonitorToolStripMenuItem.Enabled = false;
            iPLoggingCurveToolStripMenuItem.Enabled = false;
            parameterLoggingToolStripMenuItem.Enabled = false;
            nEMonitorToolStripMenuItem.Enabled = false;

            // Enable some GUI elements.
            btnSetpoints.Enabled = false;
            buttonRestart.Enabled = false;
            buttonAbort.Enabled = false;
            buttonClearFault.Enabled = false;
            buttonLaserOff.Enabled = false;
            buttonLaserOn.Enabled = false;
            cbTec1Ena.Enabled = false;
            cbTec2Ena.Enabled = false;
            cbTec3Ena.Enabled = false;
            cbTec4Ena.Enabled = false;
            cbTec1Running.Enabled = false;
            cbTec2Running.Enabled = false;
            cbTec3Running.Enabled = false;
            cbTec4Running.Enabled = false;
            nudTec1SetTemp.Enabled = false;
            nudTec2SetTemp.Enabled = false;
            nudTec3SetTemp.Enabled = false;
            nudTec4SetTemp.Enabled = false;
            buttonSaveCmd.Enabled = false;
            buttonCmdFile.Enabled = false;
            tbLastCmdFile.Enabled = false;
            labelRmsPercent.Visible = false;
            tbRmsPercent.Visible = false;
            labelRmsVoltage.Visible = false;
            tbRmsVoltage.Visible = false;
            rbSetPower.Enabled = false;
            rbSetCurrent.Enabled = false;
            tbSetPower.Enabled = false;
            tbSetCurrent.Enabled = false;
            tbCalibratePower.Enabled = false;
            tbSetWarmUpCurrent.Enabled = false;
            tbSetWarmUpTime.Enabled = false;
            cbAutostartEnable.Enabled = false;
            cbCdrh.Enabled = false;
            cbDirectInput.Enabled = false;

            toolStripStatusLabel3.Visible = false;
            toolStripStatusLabelHours.Visible = false;
        }

        private void SetAuthorizedState(bool enabled)
        {
            coboltMonitor.LoggedIn = enabled;
            loginToolStripMenuItem.Enabled = !enabled;
            logoutToolStripMenuItem.Enabled = enabled;

            if (laser == null)
                return;

            if (!laser.IsConnected())
                return;

            communicationsLogToolStripMenuItem.Enabled = enabled;           

            // Enable some Tools menu choices.
            pIDMonitorToolStripMenuItem.Enabled = enabled;
            iPLoggingCurveToolStripMenuItem.Enabled = enabled;
            parameterLoggingToolStripMenuItem.Enabled = enabled;
            nEMonitorToolStripMenuItem.Enabled = enabled;

            // Enable some GUI elements.
            cbTec1Ena.Enabled = enabled;
            cbTec2Ena.Enabled = enabled;
            cbTec3Ena.Enabled = enabled;
            cbTec4Ena.Enabled = enabled;
            cbTec1Running.Enabled = enabled;
            cbTec2Running.Enabled = enabled;
            cbTec3Running.Enabled = enabled;
            cbTec4Running.Enabled = enabled;
            nudTec1SetTemp.Enabled = enabled;
            nudTec2SetTemp.Enabled = enabled;
            nudTec3SetTemp.Enabled = enabled;
            nudTec4SetTemp.Enabled = enabled;
            buttonSaveCmd.Enabled = enabled;
            buttonCmdFile.Enabled = enabled;
            tbLastCmdFile.Enabled = enabled;
            labelRmsPercent.Visible = enabled;
            tbRmsPercent.Visible = enabled;
            labelRmsVoltage.Visible = enabled;
            tbRmsVoltage.Visible = enabled;
            rbSetCurrent.Enabled = enabled;
            tbSetCurrent.Enabled = enabled;
            tbCalibratePower.Enabled = enabled;
            tbSetWarmUpCurrent.Enabled = enabled;
            tbSetWarmUpTime.Enabled = enabled;
            cbAutostartEnable.Enabled = enabled;
            cbCdrh.Enabled = enabled;
            cbDirectInput.Enabled = enabled;

            toolStripStatusLabel3.Visible = enabled;
            toolStripStatusLabelHours.Visible = enabled;

            if (!enabled)
            {
                if (_PidMon != null)
                {
                    if (_PidMon.Visible)
                        _PidMon.Close();
                }
                if (_ipLog != null)
                {
                    if (_ipLog.Visible)
                        _ipLog.Close();
                }
            }

            if (_log != null)
            {
                if (_log.Visible)
                    _log.AuthChg();
            }
            if (laser != null)
            {
                laser.GetValues = true;
            }
        }

        private void UpdateStatusBar()
        {
            tsPort.Text = SerialPortName;
            tsBaudRate.Text = SerialBaudRate.ToString();
        }

        /// <summary>
        /// Gets a stream to selected file.
        /// </summary>
        /// <param name="Name">name of file to get stream for</param>
        /// <returns>A stream to the file</returns>
        public Stream GetEmbeddedFile(string Name)
        {
            return Assembly.GetExecutingAssembly().GetManifestResourceStream(Name);
        }

        #region TEC Control Events
        private void cbTec1Ena_Click(object sender, EventArgs e)
        {
            if (cbTec1Ena.Checked)
                laser.EnableTec1();
            else
                laser.DisableTec1();
            laser.GetTec1EnabledState();
        }

        private void cbTec2Ena_Click(object sender, EventArgs e)
        {
            if (cbTec2Ena.Checked)
                laser.EnableTec2();
            else
                laser.DisableTec2();
            laser.GetTec2EnabledState();
        }

        private void cbTec3Ena_Click(object sender, EventArgs e)
        {
            if (cbTec3Ena.Checked)
                laser.EnableTec3();
            else
                laser.DisableTec3();
            laser.GetTec3EnabledState();
        }

        private void cbTec4Ena_Click(object sender, EventArgs e)
        {
            if (cbTec4Ena.Checked)
                laser.EnableTec4();
            else
                laser.DisableTec4();
            laser.GetTec4EnabledState();
        }

        private void cbTec1Running_Click(object sender, EventArgs e)
        {
            if (cbTec1Running.Checked)
                laser.RunTec1();
            else
                laser.StopTec1();
        }

        private void cbTec2Running_Click(object sender, EventArgs e)
        {
            if (cbTec2Running.Checked)
                laser.RunTec2();
            else
                laser.StopTec2();
        }

        private void cbTec3Running_Click(object sender, EventArgs e)
        {
            if (cbTec3Running.Checked)
                laser.RunTec3();
            else
                laser.StopTec3();
        }

        private void cbTec4Running_Click(object sender, EventArgs e)
        {
            if (cbTec4Running.Checked)
                laser.RunTec4();
            else
                laser.StopTec4();
        }
        #endregion

        #region TEC SetTemp Events
        private void nudTec1SetTemp_ValueChanged(object sender, EventArgs e)
        {
            laser.SetTec1Temp((sender as NumericUpDown).Value.ToString(new CultureInfo("en-US")));
            laser.GetTec1Temp();
        }

        private void nudTec2SetTemp_ValueChanged(object sender, EventArgs e)
        {
            laser.SetTec2Temp((sender as NumericUpDown).Value.ToString(new CultureInfo("en-US")));
            laser.GetTec2Temp();
        }

        private void nudTec3SetTemp_ValueChanged(object sender, EventArgs e)
        {
            laser.SetTec3Temp((sender as NumericUpDown).Value.ToString(new CultureInfo("en-US")));
            laser.GetTec3Temp();
        }

        private void nudTec4SetTemp_ValueChanged(object sender, EventArgs e)
        {
            laser.SetTec4Temp((sender as NumericUpDown).Value.ToString(new CultureInfo("en-US")));
            laser.GetTec4Temp();
        }
        #endregion

        #region KeyPress Events
        private void tbSetPower_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (laser == null)
                return;

            float res;
            string text = tbSetPower.Text;

            if (e.KeyChar == (char)Keys.Return)
            {
                if (float.TryParse(text, NumberStyles.Float, new CultureInfo("en-US"), out res))
                {
                    if(!laser.SetPower(res))
                    {
                        MessageBox.Show("Laser power should be between " + laser.PowerRangeMin + " and " + laser.PowerRangeMax + " mW", "Bad Value");
                    }
                }
                else
                {
                    MessageBox.Show("'" + text + "' is not a valid floating point number.", "Bad number format");
                }

                laser.GetPower();
                e.Handled = true;
            }
        }

        /// <summary>
        /// Handles the KeyPress event of the tbSetCurrent control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.Windows.Forms.KeyPressEventArgs"/> instance containing the event data.</param>
        private void tbSetCurrent_KeyPress(object sender, KeyPressEventArgs e)
        {
            float res;
            string text = tbSetCurrent.Text;

            if (e.KeyChar == (char)Keys.Return)
            {
                if (float.TryParse(text, NumberStyles.Float, new CultureInfo("en-US"), out res))
                {
                    // Check that new set-point is within range.
                    if (!laser.SetCurrent(res))
                    {
                        MessageBox.Show("'" + text + "' is not a valid set-point value.", "Laser current should be between " +
                            laser.CurrentRangeMin + " and " + laser.CurrentRangeMax + " mA");
                    }
                }
                else
                {
                    MessageBox.Show("'" + text + "' is not a valid floating point number.", "Bad number format");
                }

                laser.GetCurrent();

                e.Handled = true;
            }
        }

        /// <summary>
        /// Handles the KeyPress event of the tbSetWarmUpCurrent control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.Windows.Forms.KeyPressEventArgs"/> instance containing the event data.</param>
        private void tbSetWarmUpCurrent_KeyPress(object sender, KeyPressEventArgs e)
        {
            float res;
            string text = tbSetWarmUpCurrent.Text;

            if (e.KeyChar == (char)Keys.Return)
            {
                if (float.TryParse(text, NumberStyles.Float, new CultureInfo("en-US"), out res))
                {
                    // Check that new set-point is within range.
                    if (!laser.SetWarmUpCurrent(res))
                    {
                        MessageBox.Show("'" + text + "' is not a valid set-point value.", "Laser current should be between " +
                             laser.CurrentRangeMin + " and " + laser.CurrentRangeMax + " mA");
                    }
                }
                else
                {
                    MessageBox.Show("'" + text + "' is not a valid floating point number.", "Bad number format");
                }

                laser.GetWarmUpCurrent();

                e.Handled = true;
            }
        }

        private void tbSetWarmUpTime_KeyPress(object sender, KeyPressEventArgs e)
        {
            int res;
            string text = tbSetWarmUpTime.Text;

            if (e.KeyChar == (char)Keys.Return)
            {
                if (int.TryParse(text, out res))
                {
                    // Check that new set-point is within range.
                    if (!laser.SetWarmUpDelay(res))
                    {
                        MessageBox.Show("'" + text + "' is not a valid set-point value.", "Value must be greater than or equal to 0.");
                    }
                }
                else
                {
                    MessageBox.Show("'" + text + "' is not a valid integer number.", "Bad number format");
                }

                laser.GetWarmUpDelay();

                e.Handled = true;
            }
        }

        private void tbCalibratePower_KeyPress(object sender, KeyPressEventArgs e)
        {
            float res;
            string text = tbCalibratePower.Text;

            if (e.KeyChar == (char)Keys.Return)
            {
                if (float.TryParse(text, NumberStyles.Float, new CultureInfo("en-US"), out res))
                {
                    // Check that new power value is within range.
                    if (!laser.SetCalibratePower(res))
                    {
                        MessageBox.Show("Laser power should be between " + laser.PowerRangeMin + " and " + laser.PowerRangeMax + " mW", "Bad Value");
                    }
                }
                else
                {
                    MessageBox.Show("'" + text + "' is not a valid floating point number.", "Bad number format");
                    tbCalibratePower.Text = "";
                }
                tbCalibratePower.Text = "";
                e.Handled = true;
            }
        }
        #endregion

        #region Radio button Events
        private void rbSetPower_MouseClick(object sender, MouseEventArgs e)
        {
            if (laser != null)
            {
                laser.EnterCpMode();
            }
        }

        private void rbSetCurrent_MouseClick(object sender, MouseEventArgs e)
        {
            if (laser != null)
            {
                laser.EnterCcMode();
            }
        }
        #endregion


        #region Listbox Events
        private void lbAutostartState_SelectedIndexChanged(object sender, EventArgs e)
        {
            lbAutostartState.SelectedIndex = (int)laser.OperatingMode;
        }
        #endregion




        private void pIDMonitorToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (!_PidMon.Visible)
            {
                _PidMon.Show();
            }
            _PidMon.BringToFront();
        }

        private void iPLoggingCurveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (!_ipLog.Visible)
            {
                if (!laser.LaserEnabled)
                    MessageBox.Show("Laser needs to be on to get a ip log", "CoboltMonitor"); 
                
                if (_ipLog.IsDisposed)
                    _ipLog = new IPLog2(laser,false);
                _ipLog.Show();
            }
            _ipLog.BringToFront();
        }

        private void parameterLoggingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (!_parLog.Visible)
            {
                _parLog.Show();
            }
            _parLog.BringToFront();
        }

        private void nEMonitorToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (!_neMon.Visible)
            {
                _neMon.Show();
            }
            _neMon.BringToFront();
        }

        private void buttonAbort_Click(object sender, EventArgs e)
        {
            laser.Abort();
        }

        private void buttonRestart_Click(object sender, EventArgs e)
        {
            laser.Restart();
        }

        /// <summary>
        /// Save commands to file
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonSaveCmd_Click(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.Filter = "Text Only (*.txt)|*.txt|All files (*.*)|*.*";
            sfd.FilterIndex = 0;

            if (sfd.ShowDialog() == DialogResult.OK)
            {
                string message = String.Empty;
                if (!laser.SaveCommand(sfd.FileName, out message))
                {
                    MessageBox.Show(message, "Error saving file");
                }
            }

        }

        private void buttonCmdFile_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Filter = "Text Only (*.txt)|*.txt|All files (*.*)|*.*";
            ofd.FilterIndex = 0;

            if (ofd.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    string message = String.Empty;
                    if (File.Exists(ofd.FileName))
                    {
                        if (!laser.LoadCommand(ofd.FileName, out message))
                        {
                            MessageBox.Show(message, "Script File Error");
                        }
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message,"Exception loading command file");
                }
            }
        }

        private void btnSetpoints_Click(object sender, EventArgs e)
        {
            laser.GetSetPoints();
        }

        /// <summary>
        /// Handles the Click event of the buttonLaserOn control.
        /// 
        /// Turns on laser.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void buttonLaserOn_Click(object sender, EventArgs e)
        {
            laser.SetLaserOn();
        }

        /// <summary>
        /// Handles the Click event of the buttonLaserOff control.
        /// 
        /// Turns off laser.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void buttonLaserOff_Click(object sender, EventArgs e)
        {
            laser.SetLaserOff();
        }

        private void buttonClearFault_Click(object sender, EventArgs e)
        {
            laser.ClearFault();
        }

        /// <summary>
        /// Handles the CheckedChanged event of the cbAutostartEnable control.
        /// Enables or disables the laser driver autostart function.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void cbAutostartEnable_Click(object sender, EventArgs e)
        {
            if (cbAutostartEnable.Checked)
                laser.EnableAutoStart();
            else
                laser.DisableAutoStart();
        }

        private void cbCdrh_Click(object sender, EventArgs e)
        {
            if (cbCdrh.Checked)
                laser.EnableKeySwitch();
            else
                laser.DisableKeySwitch();
        }

        private void cbDirectInput_CheckedChanged(object sender, EventArgs e)
        {
            if (cbDirectInput.Checked)
                laser.Enable5VDirectInputStart();
            else
                laser.Disable5VDirectInputStart();
        }

        /// <summary>
        /// Gets laser fault description.
        /// </summary>
        public string GetLaserFault(int fault)
        {
            return _strLaserFault[fault];
        }

        /// <summary>
        /// Gets TEC fault description.
        /// </summary>
        public string GetTecFault(int fault)
        {
            return _strTecFault[fault];
        }


    }
}
