using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.IO;
using System.IO.Ports;
using System.Configuration;
using System.Globalization;
using System.Threading;
using System.Resources;
using System.Reflection;

using CoboltMonitor.Laser;
using CoboltMonitor.Gui;

namespace CoboltMonitorMLD.Gui
{
	public partial class MainForm : Form
	{
        #region Enumerations
	    public enum OperatingMode
	    {
		    mode_off,
		    //        mode_wait_tec,
		    mode_wait_key,
		    mode_completed,
		    mode_onoff_mod,
		    mode_modulation,
		    mode_fault,
		    mode_aborted,
		    mode_count
	    };

	    public enum TecFaults
	    {
		    fault_none,
		    fault_upper_rail,
		    fault_lower_rail,
		    fault_no_conv,
		    fault_NTC_bad_conn,
		    fault_NTC_short,
		    fault_driver_overcurr,
		    fault_driver_undervolt,
		    fault_driver_overtemp,
		    fault_count
	    };

	    public enum LaserFaults
	    {
		    fault_none,
		    fault_no_completion,
		    fault_interlock,
		    fault_current_clamp,
		    fault_count
	    };
	    #endregion
        
        #region Delegates
		/// <summary>
		/// Helper delegate to update UI and be thread safe.
		/// </summary>
        public delegate void UpdateSerialData(Object objSender, LaserUpdatedEvent dataReceivedEventArgs);
        #endregion

		#region Instance variables

        CoboltMonitor.Laser.CoboltMonitor coboltMonitor;
        Gen6Laser laser;

        LogTerm _log;
        PIDMonitor1Tec _PidMon;
        ParLogMLD _parLog;

        Image _ledGray;
        Image _ledGreen;
        Image _ledOrange;
        Image _ledRed;

        Image _led1;
        Image _led1Off;
        Image _led2;
        Image _led2Off;
        Image _led3;
        Image _led4;
        Image _led4Off;

        bool _led1Blinks = false;
        bool _led2Blinks = false;
        bool _led4Blinks = false;
        private int _blinkTimer = 0;
        private String serialPortName;
        private int serialBaudRate;

        string[] _strTecFault;
        string[] _strLaserFault;

        public String SerialPortName
        {
            get
            {
                return serialPortName;
            }
            set
            {
                serialPortName = value;
            }
        }

        public int SerialBaudRate
        {
            get
            {
                return serialBaudRate;
            }
            set
            {
                serialBaudRate = value;
            }
        }

		#endregion

		#region Constructors
		/// <summary>
		/// Initializes a new instance of the <see cref="MainForm"/> class.
		/// </summary>
		public MainForm()
		{
            Thread.CurrentThread.CurrentUICulture = Thread.CurrentThread.CurrentCulture = new CultureInfo("en-US");
            InitializeComponent();
            coboltMonitor = new CoboltMonitor.Laser.CoboltMonitor();
            _ledGray = Image.FromStream(GetEmbeddedFile("CoboltMonitorMLD.Resources.LedGray.bmp"));
            _led1 = _ledGreen = Image.FromStream(GetEmbeddedFile("CoboltMonitorMLD.Resources.LedGreen.bmp"));
            _led2 = _led3 = _ledOrange = Image.FromStream(GetEmbeddedFile("CoboltMonitorMLD.Resources.LedOrange.bmp"));
            _led4 = _ledRed = Image.FromStream(GetEmbeddedFile("CoboltMonitorMLD.Resources.LedRed.bmp"));
            _led1Off = _ledGray;
            _led2Off = _ledGray;
            _led4Off = _ledGray;
            pbLed1.Image = _led1Off;
            pbLed2.Image = _led2Off;
            pbLed4.Image = _led4Off;
            Disconnect();
            SerialBaudRate = Properties.Settings.Default.SerialBaudRate;
            SerialPortName = Properties.Settings.Default.SerialPortName;
            UpdateStatusBar();

            _strTecFault = new string[(int)TecFaults.fault_count];
            _strTecFault[(int)TecFaults.fault_none] = "None";
            _strTecFault[(int)TecFaults.fault_upper_rail] = "Upper rail";
            _strTecFault[(int)TecFaults.fault_lower_rail] = "Lower rail";
            _strTecFault[(int)TecFaults.fault_no_conv] = "No convergence";
            _strTecFault[(int)TecFaults.fault_NTC_bad_conn] = "NTC bad conn";
            _strTecFault[(int)TecFaults.fault_NTC_short] = "NTC shorted";
            _strTecFault[(int)TecFaults.fault_driver_overcurr] = "Over-current";
            _strTecFault[(int)TecFaults.fault_driver_undervolt] = "Under-voltage";
            _strTecFault[(int)TecFaults.fault_driver_overtemp] = "Over-temp";

            _strLaserFault = new string[(int)LaserFaults.fault_count];
            _strLaserFault[(int)LaserFaults.fault_none] = "None";
            _strLaserFault[(int)LaserFaults.fault_no_completion] = "Unstable";
            _strLaserFault[(int)LaserFaults.fault_interlock] = "Interlock open";
            _strLaserFault[(int)LaserFaults.fault_current_clamp] = "Current Clamp";
        }
		#endregion

        /// <summary>
        /// Gets a stream to selected file.
        /// </summary>
        /// <param name="Name">name of file to get stream for</param>
        /// <returns>A stream to the file</returns>
        public Stream GetEmbeddedFile(string Name)
        {
            return Assembly.GetExecutingAssembly().GetManifestResourceStream(Name);
        }

		#region Form Events
		/// <summary>
		/// Handles the FormClosing event of the MainForm control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.Windows.Forms.FormClosingEventArgs"/> instance containing the event data.</param>
		private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
		{
            if (laser != null)
                laser.Disconnect();
		}

		/// <summary>
		/// Handles the Load event of the MainForm control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void MainForm_Load(object sender, EventArgs e)
		{
            this.Text += " - " + Application.ProductVersion.ToString();
        }

        #endregion

		#region Menu Events
		// The Login choice was selected from the File menu.
		private void loginToolStripMenuItem_Click(object sender, EventArgs e)
		{
            LogInDlg _logIn = new LogInDlg();
            if (_logIn.ShowDialog(this) == DialogResult.OK)
            {
                if (_logIn.InputString.Equals("1960"))
                {
                    SetAuthorizedState(true);
                }
                else
                {
                    MessageBox.Show("Incorrect password");
                    SetAuthorizedState(false);
                    coboltMonitor.LoggedIn = false;
                }
            }
		}

		// The Logout choice was selected from the File menu.
		private void logoutToolStripMenuItem_Click(object sender, EventArgs e)
		{
			SetAuthorizedState(false);
		}

		private void optionsToolStripMenuItem_Click(object sender, EventArgs e)
		{
            OptionsDlg _opts = new OptionsDlg();
            _opts.SerialBaudRate = SerialBaudRate;
            _opts.SerialPortName = SerialPortName;
            if (_opts.ShowDialog(this) == DialogResult.OK)
            {
                SerialPortName = _opts.SerialPortName;
                SerialBaudRate = _opts.SerialBaudRate;
                UpdateStatusBar();
            }
            Properties.Settings.Default.SerialBaudRate = SerialBaudRate;
            Properties.Settings.Default.SerialPortName = SerialPortName;
            Properties.Settings.Default.Save();
		}

		private void exitToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Close();
		}

		private void connectToolStripMenuItem_Click(object sender, EventArgs e)
		{
            Connect();
		}

		private void disconnectToolStripMenuItem_Click(object sender, EventArgs e)
		{
            Disconnect();
            laser.Disconnect();
            pbLed1.Image = _led1Off;
            connectToolStripMenuItem.Enabled = true;
            disconnectToolStripMenuItem.Enabled = false;
            _log.Connected = laser.IsConnected();
		}

        private void pIDMonitorToolStripMenuItem_Click(object sender, EventArgs e)
		{
			if (!_PidMon.Visible) {
				_PidMon.Show();
			}
            _PidMon.BringToFront();
        }

        private void communicationsLogToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (!_log.Visible)
            {
                _log.Show();
            }
            _log.BringToFront();
        }
		private void parameterLoggingToolStripMenuItem_Click(object sender, EventArgs e)
		{
			if (!_parLog.Visible) {
				_parLog.Show();
			}
            _parLog.BringToFront();
        }

		private void aboutCoboltMonitorToolStripMenuItem_Click(object sender, EventArgs e)
		{
            AboutDlg about = new AboutDlg();
            about.ShowDialog(this);
        }
		#endregion

		#region TEC Control Events
		private void cbTecEna_Click(object sender, EventArgs e)
		{
			if (cbTecEna.Checked)
				laser.EnableTec();
			else
				laser.DisableTec();
			laser.GetTecEnabledState();
		}

		private void cbTecRunning_Click(object sender, EventArgs e)
		{
			if (cbTecRunning.Checked)
                laser.RunTec();
			else
                laser.StopTec();
		}
		#endregion

		#region TEC SetTemp Events
		private void nudTecSetTemp_ValueChanged(object sender, EventArgs e)
		{
            laser.SetTecTemp((sender as NumericUpDown).Value.ToString(new CultureInfo("en-US")));
            laser.GetTecTemp();
        }
		#endregion

		#region KeyPress Events
		private void tbModPower_KeyPress(object sender, KeyPressEventArgs e)
		{
            if (laser == null)
                return; 
            
            float res;
			string text = tbModPower.Text;

			if (e.KeyChar == (char)Keys.Return) {
				if (float.TryParse(text, NumberStyles.Float, new CultureInfo("en-US"), out res)) {
					// Check that new set-point is within range.
                    if (!laser.SetLaserModPower(res))
                    {
						MessageBox.Show("Laser modulated power should be between " + laser.ModPowerRangeMin + " and " + laser.ModPowerRangeMax + " mW", "Bad Value");
					}
				} else {
					MessageBox.Show("'" + text + "' is not a valid floating point number.", "Bad number format");
				}

                laser.GetLaserModPower();

				e.Handled = true;
			}
		}

        private void tbSetPower_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (laser == null)
                return;

            float res;
            string text = tbSetPower.Text;

            if (e.KeyChar == (char)Keys.Return)
            {
                if (float.TryParse(text, NumberStyles.Float, new CultureInfo("en-US"), out res))
                {
                    if (!laser.SetPower(res))
                    {
                        MessageBox.Show("Laser power should be between " + laser.PowerRangeMin + " and " + laser.PowerRangeMax + " mW", "Bad Value");
                    }
                }
                else
                {
                    MessageBox.Show("'" + text + "' is not a valid floating point number.", "Bad number format");
                }

                laser.GetPower();
                e.Handled = true;
            }
        }
      
		/// <summary>
		/// Handles the KeyPress event of the tbSetCurrent control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.Windows.Forms.KeyPressEventArgs"/> instance containing the event data.</param>
        private void tbSetCurrent_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (laser == null)
                return;

            float res;
            string text = tbSetCurrent.Text;

            if (e.KeyChar == (char)Keys.Return)
            {
                if (float.TryParse(text, NumberStyles.Float, new CultureInfo("en-US"), out res))
                {
                    // Check that new set-point is within range.
                    if (!laser.SetCurrent(res))
                    {
                        MessageBox.Show("'" + text + "' is not a valid set-point value.", "Laser current should be between " +
                            laser.CurrentRangeMin + " and " + laser.CurrentRangeMax + " mA");
                    }
                }
                else
                {
                    MessageBox.Show("'" + text + "' is not a valid floating point number.", "Bad number format");
                }

                laser.GetCurrent();

                e.Handled = true;
            }
        }

        private void tbCalibratePower_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (laser == null)
                return;

            float res;
            string text = tbCalibratePower.Text;

            if (e.KeyChar == (char)Keys.Return)
            {
                if (float.TryParse(text, NumberStyles.Float, new CultureInfo("en-US"), out res))
                {
                    // Check that new power value is within range.
                    if (!laser.SetCalibratePower(res))
                    {
                        MessageBox.Show("Laser power should be between " + laser.PowerRangeMin + 
                            " and " + laser.PowerRangeMax + " mW", "Bad Value");
                    }
                }
                else
                {
                    MessageBox.Show("'" + text + "' is not a valid floating point number.", "Bad number format");
                    tbCalibratePower.Text = "";
                }
                tbCalibratePower.Text = "";
                e.Handled = true;
            }
        }
		#endregion

		#region Button Click Events
		/// <summary>
		/// Handles the Click event of the buttonClearFault control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void buttonClearFault_Click(object sender, EventArgs e)
		{
            laser.ClearFault();
        }

		/// <summary>
		/// Handles the Click event of the btnSetpoints control.
		/// Causes all the set points to be requested from the laser driver.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void btnSetpoints_Click(object sender, EventArgs e)
		{
            laser.GetSetPoints();
        }

		/// <summary>
		/// Handles the Click event of the buttonAbort control.
		/// 
		/// Aborts autostart sequence.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void buttonAbort_Click(object sender, EventArgs e)
		{
            laser.Abort();
        }

		/// <summary>
		/// Handles the Click event of the buttonRestart control.
		/// 
		/// Restarts autostart sequence.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void buttonRestart_Click(object sender, EventArgs e)
		{
            laser.Restart();
        }

		/// <summary>
		/// Handles the Click event of the buttonLaserOn control.
		/// 
		/// Turns on laser.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void buttonLaserOn_Click(object sender, EventArgs e)
		{
            laser.SetLaserOn();
        }

		/// <summary>
		/// Handles the Click event of the buttonLaserOff control.
		/// 
		/// Turns off laser.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void buttonLaserOff_Click(object sender, EventArgs e)
		{
            laser.SetLaserOff();
        }

		/// <summary>
		/// Handles the Click event of the buttonCmdFile control.
		/// 
		/// Opens a script file and sends the contained commands to laser driver.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void buttonCmdFile_Click(object sender, EventArgs e)
		{
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Filter = "Text Only (*.txt)|*.txt|All files (*.*)|*.*";
            ofd.FilterIndex = 0;

            if (ofd.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    string message = String.Empty;
                    if (File.Exists(ofd.FileName))
                    {
                        if (!laser.LoadCommand(ofd.FileName, out message))
                        {
                            MessageBox.Show(message, "Script File Error");
                        }
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message, "Exception loading command file");
                }
            } 
 		}

		/// <summary>
		/// Handles the Click event of the buttonSaveCmd control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void buttonSaveCmd_Click(object sender, EventArgs e)
		{
			SaveFileDialog sfd = new SaveFileDialog();
            sfd.Filter = "Text Only (*.txt)|*.txt|All files (*.*)|*.*";
            sfd.FilterIndex = 0;

            if (sfd.ShowDialog() == DialogResult.OK)
            {
                string message = String.Empty;
                if (!laser.SaveCommand(sfd.FileName, out message))
                {
                    MessageBox.Show(message, "Error saving file");
                }
            }
        }
		#endregion

		#region Radio button Events
		private void rbModulation_MouseClick(object sender, MouseEventArgs e)
		{
            laser.EnterModMode();
		}

		private void rbSetPower_MouseClick(object sender, MouseEventArgs e)
		{
			laser.EnterCpMode();
		}

		private void rbSetCurrent_MouseClick(object sender, MouseEventArgs e)
		{
            laser.EnterCcMode();
        }

		private void rbDigital_MouseClick(object sender, MouseEventArgs e)
		{
            laser.SetDigitalModulationEnabledState("1");
            laser.SetAnalogModulationEnabledState("0");
		}

		private void rbAnalog_MouseClick(object sender, MouseEventArgs e)
		{
            laser.SetDigitalModulationEnabledState("0");
            laser.SetAnalogModulationEnabledState("1");
		}

		private void rbDigitalAndAnalog_MouseClick(object sender, MouseEventArgs e)
		{
            laser.SetDigitalModulationEnabledState("1");
            laser.SetAnalogModulationEnabledState("1");
		}

		private void rb50ohm_MouseClick(object sender, MouseEventArgs e)
		{
            laser.SetAnalogLowImpedanceState("1");
		}

		private void rb1kOhm_MouseClick(object sender, MouseEventArgs e)
		{
            laser.SetAnalogLowImpedanceState("0");
		}
		#endregion

		#region Listbox Events
		private void lbAutostartState_SelectedIndexChanged(object sender, EventArgs e)
		{
			lbAutostartState.SelectedIndex = (int)laser.OperatingMode;
		}
		#endregion

		#region CheckBox Events
		/// <summary>
		/// Handles the CheckedChanged event of the cbAutostartEnable control.
		/// Enables or disables the laser driver autostart function.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void cbAutostartEnable_Click(object sender, EventArgs e)
		{
            if (cbAutostartEnable.Checked)
                laser.EnableAutoStart();
            else
                laser.DisableAutoStart();
        }

		private void cbOnOffMod_Click(object sender, EventArgs e)
		{
			if (cbOnOffMod.Checked)
				laser.EnterOnOffMode();
			else
				laser.ExitOnOffMode();
		}

		/// <summary>
		/// Handles the Click event of the cbCdrh control.
		/// 
		/// Toggles laser driver CDRH mode.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void cbCdrh_Click(object sender, EventArgs e)
		{
            if (cbCdrh.Checked)
                laser.EnableKeySwitch();
            else
                laser.DisableKeySwitch();
        }
		#endregion

        private void Connect()
        {
            laser = (Gen6Laser)coboltMonitor.ConnectLaserMLD(SerialPortName, SerialBaudRate);
            if (laser != null)
            {
                laser.LaserUpdatedEvent += new EventHandler<LaserUpdatedEvent>(UpdateGui);

                _PidMon = new PIDMonitor1Tec(laser);
                _parLog = new ParLogMLD(laser);
                _log = new LogTerm(laser);

                coboltMonitor.LoggedInEvent += new EventHandler<LoggedIn>(_log.LoggedIn_Received);
                SetAuthorizedState(coboltMonitor.LoggedIn);

                if (laser.Connect(SerialPortName, SerialBaudRate))
                {
                    rbSetPower.Enabled = true;
                    tbSetPower.Enabled = true;
                    buttonClearFault.Enabled = true;
                    btnSetpoints.Enabled = true;
                    communicationsLogToolStripMenuItem.Enabled = true;
                    connectToolStripMenuItem.Enabled = false;
                    disconnectToolStripMenuItem.Enabled = true;
                    _log.Connected = laser.IsConnected();
                }
            }
            else
            {
                MessageBox.Show("Could not connect with laser");
            }
        }
        private void Disconnect()
        {
            communicationsLogToolStripMenuItem.Enabled = false;
            connectToolStripMenuItem.Enabled = true;
            disconnectToolStripMenuItem.Enabled = false;

            // Enable some Tools menu choices.
            pIDMonitorToolStripMenuItem.Enabled = false;
            parameterLoggingToolStripMenuItem.Enabled = false;

            // Enable some GUI elements.
            buttonRestart.Enabled = false;
            buttonAbort.Enabled = false;
            buttonLaserOff.Enabled = false;
            cbTecEna.Enabled = false;
            cbTecRunning.Enabled = false;
            nudTecSetTemp.Enabled = false;
            buttonSaveCmd.Enabled = false;
            buttonCmdFile.Enabled = false;
            tbLastCmdFile.Enabled = false;
            tbCalibratePower.Enabled = false;
            cbAutostartEnable.Enabled = false;
            cbCdrh.Enabled = false;

            toolStripStatusLabel3.Visible = false;
            toolStripStatusLabelHours.Visible = false;

            tcTemps.TabPages.Clear();
            tcTemps.TabPages.Add(tpBasePlate);
        }

        private void ShowOpMode()
        {
            lbAutostartState.SelectedIndex = laser.OperatingMode;
        }

        private void SetReadTempToolTip(TextBox tb, int idx, Tec[] tecs)
        {
            toolTip1.SetToolTip(tb, "Actual TEC" + (idx + 1) + " temperature\rDiff: " + tecs[idx].StrDiff + "\rDelta: " + tecs[idx].StrDelta);
            if (tecs[idx].Locked)
                tb.ForeColor = Color.LightGreen;
            else
                tb.ForeColor = Color.HotPink;
        }
        
        private void SetAuthorizedState(bool enabled)
        {
            coboltMonitor.LoggedIn = enabled;
            loginToolStripMenuItem.Enabled = !enabled;
            logoutToolStripMenuItem.Enabled = enabled;

            if (laser == null)
                return;

            if (!laser.IsConnected())
                return;

            communicationsLogToolStripMenuItem.Enabled = enabled;

            // Enable some Tools menu choices.
            pIDMonitorToolStripMenuItem.Enabled = enabled;
            parameterLoggingToolStripMenuItem.Enabled = enabled;

            // Enable some GUI elements.
            cbTecEna.Enabled = enabled;
            cbTecRunning.Enabled = enabled;
            nudTecSetTemp.Enabled = enabled;
            buttonSaveCmd.Enabled = enabled;
            buttonCmdFile.Enabled = enabled;
            tbLastCmdFile.Enabled = enabled;
            rbSetPower.Enabled = true;
            tbSetPower.Enabled = true;
            rbSetCurrent.Enabled = true;
            tbSetCurrent.Enabled = true;
            tbCalibratePower.Enabled = enabled;
            rbModulation.Enabled = true;
            cbAutostartEnable.Enabled = enabled;
            cbCdrh.Enabled = enabled;

            tcTemps.TabPages.Clear();
            tcTemps.TabPages.Add(tpBasePlate);
            if (enabled)
            {
                tcTemps.TabPages.Add(tpTec);
                tcTemps.TabPages.Add(tpBoard);
            } 
            
            toolStripStatusLabel3.Visible = enabled;
            toolStripStatusLabelHours.Visible = enabled;

            if (!enabled)
            {
                if (_PidMon != null)
                {
                    if (_PidMon.Visible)
                        _PidMon.Close();
                }
            }

            if (laser != null)
            {
                laser.GetValues = true;
            }
        }

        private void UpdateStatusBar()
        {
            tsPort.Text = SerialPortName;
            tsBaudRate.Text = SerialBaudRate.ToString();
        }

        private void UpdateGui(object sender, LaserUpdatedEvent e)
        {
            if (this.InvokeRequired == true)
            {
                UpdateSerialData updateHandler = new UpdateSerialData(UpdateGui);
                this.BeginInvoke(updateHandler, new Object[] { sender, e });
                return;
            }

            Tec[] tecs = laser.GetTecs();
            pbLed1.Image = _led1;
            switch (e.CommandResponse.CommandId)
            {
                case CommandId.GetCmdName:
                    tbLastCmdFile.Text = laser.CmdName;
                    break;

                case CommandId.GetOperatingHours:
                    toolStripStatusLabelHours.Text = laser.OperatingHours;
                    break;

                case CommandId.GetCalibrateLdCurr:
                    toolStripStatusLabelCalibrated.Text = "Calibrate Laser Driver";
                    break;
                
                case CommandId.LaserDiodeEnable:
                    buttonLaserOn.Enabled = false;
                    buttonLaserOff.Enabled = true;
                    break;

                case CommandId.LaserDisable:
                    buttonLaserOn.Enabled = true;
                    buttonLaserOff.Enabled = false;
                    break;

                case CommandId.GetAutostartState:
                    cbAutostartEnable.Checked = laser.AutoStart;
                    buttonLaserOn.Enabled = !laser.AutoStart;
                    buttonLaserOff.Enabled = !laser.AutoStart;
                    cbCdrh.Enabled = laser.AutoStart && coboltMonitor.LoggedIn;
                    buttonAbort.Enabled = laser.AutoStart;
                    buttonRestart.Enabled = laser.AutoStart;
                    break;

                case CommandId.DisableAutostart:
                    buttonLaserOn.Enabled = true;
                    buttonLaserOff.Enabled = false;
                    break;

                case CommandId.GetLaserCurrent:			// Laser current set-point
                    tbSetCurrent.Text = laser.LaserCurrentSetPoint.ToString("0.0", new CultureInfo("en-US"));
                    break;

                case CommandId.ReadLaserCurrent:		// Laser current value
                    tbCurrent.Text = laser.LaserCurrent.ToString("0.0", new CultureInfo("en-US"));
                    break;
              
                case CommandId.ReadTecTemperature:
                    tbTecTemp.Text = tecs[0].Temp.ToString("0.00", new CultureInfo("en-US"));
                    SetReadTempToolTip(tbTecTemp, 0, tecs);
                    break;

                case CommandId.GetTecEnabledState:
                    cbTecEna.Checked = tecs[0].Enabled;
                    break;

                case CommandId.ReadTecDrive:
                    tbTecDrive.Text = tecs[0].Drive.ToString("0.00", new CultureInfo("en-US"));
                    break;

                case CommandId.GetLaserPower:
                    tbSetPower.Text = laser.LaserPowerSetPoint.ToString("0.0", new CultureInfo("en-US"));
                    break;

                case CommandId.GetLaserModPower:
                    tbModPower.Text = laser.LaserModPower.ToString("0.0", new CultureInfo("en-US"));
                    break;

                case CommandId.GetAnalogLowImpedanceState:
                    if (laser.Analog50Ohm)
                    {
                        rb50ohm.Select();
                    }
                    else if (laser.Analog1kOhm)
                    {
                        rb1kOhm.Select();
                    }
                    break;
                
                case CommandId.ReadAll:
                    // Base Plate temperature reading.
                    lbBasePlateTemp.Text = laser.BasePlateTemp.ToString("0.00", new CultureInfo("en-US")) + " C";

                    // Board temperature reading.
                    lbBoardTemp.Text = laser.BoardTemp.ToString("0.00", new CultureInfo("en-US")) + " C";

                    tbTecTemp.Text = tecs[0].Temp.ToString("0.00", new CultureInfo("en-US"));
                    SetReadTempToolTip(tbTecTemp, 0, tecs);

                    // TEC drive readings.
                    tbTecDrive.Text = tecs[0].Drive.ToString("0.00", new CultureInfo("en-US"));

                    // Laser Current, Power, Diode Volt, RMS %, RMS Volt, Photo Diode Vol readings.
                    tbCurrent.Text = laser.LaserCurrent.ToString("0.0", new CultureInfo("en-US"));
                    tbPower.Text = laser.LaserPower.ToString("0.0", new CultureInfo("en-US"));

                    // Keep text color upated (laser current).
                    if (!rbSetCurrent.Checked || buttonLaserOn.Enabled)
                        tbCurrent.ForeColor = Color.Yellow;
                    else if (Math.Abs(laser.LaserCurrentSetPoint - laser.LaserCurrent) <= 0.5f)
                        tbCurrent.ForeColor = Color.LightGreen;
                    else
                        tbCurrent.ForeColor = Color.HotPink;

                    // Keep text color upated (laser power).
                    if (!rbSetPower.Checked || buttonLaserOn.Enabled)
                        tbPower.ForeColor = Color.Yellow;
                    else if (Math.Abs(laser.LaserPowerSetPoint - laser.LaserPower) / laser.LaserPowerSetPoint <= 0.03f)
                        tbPower.ForeColor = Color.LightGreen;
                    else
                        tbPower.ForeColor = Color.HotPink;

                    // Laser Operating mode.
                    OperatingMode om = (OperatingMode)laser.OperatingMode;
                    if (om >= (int)OperatingMode.mode_off && laser.OperatingMode < (int)OperatingMode.mode_count)
                    {
                        ShowOpMode();
                        switch (om)
                        {
                            case OperatingMode.mode_completed:
                            case OperatingMode.mode_modulation:
                            case OperatingMode.mode_onoff_mod:
                                buttonLaserOff.Enabled = true;
                                buttonLaserOn.Enabled = false;
                                break;

                            default:
                                buttonLaserOff.Enabled = false;
                                buttonLaserOn.Enabled = !laser.AutoStart; // true;
                                break;
                        }
                        // Laser LED On indication.
                        switch (om)
                        {
                            case OperatingMode.mode_off:
                            case OperatingMode.mode_fault:
                            case OperatingMode.mode_aborted:
                                pbLed2.Image = _led2Off;
                                break;

                            case OperatingMode.mode_completed:
                            case OperatingMode.mode_onoff_mod:
                            case OperatingMode.mode_modulation:
                                pbLed2.Image = _led2;
                                break;

                            default:
                                pbLed2.Image = _led2Off;
                                break;
                        }
                        // On/Off Modulation mode handling.
                        if (om == OperatingMode.mode_onoff_mod)
                        {
                            cbOnOffMod.Checked = true;
                            rbModulation.Enabled = false;
                            rbSetPower.Enabled = false;
                            rbSetCurrent.Enabled = false;
                            tbSetPower.Enabled = false;
                            tbSetCurrent.Enabled = false;
                        }
                        else
                        {
                            cbOnOffMod.Checked = false;
                            rbModulation.Enabled = true;
                            rbSetPower.Enabled = true;
                            rbSetCurrent.Enabled = true;
                            tbSetPower.Enabled = true;
                            tbSetCurrent.Enabled = true;
                        }
                    }

                    // Constant Current or Constant Power.
                    if (laser.ConstantCurrent)
                    {
                        rbSetCurrent.Checked = true;
                    }
                    else if (laser.ConstantPower)
                    {
                        rbSetPower.Checked = true;
                    }
                    else if (laser.Modulation)
                    {
                        rbModulation.Checked = true;
                    }

                    if (tecs[0].Fault >= (int)TecFaults.fault_none && tecs[0].Fault < (int)TecFaults.fault_count)
                    {
                        lbTecFault.Text = GetTecFault(tecs[0].Fault);
                        lbTecFault.ForeColor = (tecs[0].Fault == (int)TecFaults.fault_none) ? Color.Green : Color.Red;
                    }
                    else
                    {
                        lbTecFault.Text = "Unknown";
                        lbTecFault.ForeColor = Color.Red;
                    }

                    if (laser.LaserFault >= (int)LaserFaults.fault_none && laser.LaserFault < (int)LaserFaults.fault_count)
                    {
                        lbLaserFault.Text = GetLaserFault(laser.LaserFault);
                        lbLaserFault.ForeColor = (laser.LaserFault == (int)LaserFaults.fault_none || laser.LaserFault == (int)LaserFaults.fault_current_clamp) ? Color.Green : Color.Red;
                    }
                    else
                    {
                        lbLaserFault.Text = "Unknown";
                        lbLaserFault.ForeColor = Color.Red;
                    }


                    if (laser.LaserFault > (int)LaserFaults.fault_none && laser.LaserFault != (int)LaserFaults.fault_current_clamp)
                        buttonClearFault.Enabled = (tecs[0].Fault != 0);

                    // Set Error LED.
                    if (tecs[0].Fault + laser.LaserFault > 0)
                    {
                        pbLed4.Image = _led4;
                    }
                    else
                    {
                        pbLed4.Image = _led4Off;
                    }

                    cbTecRunning.Checked = tecs[0].Running;

                    // Modulation type.
                    switch (laser.ModulationSettings & 3)
                    {
                        case 0:
                            rbDigital.Checked = false;
                            rbAnalog.Checked = false;
                            rbDigitalAndAnalog.Checked = false;
                            break;

                        case 1:
                            rbDigital.Checked = true;
                            break;

                        case 2:
                            rbAnalog.Checked = true;
                            break;

                        case 3:
                            rbDigitalAndAnalog.Checked = true;
                            break;
                    }

                    // Analog input impedance.
                    if ((laser.ModulationSettings & 4) != 0)
                        rb50ohm.Checked = true;
                    else
                        rb1kOhm.Checked = true;

                    if ((_blinkTimer++ & 2) == 2)
                    {
                        if (_led1Blinks) pbLed1.Image = _led1;
                        if (_led2Blinks) pbLed2.Image = _led2;
                        if (_led4Blinks) pbLed4.Image = _led4;
                    }
                    else
                    {
                        if (_led1Blinks) pbLed1.Image = _led1Off;
                        if (_led2Blinks) pbLed2.Image = _led2Off;
                        if (_led4Blinks) pbLed4.Image = _led4Off;
                    }
                    break;

                case CommandId.ReadAllSetPoints:
                    nudTecSetTemp.Value = (decimal)tecs[0].SetPoint;
                    cbTecEna.Checked = tecs[0].Enabled;

                    cbAutostartEnable.Checked = laser.AutoStart;

                    if (laser.AutoStart)
                    {
                        buttonLaserOn.Enabled = false;
                        buttonLaserOff.Enabled = false;
                    }
                    buttonAbort.Enabled = laser.AutoStart;
                    buttonRestart.Enabled = laser.AutoStart;
                    cbCdrh.Enabled = laser.AutoStart && coboltMonitor.LoggedIn;
                    cbCdrh.Checked = laser.Cdrh;

                    if (laser.AutoStart)
                    {
                        if (cbCdrh.Checked)
                            lbAutostartState.Items[1] = "Waiting for Key";
                        else
                            lbAutostartState.Items[1] = "---";
                    }
                    else
                    {
                        lbAutostartState.Items[1] = "---";
                    }

                    lblMode.Text = laser.AutoStart ? (cbCdrh.Checked ? "CDRH Mode" : "OEM Mode") : "Manual Mode";

                    cbTecEna.Enabled = coboltMonitor.LoggedIn && !laser.AutoStart;
                    cbTecRunning.Enabled = coboltMonitor.LoggedIn && !laser.AutoStart;

                    // Laser Current set point.
                    tbSetCurrent.Text = laser.LaserCurrentSetPoint.ToString("0.0", new CultureInfo("en-US"));

                    // Laser Power set point.
                    tbSetPower.Text = laser.LaserPowerSetPoint.ToString("0.0", new CultureInfo("en-US"));
                    break;

                case CommandId.GetSerialNumber:
                    tsSerNo.Text = laser.SerialNumber.ToString();
                    break;

                case CommandId.GetFirmwareVersion:
                    tsFwVer.Text = laser.FirmwareVersion;
                    break;

                case CommandId.GetFirmwareBuildDate:
                    toolTip1.SetToolTip(statusStrip1, "Firmware Date: " + laser.FirmwareDate + "\nFirmware Build Date: " + laser.FirmwareBuildDate);
                    break;
            }
        }
        
        private void tcTemps_DrawItem(object sender, DrawItemEventArgs e)
		{
			Graphics g = e.Graphics;
			Pen p = new Pen(Color.Blue);
			Font font = tcTemps.TabPages[e.Index].Font;
			Color c = tcTemps.TabPages[e.Index].ForeColor;
			String text = tcTemps.TabPages[e.Index].Text;
			SolidBrush brush = new SolidBrush(c);
			g.FillRectangle(new SolidBrush(Color.LightGray), tcTemps.GetTabRect(e.Index));//.DrawRectangle(p, tcTemps.GetTabRect(e.Index));
			g.DrawString(text, font, brush, (RectangleF)tcTemps.GetTabRect(e.Index));
		}

        /// <summary>
        /// Gets laser fault description.
        /// </summary>
        public string GetLaserFault(int fault)
        {
            return _strLaserFault[fault];
        }

        /// <summary>
        /// Gets TEC fault description.
        /// </summary>
        public string GetTecFault(int fault)
        {
            return _strTecFault[fault];
        }

	}
}