using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Data;
using System.Text;
using System.Windows.Forms;

namespace Graph
{
	public partial class UsGraph : PictureBox //UserControl
	{
		#region Enumerations
		public enum GridStyle
		{
			None,
			Independent,
			XBased,
			YBased,
			Scope
		};
		#endregion

		#region Member variables
		private bool _needsUpdate = false;
		private Bitmap _bm = null;
		private Bitmap _bmGrid = null;
		private Graphics _gr = null;
		private Graphics _grGrid = null;
		private float _xtick = 1.0F;
		private float _ytick = 1.0F;
		private GridStyle _gridStyle = GridStyle.Independent;
		private float _xmin = -10.0F;
		private float _xmax = 10.0F;
		private float _ymin = -10.0F;
		private float _ymax = 10.0F;
		private RectangleF _dataRect;
		private RectangleF _screenRect;
		private Color _gridColor = Color.Gray;
		#endregion

		#region Properties
		[DefaultValue(UsGraph.GridStyle.Independent)]
		public GridStyle Style
		{
			get { return _gridStyle; }
			set { _gridStyle = value; _needsUpdate = true; updateGraph(); Invalidate(); Update(); }
		}

		[DefaultValue(1.0)]
		public float XTick
		{
			get { return _xtick; }
			set { _xtick = value; _needsUpdate = true; updateGraph(); Invalidate(); Update(); }
		}

		[DefaultValue(1.0)]
		public float YTick
		{
			get { return _ytick; }
			set { _ytick = value; _needsUpdate = true; updateGraph(); Invalidate(); Update(); }
		}

		[DefaultValue(-10.0)]
		public float XMin
		{
			get { return _xmin; }
			set { _xmin = value; _needsUpdate = true; updateGraph(); Invalidate(); Update(); }
		}

		[DefaultValue(10.0)]
		public float XMax
		{
			get { return _xmax; }
			set { _xmax = value; _needsUpdate = true; updateGraph(); Invalidate(); Update(); }
		}

		[DefaultValue(-10.0)]
		public float YMin
		{
			get { return _ymin; }
			set { _ymin = value; _needsUpdate = true; updateGraph(); Invalidate(); Update(); }
		}

		[DefaultValue(10.0)]
		public float YMax
		{
			get { return _ymax; }
			set { _ymax = value; _needsUpdate = true; updateGraph(); Invalidate(); Update(); }
		}

		public Color GridColor
		{
			get { return _gridColor; }
			set { _gridColor = value; _needsUpdate = true; updateGraph(); Invalidate(); Update(); }
		}
		#endregion

		#region Constructors
		public UsGraph()
		{
			InitializeComponent();

			SetStyle(ControlStyles.DoubleBuffer | ControlStyles.UserPaint | ControlStyles.AllPaintingInWmPaint, true);
			UpdateStyles();

			this.BackColorChanged += new EventHandler(UsGraph_BackColorChanged);
			updateGraph();
		}
		#endregion

		#region Helper functions
		public PointF Transform(RectangleF from, RectangleF to, PointF pt_src)
		{
			PointF pt_dst = new PointF();

			pt_dst.X = (((pt_src.X - from.Left) / from.Width) * to.Width) + to.Left;
			pt_dst.Y = (((from.Top - pt_src.Y) / from.Height) * to.Height) + to.Bottom;

			return pt_dst;
		}

		public void SetHorizontalBounds(float xmin, float xmax)
		{
			_xmin = xmin;
			_xmax = xmax;
			_dataRect = new RectangleF(_xmin, _ymin, _xmax - _xmin, _ymax - _ymin);
			//_needsUpdate = true;
		}

		public void SetVerticalBounds(float ymin, float ymax)
		{
			_ymin = ymin;
			_ymax = ymax;
			_dataRect = new RectangleF(_xmin, _ymin, _xmax - _xmin, _ymax - _ymin);
			//_needsUpdate = true;
		}

		private void updateGraph()
		{
			if (_needsUpdate)
			{
				_needsUpdate = false;
				if (_grGrid != null)
					_grGrid.Dispose();
				_bmGrid = null;
			}

			if (_bmGrid == null)
			{
				_bmGrid = new Bitmap(this.Width, this.Height);
				_grGrid = Graphics.FromImage(_bmGrid);
				drawGrid();
			}

			_bm = new Bitmap(this.Width, this.Height);
			if (_gr != null)
				_gr.Dispose();
			_gr = Graphics.FromImage(_bm);
		}

		public void Clear()
		{
			_gr.Clear(Color.Transparent);
			Invalidate();
			//updateGraph();
		}

		public void SaveImage(string filename)
		{
			_bm.Save(filename/*, ImageFormat.Jpeg*/);
		}

		public void Line(PointF p0, PointF p1, Color color)
		{
			_gr.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
			PointF _p0 = Transform(_dataRect, _screenRect, p0);
			PointF _p1 = Transform(_dataRect, _screenRect, p1);
			_gr.DrawLine(new Pen(color), _p0, _p1);
		}

		public void Lines(PointF[] p)
		{
		}

		public void Axis(PointF p0, PointF p1, Color color)
		{
			_gr.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
			PointF _p0 = Transform(_dataRect, _screenRect, p0);
			PointF _p1 = Transform(_dataRect, _screenRect, p1);
			Pen pen = new Pen(color);
			pen.DashPattern = new float[] { 2.0f, 2.0f };
			_gr.DrawLine(pen, _p0, _p1);
		}

		public void Legend(string text, Font font, Brush brush, PointF p)
		{
			PointF _p = Transform(_dataRect, _screenRect, p);
			_gr.DrawString(text, font, brush, _p); 
		}

		private void drawGrid()
		{
			// Clear screen
			_grGrid.Clear(this.BackColor);

			// Get pen for drawing the grid.
			Pen gridPen = new Pen(this.GridColor);

			float xt = (float)(Size.Width / (XMax - XMin));
			float [] dpx = new float [2];
			dpx[0] = 1.0f;
			dpx[1] = xt * 2 - 1.0f;

			float yt = (float)(Size.Height / (YMax - YMin));
			float[] dpy = new float[2];
			dpy[0] = 1.0f;
			dpy[1] = yt * 2 - 1.0f;

			_screenRect = _grGrid.VisibleClipBounds;
			//_screenRect.Width -= 3;
			//_screenRect.Height -= 3;
			_dataRect = new RectangleF(_xmin, _ymin, _xmax - _xmin, _ymax - _ymin);

			float xstart = (float)Math.Floor(_xmin / _xtick) * _xtick;
			float ystart = (float)Math.Floor(_ymin / _ytick) * _ytick;

			// Draw grid.
			switch (_gridStyle)
			{
				case GridStyle.Scope:
					gridPen.DashPattern = dpx;
					for (float x = _xmin; x < _xmax; x += _xtick)
					{
						PointF p0 = Transform(_dataRect, _screenRect, new PointF(x, _dataRect.Bottom));
						PointF p1 = Transform(_dataRect, _screenRect, new PointF(x, _dataRect.Top));
						_grGrid.DrawLine(gridPen, p0, p1);
					}
					for (float y = _ymin; y < _ymax; y += _ytick)
					{
						PointF p0 = Transform(_dataRect, _screenRect, new PointF(_dataRect.Left, y));
						PointF p1 = Transform(_dataRect, _screenRect, new PointF(_dataRect.Right, y));
						_grGrid.DrawLine(gridPen, p0, p1);
					}
					break;

				case GridStyle.Independent:
					gridPen.DashPattern = dpy;
					for (float x = _xmin; x < _xmax; x += _xtick)
					{
						PointF p0 = Transform(_dataRect, _screenRect, new PointF(x, _dataRect.Bottom));
						PointF p1 = Transform(_dataRect, _screenRect, new PointF(x, _dataRect.Top));
						_grGrid.DrawLine(gridPen, p0, p1);
					}
					gridPen.DashPattern = dpx;
					for (float y = _ymin; y < _ymax; y += _ytick)
					{
						PointF p0 = Transform(_dataRect, _screenRect, new PointF(_dataRect.Left, y));
						PointF p1 = Transform(_dataRect, _screenRect, new PointF(_dataRect.Right, y));
						_grGrid.DrawLine(gridPen, p0, p1);
					}
					break;

				case GridStyle.XBased:
					gridPen.DashPattern = dpx;
					for (float x = _xmin; x < _xmax; x += _xtick)
					{
						PointF p0 = Transform(_dataRect, _screenRect, new PointF(x, _dataRect.Bottom));
						PointF p1 = Transform(_dataRect, _screenRect, new PointF(x, _dataRect.Top));
						_grGrid.DrawLine(gridPen, p0, p1);
					}
					for (float y = _ymin; y < _ymax; y += _xtick)
					{
						PointF p0 = Transform(_dataRect, _screenRect, new PointF(_dataRect.Left, y));
						PointF p1 = Transform(_dataRect, _screenRect, new PointF(_dataRect.Right, y));
						_grGrid.DrawLine(gridPen, p0, p1);
					}
					//for (int x = 0; x < _bm.Width; x += _bm.Width / _xdivs)
					//{
					//    _gr.DrawLine(gridPen, x, 0, x, _bm.Height);
					//}
					//for (int y = 0; y < _bm.Height; y += _bm.Width / _xdivs)
					//{
					//    _gr.DrawLine(gridPen, 0, y, _bm.Width, y);
					//}
					break;

				case GridStyle.YBased:
					for (float x = _xmin; x < _xmax; x += _ytick)
					{
						PointF p0 = Transform(_dataRect, _screenRect, new PointF(x, _dataRect.Bottom));
						PointF p1 = Transform(_dataRect, _screenRect, new PointF(x, _dataRect.Top));
						_grGrid.DrawLine(gridPen, p0, p1);
					}
					for (float y = _ymin; y < _ymax; y += _ytick)
					{
						PointF p0 = Transform(_dataRect, _screenRect, new PointF(_dataRect.Left, y));
						PointF p1 = Transform(_dataRect, _screenRect, new PointF(_dataRect.Right, y));
						_grGrid.DrawLine(gridPen, p0, p1);
					}
					//for (int x = 0; x < _bm.Width; x += _bm.Height / _ydivs)
					//{
					//    _gr.DrawLine(gridPen, x, 0, x, _bm.Height);
					//}
					//for (int y = 0; y < _bm.Height; y += _bm.Height / _ydivs)
					//{
					//    _gr.DrawLine(gridPen, 0, y, _bm.Width, y);
					//}
					break;
			}

			gridPen.Dispose();
		}
		#endregion

		private void UsGraph_Paint(object sender, PaintEventArgs e)
		{
			Graphics g = e.Graphics;
			g.DrawImageUnscaled(_bmGrid, 0, 0);
			g.DrawImageUnscaled(_bm, 0, 0);
		}

		private void UsGraph_ClientSizeChanged(object sender, EventArgs e)
		{
			updateGraph();
		}

		void UsGraph_BackColorChanged(object sender, EventArgs e)
		{
			_needsUpdate = true;
			updateGraph();
		}
	}
}